Package: mlr3fda
Title: Extending 'mlr3' to Functional Data Analysis
Version: 0.6.0
Authors@R: c(
    person("Maximilian", "Mücke", , "muecke.maximilian@gmail.com", role = c("aut", "cre"),
           comment = c(ORCID = "0009-0000-9432-9795")),
    person("Sebastian", "Fischer", , "sebf.fischer@gmail.com", role = "aut",
           comment = c(ORCID = "0000-0002-9609-3197")),
    person("Fabian", "Scheipl", , "fabian.scheipl@googlemail.com", role = "ctb",
           comment = c(ORCID = "0000-0001-8172-3603")),
    person("Bernd", "Bischl", , "bernd_bischl@gmx.net", role = "ctb",
           comment = c(ORCID = "0000-0001-6002-6980"))
  )
Description: Extends the 'mlr3' ecosystem to functional analysis by adding
    support for irregular and regular functional data as defined in the
    'tf' package.  The package provides 'PipeOps' for preprocessing
    functional columns and for extracting scalar features, thereby
    allowing standard machine learning algorithms to be applied
    afterwards. Available operations include simple functional features
    such as the mean or maximum, smoothing, interpolation, flattening, and
    functional 'PCA'.
License: LGPL-3
URL: https://mlr3fda.mlr-org.com, https://github.com/mlr-org/mlr3fda
BugReports: https://github.com/mlr-org/mlr3fda/issues
Depends: mlr3 (>= 1.3.0), mlr3pipelines (>= 0.5.2), R (>= 4.1.0)
Imports: checkmate, data.table (>= 1.18.0), lgr, mlr3misc (>= 0.19.0),
        paradox, R6, tf (>= 0.4.0)
Suggests: FDboost, lme4, mboost, rpart, testthat (>= 3.3.0),
        tsfeatures, wavelets, withr
Config/roxygen2/markdown: TRUE
Config/roxygen2/r6: TRUE
Config/roxygen2/version: 8.0.0
Config/testthat/edition: 3
Encoding: UTF-8
LazyData: true
Collate: 'zzz.R' 'PipeOpFDABsignal.R' 'PipeOpFDACor.R'
        'PipeOpFDADepth.R' 'PipeOpFDADerive.R' 'PipeOpFDAExtract.R'
        'PipeOpFDAFlatten.R' 'PipeOpFDAFourier.R' 'PipeOpFDAInterpol.R'
        'PipeOpFDARandomEffect.R' 'PipeOpFDARegister.R'
        'PipeOpFDAScaleRange.R' 'PipeOpFDASmooth.R'
        'PipeOpFDATsfeatures.R' 'PipeOpFDAWavelets.R' 'PipeOpFDAZoom.R'
        'PipeOpFPCA.R' 'TaskClassif_phoneme.R' 'TaskRegr_dti.R'
        'TaskRegr_fuel.R' 'bibentries.R' 'datasets.R' 'hash_input.R'
NeedsCompilation: no
Packaged: 2026-06-01 10:02:53 UTC; mmuecke
Author: Maximilian Mücke [aut, cre] (ORCID:
    <https://orcid.org/0009-0000-9432-9795>),
  Sebastian Fischer [aut] (ORCID:
    <https://orcid.org/0000-0002-9609-3197>),
  Fabian Scheipl [ctb] (ORCID: <https://orcid.org/0000-0001-8172-3603>),
  Bernd Bischl [ctb] (ORCID: <https://orcid.org/0000-0001-6002-6980>)
Maintainer: Maximilian Mücke <muecke.maximilian@gmail.com>
Repository: CRAN
Date/Publication: 2026-06-01 10:40:02 UTC
Built: R 4.5.2; ; 2026-06-01 13:27:59 UTC; unix

Package: mlr3fda
Title: Extending 'mlr3' to Functional Data Analysis
Version: 0.6.0
Authors@R: c(
    person("Maximilian", "Mücke", , "muecke.maximilian@gmail.com", role = c("aut", "cre"),
           comment = c(ORCID = "0009-0000-9432-9795")),
    person("Sebastian", "Fischer", , "sebf.fischer@gmail.com", role = "aut",
           comment = c(ORCID = "0000-0002-9609-3197")),
    person("Fabian", "Scheipl", , "fabian.scheipl@googlemail.com", role = "ctb",
           comment = c(ORCID = "0000-0001-8172-3603")),
    person("Bernd", "Bischl", , "bernd_bischl@gmx.net", role = "ctb",
           comment = c(ORCID = "0000-0001-6002-6980"))
  )
Description: Extends the 'mlr3' ecosystem to functional analysis by adding
    support for irregular and regular functional data as defined in the
    'tf' package.  The package provides 'PipeOps' for preprocessing
    functional columns and for extracting scalar features, thereby
    allowing standard machine learning algorithms to be applied
    afterwards. Available operations include simple functional features
    such as the mean or maximum, smoothing, interpolation, flattening, and
    functional 'PCA'.
License: LGPL-3
URL: https://mlr3fda.mlr-org.com, https://github.com/mlr-org/mlr3fda
BugReports: https://github.com/mlr-org/mlr3fda/issues
Depends: mlr3 (>= 1.3.0), mlr3pipelines (>= 0.5.2), R (>= 4.1.0)
Imports: checkmate, data.table (>= 1.18.0), lgr, mlr3misc (>= 0.19.0),
        paradox, R6, tf (>= 0.4.0)
Suggests: FDboost, lme4, mboost, rpart, testthat (>= 3.3.0),
        tsfeatures, wavelets, withr
Config/roxygen2/markdown: TRUE
Config/roxygen2/r6: TRUE
Config/roxygen2/version: 8.0.0
Config/testthat/edition: 3
Encoding: UTF-8
LazyData: true
Collate: 'zzz.R' 'PipeOpFDABsignal.R' 'PipeOpFDACor.R'
        'PipeOpFDADepth.R' 'PipeOpFDADerive.R' 'PipeOpFDAExtract.R'
        'PipeOpFDAFlatten.R' 'PipeOpFDAFourier.R' 'PipeOpFDAInterpol.R'
        'PipeOpFDARandomEffect.R' 'PipeOpFDARegister.R'
        'PipeOpFDAScaleRange.R' 'PipeOpFDASmooth.R'
        'PipeOpFDATsfeatures.R' 'PipeOpFDAWavelets.R' 'PipeOpFDAZoom.R'
        'PipeOpFPCA.R' 'TaskClassif_phoneme.R' 'TaskRegr_dti.R'
        'TaskRegr_fuel.R' 'bibentries.R' 'datasets.R' 'hash_input.R'
NeedsCompilation: no
Packaged: 2026-06-01 10:02:53 UTC; mmuecke
Author: Maximilian Mücke [aut, cre] (ORCID:
    <https://orcid.org/0009-0000-9432-9795>),
  Sebastian Fischer [aut] (ORCID:
    <https://orcid.org/0000-0002-9609-3197>),
  Fabian Scheipl [ctb] (ORCID: <https://orcid.org/0000-0001-8172-3603>),
  Bernd Bischl [ctb] (ORCID: <https://orcid.org/0000-0001-6002-6980>)
Maintainer: Maximilian Mücke <muecke.maximilian@gmail.com>
Repository: CRAN
Date/Publication: 2026-06-01 10:40:02 UTC
Built: R 4.6.0; ; 2026-06-01 15:27:08 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(hash_input,tfd_irreg)
S3method(hash_input,tfd_reg)
export(PipeOpFDABsignal)
export(PipeOpFDACor)
export(PipeOpFDADepth)
export(PipeOpFDADerive)
export(PipeOpFDAExtract)
export(PipeOpFDAFlatten)
export(PipeOpFDAFourier)
export(PipeOpFDAInterpol)
export(PipeOpFDARandomEffect)
export(PipeOpFDARegister)
export(PipeOpFDAScaleRange)
export(PipeOpFDASmooth)
export(PipeOpFDATsfeatures)
export(PipeOpFDAWavelets)
export(PipeOpFDAZoom)
export(PipeOpFPCA)
import(R6)
import(checkmate)
import(data.table)
import(mlr3)
import(mlr3misc)
import(mlr3pipelines)
import(paradox)
importFrom(tf,tf_approx_fill_extend)
importFrom(tf,tf_approx_linear)
importFrom(tf,tf_approx_locf)
importFrom(tf,tf_approx_nocb)
importFrom(tf,tf_approx_spline)
importFrom(utils,bibentry)

# mlr3fda

Package website: [release](https://mlr3fda.mlr-org.com/) \|
[dev](https://mlr3fda.mlr-org.com/dev/)

Extending mlr3 to functional data.

<!-- badges: start -->

[![RCMD
Check](https://github.com/mlr-org/mlr3fda/actions/workflows/rcmdcheck.yaml/badge.svg)](https://github.com/mlr-org/mlr3fda/actions/workflows/rcmdcheck.yaml)
[![CRAN
status](https://www.r-pkg.org/badges/version/mlr3fda)](https://CRAN.R-project.org/package=mlr3fda)
[![StackOverflow](https://img.shields.io/badge/stackoverflow-mlr3-orange.svg)](https://stackoverflow.com/questions/tagged/mlr3)
[![Mattermost](https://img.shields.io/badge/chat-mattermost-orange.svg)](https://lmmisld-lmu-stats-slds.srv.mwn.de/mlr_invite/)
<!-- badges: end -->

## Installation

Install the last release from [CRAN](https://CRAN.R-project.org):

``` r
install.packages("mlr3fda")
```

Install the development version from [GitHub](https://github.com/):

``` r
# install.packages("pak")
pak::pak("mlr-org/mlr3fda")
```

## What is mlr3fda?

The goal of `mlr3fda` is to extend `mlr3` to [functional
data](https://en.wikipedia.org/wiki/Functional_data_analysis). This is
achieved by adding support for functional feature types and providing
preprocessing `PipeOp`s that operates on functional columns. For
representing functional data, the `tfd_reg` and `tfd_irreg` datatypes
from the [tf](https://github.com/tidyfun/tf) package are used and are
available after loading `mlr3fda`:

``` r
library(mlr3fda)
mlr_reflections$task_feature_types[c("tfr", "tfi")]
#>         tfr         tfi 
#>   "tfd_reg" "tfd_irreg"
```

These datatypes can be used to represent regular and irregular
functional data respectively. Currently, `Learner`s that directly
operate on functional data are not available, so it is necessary to
first extract scalar features from the functional columns.

# Quickstart

Here we will start with the predefined `dti` (Diffusion Tensor Imaging)
task, see `tsk("dti")$help()` for more details. Besides scalar columns,
this task also contains two functional columns `cca` and `rcst`.

``` r
task = tsk("dti")
task
#> 
#> ── <TaskRegr> (340x4): Diffusion Tensor Imaging (DTI) ──────────────────────────
#> • Target: pasat
#> • Properties: groups
#> • Features (3):
#>   • tfi (2): cca, rcst
#>   • fct (1): sex
#> • Groups: subject_id
```

To train a model on this task we first need to extract scalar features
from the functions. We illustrate this below by extracting the mean
value.

``` r
po_fmean = po("fda.extract", features = "mean")

task_fmean = po_fmean$train(list(task))[[1L]]
task_fmean$head()
#>    pasat    sex  cca_mean rcst_mean
#> 1:    31 female 0.4493332 0.4968519
#> 2:    31 female 0.4441292 0.4810724
#> 3:    29 female 0.4257795 0.5102722
#> 4:    34 female 0.4418538 0.5453188
#> 5:    37 female 0.4700994 0.5471177
#> 6:    40 female 0.4873356 0.4969408
```

This can be combined with a `Learner` into a `GraphLearner` that first
extracts features and then trains a model.

``` r
# split data into train and test set
ids = partition(task)

# define a Graph and convert it to a GraphLearner
graph = po("fda.extract", features = "mean", drop = TRUE) %>>%
  po("learner", learner = lrn("regr.rpart"))

glrn = as_learner(graph)

# train the graph learner on the train set
glrn$train(task, row_ids = ids$train)

# make predictions on the test set
glrn$predict(task, row_ids = ids$test)
#> 
#> ── <PredictionRegr> for 111 observations: ──────────────────────────────────────
#>  row_ids truth response
#>       11    48 49.99174
#>       12    40 49.99174
#>       13    43 52.42105
#>      ---   ---      ---
#>      324    57 52.42105
#>      325    57 41.30769
#>      326    60 49.99174
```

## Implemented PipeOps

| Key | Label | Packages | Tags |
|:---|:---|:---|:---|
| [fda.bsignal](https://mlr3fda.mlr-org.com/dev/reference/mlr_pipeops_fda.bsignal) | B-spline Feature Extraction | [tf](https://cran.r-project.org/package=tf), [mboost](https://cran.r-project.org/package=mboost), [FDboost](https://cran.r-project.org/package=FDboost) | fda, data transform |
| [fda.cor](https://mlr3fda.mlr-org.com/dev/reference/mlr_pipeops_fda.cor) | Cross-Correlation of Functional Data | [tf](https://cran.r-project.org/package=tf) | fda, data transform |
| [fda.extract](https://mlr3fda.mlr-org.com/dev/reference/mlr_pipeops_fda.extract) | Extracts Simple Features from Functional Columns | [tf](https://cran.r-project.org/package=tf) | fda, data transform |
| [fda.flatten](https://mlr3fda.mlr-org.com/dev/reference/mlr_pipeops_fda.flatten) | Flattens Functional Columns | [tf](https://cran.r-project.org/package=tf) | fda, data transform |
| [fda.fpca](https://mlr3fda.mlr-org.com/dev/reference/mlr_pipeops_fda.fpca) | Functional Principal Component Analysis | [tf](https://cran.r-project.org/package=tf) | fda, data transform |
| [fda.interpol](https://mlr3fda.mlr-org.com/dev/reference/mlr_pipeops_fda.interpol) | Interpolate Functional Columns | [tf](https://cran.r-project.org/package=tf) | fda, data transform |
| [fda.random_effect](https://mlr3fda.mlr-org.com/dev/reference/mlr_pipeops_fda.random_effect) | Extracts Random Effects from Functional Columns | [tf](https://cran.r-project.org/package=tf), [lme4](https://cran.r-project.org/package=lme4) | fda, data transform |
| [fda.scalerange](https://mlr3fda.mlr-org.com/dev/reference/mlr_pipeops_fda.scalerange) | Linearly Transform the Domain of Functional Data | [tf](https://cran.r-project.org/package=tf) | fda, data transform |
| [fda.smooth](https://mlr3fda.mlr-org.com/dev/reference/mlr_pipeops_fda.smooth) | Smoothing Functional Columns | [tf](https://cran.r-project.org/package=tf), stats | fda, data transform |
| [fda.tsfeats](https://mlr3fda.mlr-org.com/dev/reference/mlr_pipeops_fda.tsfeats) | Time Series Feature Extraction | [tf](https://cran.r-project.org/package=tf), [tsfeatures](https://cran.r-project.org/package=tsfeatures) | fda, data transform |
| [fda.wavelets](https://mlr3fda.mlr-org.com/dev/reference/mlr_pipeops_fda.wavelets) | Discrete Wavelet transform features | [tf](https://cran.r-project.org/package=tf), [wavelets](https://cran.r-project.org/package=wavelets) | fda, data transform |
| [fda.zoom](https://mlr3fda.mlr-org.com/dev/reference/mlr_pipeops_fda.zoom) | Zoom In/Out on Functional Columns | [tf](https://cran.r-project.org/package=tf) | fda, data transform |

## Bugs, Questions, Feedback

*mlr3fda* is a free and open source software project that encourages
participation and feedback. If you have any issues, questions,
suggestions or feedback, please do not hesitate to open an “issue” about
it on the GitHub page!

In case of problems / bugs, it is often helpful if you provide a
“minimum working example” that showcases the behaviour (but don’t worry
about this if the bug is obvious).

Please understand that the resources of the project are limited:
response may sometimes be delayed by a few days, and some feature
suggestions may be rejected if they are deemed too tangential to the
vision behind the project.

## Acknowledgements

The development of this R-package was supported by Roche Diagnostics
R&D.

# mlr3fda

Package website: [release](https://mlr3fda.mlr-org.com/) \|
[dev](https://mlr3fda.mlr-org.com/dev/)

Extending mlr3 to functional data.

<!-- badges: start -->

[![RCMD
Check](https://github.com/mlr-org/mlr3fda/actions/workflows/rcmdcheck.yaml/badge.svg)](https://github.com/mlr-org/mlr3fda/actions/workflows/rcmdcheck.yaml)
[![CRAN
status](https://www.r-pkg.org/badges/version/mlr3fda)](https://CRAN.R-project.org/package=mlr3fda)
[![StackOverflow](https://img.shields.io/badge/stackoverflow-mlr3-orange.svg)](https://stackoverflow.com/questions/tagged/mlr3)
[![Mattermost](https://img.shields.io/badge/chat-mattermost-orange.svg)](https://lmmisld-lmu-stats-slds.srv.mwn.de/mlr_invite/)
<!-- badges: end -->

## Installation

Install the last release from [CRAN](https://CRAN.R-project.org):

``` r
install.packages("mlr3fda")
```

Install the development version from [GitHub](https://github.com/):

``` r
# install.packages("pak")
pak::pak("mlr-org/mlr3fda")
```

## What is mlr3fda?

The goal of `mlr3fda` is to extend `mlr3` to [functional
data](https://en.wikipedia.org/wiki/Functional_data_analysis). This is
achieved by adding support for functional feature types and providing
preprocessing `PipeOp`s that operate on functional columns. For
representing functional data, the `tfd_reg` and `tfd_irreg` datatypes
from the [tf](https://github.com/tidyfun/tf) package are used and are
available after loading `mlr3fda`:

``` r
library(mlr3fda)
mlr_reflections$task_feature_types[c("tfr", "tfi")]
#>         tfr         tfi 
#>   "tfd_reg" "tfd_irreg"
```

These datatypes can be used to represent regular and irregular
functional data respectively. Currently, `Learner`s that directly
operate on functional data are not available, so it is necessary to
first extract scalar features from the functional columns.

# Quickstart

Here we will start with the predefined `dti` (Diffusion Tensor Imaging)
task, see `tsk("dti")$help()` for more details. Besides scalar columns,
this task also contains two functional columns `cca` and `rcst`.

``` r
task = tsk("dti")
task
#> 
#> ── <TaskRegr> (340x4): Diffusion Tensor Imaging (DTI) ──────────────────────────
#> • Target: pasat
#> • Properties: groups
#> • Features (3):
#>   • tfi (2): cca, rcst
#>   • fct (1): sex
#> • Groups: subject_id
```

To train a model on this task we first need to extract scalar features
from the functions. We illustrate this below by extracting the mean
value.

``` r
po_fmean = po("fda.extract", features = "mean")

task_fmean = po_fmean$train(list(task))[[1L]]
task_fmean$head()
#>    pasat    sex  cca_mean rcst_mean
#> 1:    31 female 0.4493332 0.4968519
#> 2:    31 female 0.4441292 0.4810724
#> 3:    29 female 0.4257795 0.5102722
#> 4:    34 female 0.4418538 0.5453188
#> 5:    37 female 0.4700994 0.5471177
#> 6:    40 female 0.4873356 0.4969408
```

This can be combined with a `Learner` into a `GraphLearner` that first
extracts features and then trains a model.

``` r
# split data into train and test set
ids = partition(task)

# define a Graph and convert it to a GraphLearner
graph = po("fda.extract", features = "mean", drop = TRUE) %>>%
  po("learner", learner = lrn("regr.rpart"))

glrn = as_learner(graph)

# train the graph learner on the train set
glrn$train(task, row_ids = ids$train)

# make predictions on the test set
glrn$predict(task, row_ids = ids$test)
#> 
#> ── <PredictionRegr> for 111 observations: ──────────────────────────────────────
#>  row_ids truth response
#>       11    48 49.99174
#>       12    40 49.99174
#>       13    43 52.42105
#>      ---   ---      ---
#>      324    57 52.42105
#>      325    57 41.30769
#>      326    60 49.99174
```

## Implemented PipeOps

| Key | Label | Packages | Tags |
|:---|:---|:---|:---|
| [fda.bsignal](https://mlr3fda.mlr-org.com/dev/reference/mlr_pipeops_fda.bsignal) | B-spline Feature Extraction | [tf](https://cran.r-project.org/package=tf), [mboost](https://cran.r-project.org/package=mboost), [FDboost](https://cran.r-project.org/package=FDboost) | fda, data transform |
| [fda.cor](https://mlr3fda.mlr-org.com/dev/reference/mlr_pipeops_fda.cor) | Cross-Correlation of Functional Data | [tf](https://cran.r-project.org/package=tf) | fda, data transform |
| [fda.extract](https://mlr3fda.mlr-org.com/dev/reference/mlr_pipeops_fda.extract) | Extract Simple Features from Functional Columns | [tf](https://cran.r-project.org/package=tf) | fda, data transform |
| [fda.flatten](https://mlr3fda.mlr-org.com/dev/reference/mlr_pipeops_fda.flatten) | Flatten Functional Columns | [tf](https://cran.r-project.org/package=tf) | fda, data transform |
| [fda.fourier](https://mlr3fda.mlr-org.com/dev/reference/mlr_pipeops_fda.fourier) | Fast Fourier Transform Features | [tf](https://cran.r-project.org/package=tf) | fda, data transform |
| [fda.fpca](https://mlr3fda.mlr-org.com/dev/reference/mlr_pipeops_fda.fpca) | Functional Principal Component Analysis | [tf](https://cran.r-project.org/package=tf) | fda, data transform |
| [fda.interpol](https://mlr3fda.mlr-org.com/dev/reference/mlr_pipeops_fda.interpol) | Interpolate Functional Columns | [tf](https://cran.r-project.org/package=tf) | fda, data transform |
| [fda.random_effect](https://mlr3fda.mlr-org.com/dev/reference/mlr_pipeops_fda.random_effect) | Extract Random Effects from Functional Columns | [tf](https://cran.r-project.org/package=tf), [lme4](https://cran.r-project.org/package=lme4) | fda, data transform |
| [fda.register](https://mlr3fda.mlr-org.com/dev/reference/mlr_pipeops_fda.register) | Register (Align) Functional Columns | [tf](https://cran.r-project.org/package=tf) | fda, data transform |
| [fda.scalerange](https://mlr3fda.mlr-org.com/dev/reference/mlr_pipeops_fda.scalerange) | Linearly Transform the Domain of Functional Data | [tf](https://cran.r-project.org/package=tf) | fda, data transform |
| [fda.smooth](https://mlr3fda.mlr-org.com/dev/reference/mlr_pipeops_fda.smooth) | Smooth Functional Columns | [tf](https://cran.r-project.org/package=tf), stats | fda, data transform |
| [fda.tsfeats](https://mlr3fda.mlr-org.com/dev/reference/mlr_pipeops_fda.tsfeats) | Time Series Feature Extraction | [tf](https://cran.r-project.org/package=tf), [tsfeatures](https://cran.r-project.org/package=tsfeatures) | fda, data transform |
| [fda.wavelets](https://mlr3fda.mlr-org.com/dev/reference/mlr_pipeops_fda.wavelets) | Discrete Wavelet Transform Features | [tf](https://cran.r-project.org/package=tf), [wavelets](https://cran.r-project.org/package=wavelets) | fda, data transform |
| [fda.zoom](https://mlr3fda.mlr-org.com/dev/reference/mlr_pipeops_fda.zoom) | Zoom In/Out on Functional Columns | [tf](https://cran.r-project.org/package=tf) | fda, data transform |

## Bugs, Questions, Feedback

*mlr3fda* is a free and open source software project that encourages
participation and feedback. If you have any issues, questions,
suggestions or feedback, please do not hesitate to open an “issue” about
it on the GitHub page!

In case of problems / bugs, it is often helpful if you provide a
“minimum working example” that showcases the behaviour (but don’t worry
about this if the bug is obvious).

Please understand that the resources of the project are limited:
response may sometimes be delayed by a few days, and some feature
suggestions may be rejected if they are deemed too tangential to the
vision behind the project.

## Acknowledgements

The development of this R-package was supported by Roche Diagnostics
R&D.

#' @title Extract Random Effects from Functional Columns
#'
#' @name mlr_pipeops_fda.random_effect
#'
#' @description
#' This is the class that extracts random effects, specifically random intercepts and
#' random slopes, from functional columns. This PipeOp fits a linear mixed model, specifically
#' a random intercept and random slope model, using the [lme4::lmer()] function.
#' The target variable is the value of the functional feature which is regressed on the functional feature's argument
#' while subject id determines the grouping structure. After model estimation, the random effects are extracted and
#' assigned to the correct id.
#'
#' @section Parameters:
#' The parameters are the parameters inherited from [`PipeOpTaskPreprocSimple`][mlr3pipelines::PipeOpTaskPreprocSimple].
#'
#' @section Naming:
#' The new names append `_random_intercept` and `_random_slope` to the corresponding column name of the
#' functional feature.
#'
#' @export
#' @examples
#' task = tsk("dti")
#' po_fre = po("fda.random_effect")
#' task_fre = po_fre$train(list(task))[[1L]]
PipeOpFDARandomEffect = R6Class(
  "PipeOpFDARandomEffect",
  inherit = PipeOpTaskPreprocSimple,
  public = list(
    #' @description Initializes a new instance of this Class.
    #' @param id (`character(1)`)\cr
    #'   Identifier of resulting object, default `"fda.random_effect"`.
    #' @param param_vals (named `list()`)\cr
    #'   List of hyperparameter settings, overwriting the hyperparameter settings that would
    #'   otherwise be set during construction. Default `list()`.
    initialize = function(id = "fda.random_effect", param_vals = list()) {
      super$initialize(
        id = id,
        param_vals = param_vals,
        packages = c("mlr3fda", "mlr3pipelines", "tf", "lme4"),
        feature_types = c("tfd_reg", "tfd_irreg"),
        tags = "fda"
      )
    }
  ),
  private = list(
    .transform_dt = function(dt, levels) {
      setcbindlist(imap(dt, function(x, nm) {
        tab = as.data.frame(x, unnest = TRUE)
        model = invoke(lme4::lmer, .args = list(formula = value ~ arg + (1 + arg | id), data = tab))
        feats = lme4::ranef(model)$id
        setDT(feats)
        setnames(feats, sprintf("%s_%s", nm, c("random_intercept", "random_slope")))
      }))
    }
  )
)

#' @include zzz.R
register_po("fda.random_effect", PipeOpFDARandomEffect)

#' @export
hash_input.tfd_reg = function(x) {
  # The evaluator is a function with environment, srcref, and bytecode, which we want to ignore
  attr(x, "evaluator") = hash_input(attr(x, "evaluator"))
  x
}

#' @export
hash_input.tfd_irreg = function(x) {
  # The evaluator is a function with environment, srcref, and bytecode, which we want to ignore
  attr(x, "evaluator") = hash_input(attr(x, "evaluator"))
  x
}

#' @import R6
#' @import checkmate
#' @import data.table
#' @import mlr3
#' @import mlr3misc
#' @import mlr3pipelines
#' @import paradox
#' @importFrom tf tf_approx_linear tf_approx_spline tf_approx_fill_extend tf_approx_locf tf_approx_nocb
#'
#' @section Data types:
#' To extend mlr3 to functional data, two data types from the tf package are added:
#' * `tfd_irreg` - Irregular functional data, i.e. the functions are observed for
#'   potentially different inputs for each observation.
#' * `tfd_reg` - Regular functional data, i.e. the functions are observed for the same input
#'   for each individual.
#'
#' `r tools::toRd(citation("mlr3"))`
"_PACKAGE"

mlr3fda_feature_types = c(tfr = "tfd_reg", tfi = "tfd_irreg")
mlr3fda_tasks = new.env(parent = emptyenv())
mlr3fda_pipeops = new.env(parent = emptyenv())
mlr3fda_pipeop_tags = "fda"

named_union = function(x, y) set_names(union(x, y), union(names(x), names(y)))

# metainf must be manually added in the register_mlr3pipelines function
# Because the value is substituted, we cannot pass it through this function
register_po = function(name, constructor) {
  if (name %chin% names(mlr3fda_pipeops)) {
    stopf("pipeop %s registered twice", name)
  }
  mlr3fda_pipeops[[name]] = list(constructor = constructor)
}

register_task = function(name, constructor) {
  if (name %chin% names(mlr3fda_tasks)) {
    stopf("task %s registered twice", name)
  }
  mlr3fda_tasks[[name]] = constructor
}

register_mlr3 = function() {
  # add data types
  mlr_reflections = utils::getFromNamespace("mlr_reflections", ns = "mlr3")
  mlr_reflections$task_feature_types = named_union(mlr_reflections$task_feature_types, mlr3fda_feature_types)

  # add tasks
  mlr_tasks = utils::getFromNamespace("mlr_tasks", ns = "mlr3")
  iwalk(as.list(mlr3fda_tasks), \(task, id) mlr_tasks$add(id, task))
}

register_mlr3pipelines = function() {
  mlr_reflections = utils::getFromNamespace("mlr_reflections", ns = "mlr3")
  mlr_pipeops = utils::getFromNamespace("mlr_pipeops", ns = "mlr3pipelines")
  iwalk(as.list(mlr3fda_pipeops), function(value, name) {
    mlr_pipeops$add(name, value$constructor, value$metainf)
  })
  mlr_reflections$pipeops$valid_tags = union(mlr_reflections$pipeops$valid_tags, mlr3fda_pipeop_tags)
}

.onLoad = function(libname, pkgname) {
  assign("lg", lgr::get_logger("mlr3"), envir = parent.env(environment()))
  if (Sys.getenv("IN_PKGDOWN") == "true") {
    lg$set_threshold("warn")
  }
  register_namespace_callback(pkgname, "mlr3", register_mlr3)
  register_namespace_callback(pkgname, "mlr3pipelines", register_mlr3pipelines)
}

.onUnload = function(libPaths) {
  walk(names(mlr3fda_tasks), \(nm) mlr_tasks$remove(nm))
  walk(names(mlr3fda_pipeops), \(nm) mlr_pipeops$remove(nm))
  mlr_reflections$task_feature_types =
    mlr_reflections$task_feature_types[mlr_reflections$task_feature_types %nin% mlr3fda_feature_types]
  mlr_reflections$pipeops$valid_tags = setdiff(mlr_reflections$pipeops$valid_tags, mlr3fda_pipeop_tags)
}

leanify_package()

#' @title Phoneme Classification Task
#'
#' @name mlr_tasks_phoneme
#' @format [R6::R6Class] inheriting from [mlr3::TaskClassif].
#'
#' @description
#' The task contains a single functional covariate and 5 equally big classes (aa, ao, dcl, iy, sh).
#' The aim is to predict the class of the phoneme from the functional feature, which is a
#' log-periodogram.\cr
#' This is a subset of the full dataset, which is contained in the package `fda.usc`.
#'
#' @templateVar id phoneme
#' @template task
#'
#' @references `r format_bib("ferraty2003curves")`
#'
#' @template seealso_task
NULL

load_task_phoneme = function(id = "phoneme") {
  phoneme = load_dataset("phoneme", package = "mlr3fda")
  phoneme = data.table(
    class = phoneme$classlearn,
    X = tf::tfd(as.matrix(phoneme[, -151L]))
  )
  b = as_data_backend(phoneme)

  task = TaskClassif$new(
    id = id,
    backend = b,
    target = "class",
    label = "Phoneme Curves"
  )
  b$hash = task$man = "mlr3fda::mlr_tasks_phoneme"
  task
}

#' @include zzz.R
register_task("phoneme", load_task_phoneme)

#' @title Smooth Functional Columns
#'
#' @name mlr_pipeops_fda.smooth
#'
#' @description
#' Smoothes functional data using [tf::tf_smooth()].
#' This preprocessing operator is similar to [`PipeOpFDAInterpol`], however it does not interpolate to unobserved
#' x-values, but rather smooths the observed values.
#'
#' @section Parameters:
#' The parameters are the parameters inherited from [`PipeOpTaskPreprocSimple`][mlr3pipelines::PipeOpTaskPreprocSimple],
#' as well as the following parameters:
#' * `method` :: `character(1)`\cr
#'   One of:
#'   * `"lowess"`: locally weighted scatterplot smoothing (default)
#'   * `"rollmean"`: rolling mean
#'   * `"rollmedian"`: rolling median
#'   * `"savgol"`: Savitzky-Golay filtering
#'
#'   All methods but `"lowess"` ignore non-equidistant arg values.
#' * `args` :: named `list()`\cr
#'   List of named arguments that is passed to `tf_smooth()`. See the help page of `tf_smooth()` for
#'   default values.
#' * `verbose` :: `logical(1)`\cr
#'   Whether to print messages during the transformation.
#'   Is initialized to `FALSE`.
#'
#' @export
#' @examples
#' task = tsk("fuel")
#' po_smooth = po("fda.smooth", method = "rollmean", args = list(k = 5))
#' task_smooth = po_smooth$train(list(task))[[1L]]
#' task_smooth
#' task_smooth$data(cols = c("NIR", "UVVIS"))
PipeOpFDASmooth = R6Class(
  "PipeOpFDASmooth",
  inherit = PipeOpTaskPreprocSimple,
  public = list(
    #' @description Initializes a new instance of this Class.
    #' @param id (`character(1)`)\cr
    #'   Identifier of resulting object, default `"fda.smooth"`.
    #' @param param_vals (named `list()`)\cr
    #'   List of hyperparameter settings, overwriting the hyperparameter settings that would
    #'   otherwise be set during construction. Default `list()`.
    initialize = function(id = "fda.smooth", param_vals = list()) {
      param_set = ps(
        method = p_fct(
          default = "lowess",
          c("lowess", "rollmean", "rollmedian", "savgol"),
          tags = c("train", "predict")
        ),
        args = p_uty(
          tags = c("train", "predict", "required"),
          custom_check = crate(\(x) check_list(x, names = "unique"))
        ),
        verbose = p_lgl(tags = c("train", "predict", "required"))
      )

      param_set$set_values(args = list(), verbose = FALSE)

      super$initialize(
        id = id,
        param_set = param_set,
        param_vals = param_vals,
        packages = c("mlr3fda", "mlr3pipelines", "tf", "stats"),
        feature_types = c("tfd_reg", "tfd_irreg"),
        tags = "fda"
      )
    }
  ),
  private = list(
    .transform_dt = function(dt, levels) {
      pars = self$param_set$get_values()

      if (pars$verbose) {
        for (j in seq_along(dt)) {
          set(dt, j = j, value = invoke(tf::tf_smooth, x = dt[[j]], method = pars$method, .args = pars$args))
        }
        return(dt)
      }
      for (j in seq_along(dt)) {
        set(
          dt,
          j = j,
          value = suppressMessages(invoke(tf::tf_smooth, x = dt[[j]], method = pars$method, .args = pars$args))
        )
      }
      dt
    }
  )
)

#' @include zzz.R
register_po("fda.smooth", PipeOpFDASmooth)

#' @importFrom utils bibentry
# nolint start
bibentries = c(
  ferraty2003curves = bibentry(
    "article",
    title = "Curves discrimination: a nonparametric functional approach",
    author = "Ferraty, Fr\xc3\xa9d\xc3\xa9ric and Vieu, Philippe",
    journal = "Computational Statistics & Data Analysis",
    volume = "44",
    number = "1-2",
    pages = "161--173",
    year = "2003",
    publisher = "Elsevier"
  ),
  brockhaus2015functional = bibentry(
    "article",
    title = "The functional linear array model",
    author = "Brockhaus, Sarah and Scheipl, Fabian and Hothorn, Torsten and Greven, Sonja",
    journal = "Statistical Modelling",
    volume = "15",
    number = "3",
    pages = "279--300",
    year = "2015",
    publisher = "SAGE Publications Sage India: New Delhi, India"
  ),
  goldsmith2011penalized = bibentry(
    "article",
    title = "Penalized functional regression",
    author = "Goldsmith, Jeff and Bobb, Jennifer and Crainiceanu, Ciprian M and Caffo, Brian and Reich, Daniel",
    journal = "Journal of Computational and Graphical Statistics",
    volume = "20",
    number = "4",
    pages = "830--851",
    year = "2011",
    publisher = "Taylor & Francis"
  ),
  oppenheim2010discrete = bibentry(
    "book",
    title = "Discrete-Time Signal Processing",
    author = "Oppenheim, Alan V and Schafer, Ronald W",
    year = "2010",
    edition = "3rd",
    publisher = "Pearson",
    isbn = "978-0131988422"
  )
)
# nolint end

#' @title Zoom In/Out on Functional Columns
#'
#' @name mlr_pipeops_fda.zoom
#'
#' @description
#' Zoom in or out on functional features by restricting their domain to a specified window.
#' This operation extracts a subset of each function by defining new lower and upper boundaries,
#' effectively cropping the functional data to focus on a specific region of interest.
#' Calls [tf::tf_zoom()] from package \CRANpkg{tf}.
#'
#' @section Parameters:
#' The parameters are the parameters inherited from [`PipeOpTaskPreprocSimple`][mlr3pipelines::PipeOpTaskPreprocSimple],
#' as well as the following parameters:
#' * `begin` :: `numeric()`\cr
#'   The lower limit of the domain. Can be a single value applied to all
#'   functional columns, or a numeric of length equal to the number of observations.
#'   The window includes all values where argument >= `begin`. If not specified,
#'   defaults to the lower limit of each function's domain.
#' * `end` :: `numeric()`\cr
#'   The upper limit of the domain.
#'
#' @export
#' @examples
#' task = tsk("fuel")
#' pop = po("fda.zoom", begin = 50, end = 100)
#' task_zoom = pop$train(list(task))[[1L]]
#' task_zoom$data()
PipeOpFDAZoom = R6Class(
  "PipeOpFDAZoom",
  inherit = PipeOpTaskPreprocSimple,
  public = list(
    #' @description Initializes a new instance of this Class.
    #' @param id (`character(1)`)\cr
    #'   Identifier of resulting object, default `"fda.zoom"`.
    #' @param param_vals (named `list()`)\cr
    #'   List of hyperparameter settings, overwriting the hyperparameter settings that would
    #'   otherwise be set during construction. Default `list()`.
    initialize = function(id = "fda.zoom", param_vals = list()) {
      param_set = ps(
        begin = p_uty(
          tags = c("train", "predict"),
          custom_check = crate(\(x) check_numeric(x, finite = TRUE, any.missing = FALSE, min.len = 1))
        ),
        end = p_uty(
          tags = c("train", "predict"),
          custom_check = crate(\(x) check_numeric(x, finite = TRUE, any.missing = FALSE, min.len = 1))
        )
      )

      super$initialize(
        id = id,
        param_set = param_set,
        param_vals = param_vals,
        packages = c("mlr3fda", "mlr3pipelines", "tf"),
        feature_types = c("tfd_reg", "tfd_irreg"),
        tags = "fda"
      )
    }
  ),
  private = list(
    .transform_dt = function(dt, levels) {
      pars = self$param_set$get_values()
      begin = pars$begin
      end = pars$end
      assert_true(length(begin) == length(end))
      if (!is.null(begin) && !is.null(end)) {
        assert_true(all(begin < end))
      }

      for (j in seq_along(dt)) {
        set(dt, j = j, value = invoke(tf::tf_zoom, f = dt[[j]], .args = pars))
      }
      dt
    }
  )
)

#' @include zzz.R
register_po("fda.zoom", PipeOpFDAZoom)

#' @title Cross-Correlation of Functional Data
#'
#' @name mlr_pipeops_fda.cor
#'
#' @description
#' Calculates the cross-correlation between two functional vectors using [tf::tf_crosscor()].
#' Note that it only operates on regular data and that the cross-correlation assumes that each column
#' has the same domain.
#'
#' To apply this `PipeOp` to irregular data, convert it to a regular grid first using [`PipeOpFDAInterpol`].
#' If you need to change the domain of the columns, use [`PipeOpFDAScaleRange`].
#'
#' @section Parameters:
#' The parameters are the parameters inherited from [`PipeOpTaskPreprocSimple`][mlr3pipelines::PipeOpTaskPreprocSimple],
#' as well as the following parameters:
#' * `arg` :: `numeric()`\cr
#'   Grid to use for the cross-correlation.
#'
#' @export
#' @examples
#' set.seed(1234L)
#' dt = data.table(y = 1:100, x1 = tf::tf_rgp(100L), x2 = tf::tf_rgp(100L))
#' task = as_task_regr(dt, target = "y")
#' po_cor = po("fda.cor")
#' task_cor = po_cor$train(list(task))[[1L]]
#' task_cor
PipeOpFDACor = R6Class(
  "PipeOpFDACor",
  inherit = PipeOpTaskPreprocSimple,
  public = list(
    #' @description Initializes a new instance of this Class.
    #' @param id (`character(1)`)\cr
    #'   Identifier of resulting object, default `"fda.cor"`.
    #' @param param_vals (named `list()`)\cr
    #'   List of hyperparameter settings, overwriting the hyperparameter settings that would
    #'   otherwise be set during construction. Default `list()`.
    initialize = function(id = "fda.cor", param_vals = list()) {
      param_set = ps(
        arg = p_uty(tags = c("train", "predict"), custom_check = check_numeric)
      )

      super$initialize(
        id = id,
        param_set = param_set,
        param_vals = param_vals,
        packages = c("mlr3fda", "mlr3pipelines", "tf"),
        feature_types = "tfd_reg",
        tags = "fda"
      )
    }
  ),
  private = list(
    .transform_dt = function(dt, levels) {
      pars = self$param_set$get_values()

      k = ncol(dt)
      if (k < 2L) {
        warning_input("task has fewer than 2 columns")
        return(dt)
      }

      nms = names(dt)
      res = list()
      for (i in 2:k) {
        for (j in 1:(i - 1L)) {
          x = dt[[i]]
          y = dt[[j]]
          if (!all(tf::tf_domain(x) == tf::tf_domain(y))) {
            error_input("Domain of %s and %s do not match", nms[[j]], nms[[i]])
          }
          nm = sprintf("%s_%s_cor", nms[[j]], nms[[i]])
          res[[nm]] = invoke(tf::tf_crosscor, x = x, y = y, .args = pars)
        }
      }
      setDT(res)
    }
  )
)

#' @include zzz.R
register_po("fda.cor", PipeOpFDACor)

#' The fuel dataset
#'
#' See [mlr_tasks_fuel] for a description of the dataset.
#'
#' @docType data
#' @keywords datasets
#' @name fuel
#' @usage data(fuel)
#' @format A data frame with 129 rows and 4 variables
#' @keywords internal
NULL

#' The phoneme dataset
#'
#' See [mlr_tasks_phoneme] for a description of the dataset.
#'
#' @docType data
#' @keywords datasets
#' @name phoneme
#' @usage data(phoneme)
#' @format A data frame with 250 rows and 2 variables
#' @keywords internal
NULL

#' The dti dataset
#'
#' See [mlr_tasks_dti] for a description of the dataset.
#'
#' @docType data
#' @keywords datasets
#' @name dti
#' @usage data(dti)
#' @format A data frame with 340 rows and 5 variables
#' @keywords internal
NULL

#' @title Diffusion Tensor Imaging (DTI) Regression Task
#'
#' @name mlr_tasks_dti
#' @format [R6::R6Class] inheriting from [mlr3::TaskRegr].
#'
#' @description
#' This dataset contains two functional covariates and three scalar covariates. The goal is
#' to predict the PASAT score. `pasat` represents the PASAT score at each visit.
#' `subject_id` represents the subject ID. `cca` represents the fractional anisotropy tract profiles from the corpus
#' callosum. `sex` indicates subject's sex. `rcst` represents the fractional anisotropy tract profiles from the right
#' corticospinal tract. Rows containing NAs are removed.
#'
#' This is a subset of the full dataset, which is contained in the package `refund`.
#'
#' @templateVar id dti
#' @template task
#'
#' @references `r format_bib("goldsmith2011penalized")`
#'
#' Brain dataset courtesy of Gordon Kindlmann at the Scientific Computing and Imaging Institute, University of Utah,
#' and Andrew Alexander, W. M. Keck Laboratory for Functional Brain Imaging and Behavior, University of
#' Wisconsin-Madison.
#'
#' @template seealso_task
NULL

load_task_dti = function(id = "dti") {
  dti = load_dataset("dti", package = "mlr3fda")
  dti = data.table(
    subject_id = as.factor(dti$ID),
    pasat = dti$pasat,
    cca = tf::tfd(dti$cca, arg = seq(0L, 1L, length.out = 93L)),
    rcst = tf::tfd(dti$rcst, arg = seq(0L, 1L, length.out = 55L)),
    sex = dti$sex
  )
  dti = stats::na.omit(dti)
  b = as_data_backend(dti)

  task = TaskRegr$new(
    id = id,
    backend = b,
    target = "pasat",
    label = "Diffusion Tensor Imaging (DTI)"
  )
  b$hash = task$man = "mlr3fda::mlr_tasks_dti"
  task$col_roles$group = "subject_id"
  task$col_roles$feature = setdiff(task$col_roles$feature, "subject_id")
  task
}

#' @include zzz.R
register_task("dti", load_task_dti)

#' @title Functional Principal Component Analysis
#'
#' @name mlr_pipeops_fda.fpca
#'
#' @description
#' This `PipeOp` applies a functional principal component analysis (FPCA) to functional columns and then
#' extracts the principal components as features. This is done using a (truncated) weighted SVD.
#'
#' To apply this `PipeOp` to irregular data, convert it to a regular grid first using [`PipeOpFDAInterpol`].
#'
#' For more details, see [tf::tfb_fpc()], which is called internally.
#'
#' @section Parameters:
#' The parameters are the parameters inherited from [`PipeOpTaskPreproc`][mlr3pipelines::PipeOpTaskPreproc],
#' as well as the following parameters:
#' * `pve` :: `numeric(1)`\cr
#'   The percentage of variance explained that should be retained. Default is `0.995`.
#' * `n_components` :: `integer(1)`\cr
#'   The number of principal components to extract. This parameter is initialized to `Inf`.
#'
#' @section Naming:
#' The new names generally append a `_pc_{number}` to the corresponding column name.
#' If a column was called `"x"` and there are three principal components, the corresponding
#' new columns will be called `"x_pc_1", "x_pc_2", "x_pc_3"`.
#'
#' @export
#' @examples
#' task = tsk("fuel")
#' po_fpca = po("fda.fpca", n_components = 3L)
#' task_fpca = po_fpca$train(list(task))[[1L]]
#' task_fpca$data()
PipeOpFPCA = R6Class(
  "PipeOpFPCA",
  inherit = PipeOpTaskPreproc,
  public = list(
    #' @description Initializes a new instance of this Class.
    #' @param id (`character(1)`)\cr
    #'   Identifier of resulting object, default `"fda.fpca"`.
    #' @param param_vals (named `list()`)\cr
    #'   List of hyperparameter settings, overwriting the hyperparameter settings that would
    #'   otherwise be set during construction. Default `list()`.
    initialize = function(id = "fda.fpca", param_vals = list()) {
      param_set = ps(
        pve = p_dbl(default = 0.995, lower = 0, upper = 1, tags = "train"),
        n_components = p_int(1L, special_vals = list(Inf), tags = c("train", "required"))
      )
      param_set$set_values(n_components = Inf)

      super$initialize(
        id = id,
        param_set = param_set,
        param_vals = param_vals,
        packages = c("mlr3fda", "mlr3pipelines", "tf"),
        feature_types = "tfd_reg",
        tags = "fda"
      )
    }
  ),
  private = list(
    .train_dt = function(dt, levels, target) {
      pars = self$param_set$get_values(tags = "train")

      dt = map_dtc(dt, function(x, nm) {
        invoke(tf::tfb_fpc, data = x, .args = remove_named(pars, "n_components"))
      })
      self$state = list(fpc = dt)

      dt = imap_dtc(dt, function(col, nm) {
        map(col, function(x) {
          pc = as.list(x[2:min(pars$n_components + 1L, length(x))])
          set_names(pc, sprintf("%s_pc_%i", nm, seq_along(pc)))
        })
      })
      unnest(dt, names(dt))
    },

    .predict_dt = function(dt, levels) {
      pars = self$param_set$get_values()

      dt = imap_dtc(dt, function(col, nm) {
        fpc = invoke(
          tf::tf_rebase,
          object = col,
          basis_from = self$state$fpc[[nm]],
          arg = tf::tf_arg(col)
        )
        map(fpc, function(x) {
          pc = as.list(x[2:min(pars$n_components + 1L, length(x))])
          set_names(pc, sprintf("%s_pc_%i", nm, seq_along(pc)))
        })
      })
      unnest(dt, names(dt))
    }
  )
)

#' @include zzz.R
register_po("fda.fpca", PipeOpFPCA)

#' @title B-spline Feature Extraction
#'
#' @name mlr_pipeops_fda.bsignal
#'
#' @description
#' This `PipeOp` extracts features from functional data using B-spline basis functions.
#' The extracted features are B-spline coefficients that represent the functional data in the B-spline basis space.
#' For more details, see [FDboost::bsignal()], which is called internally.
#'
#' @section Parameters:
#' The parameters are the parameters inherited from [`PipeOpTaskPreprocSimple`][mlr3pipelines::PipeOpTaskPreprocSimple],
#' as well as the following parameters:
#' * `inS` :: `character(1)`\cr
#'   Type of effect in the covariate index: one of `"smooth"`, `"linear"`, `"constant"`. Default is `"smooth"`.
#' * `knots` :: `numeric()`\cr
#'   Either the number of interior knots or a vector of their positions.
#' * `boundary.knots` :: `numeric(2)`\cr
#'   Boundary points at which to anchor the B-spline basis.
#'   Lower and upper boundary points for the spline basis. Defaults to the range of the data.
#' * `degree` :: `integer(1)`\cr
#'   The degree of the regression spline. Default is `3L`.
#' * `differences` :: `integer(1)`\cr
#'   Order of difference penalty. Default is `1L`.
#' * `df` :: `numeric(1)`\cr
#'    Trace of the hat matrix, controlling smoothness. Default is `4`.
#' * `lambda` :: `any`\cr
#'   Smoothing parameter of the penalty term.
#' * `center` :: `logical(1)`\cr
#'   Reparameterize the unpenalized part to zero-mean? Default is `FALSE`.
#' * `cyclic` :: `logical(1)`\cr
#'   If true the fitted coefficient function coincides at the boundaries.
#' * `Z` :: `any`\cr
#'   Custom transformation matrix for the spline design.
#' * `penalty` :: `character(1)`\cr
#'   The penalty type: `"ps"` (P-spline) or `"pss"` (shrinkage). Default is `"ps"`.
#' * `check.ident` :: `logical(1)`\cr
#'   Use checks for identifiability of the effect. Default is `FALSE`.
#'
#' @export
#' @examples
#' task = tsk("fuel")
#' po_bsignal = po("fda.bsignal")
#' task_bsignal = po_bsignal$train(list(task))[[1L]]
#' task_bsignal$data()
PipeOpFDABsignal = R6Class(
  "PipeOpFDABsignal",
  inherit = PipeOpTaskPreprocSimple,
  public = list(
    #' @description Initializes a new instance of this Class.
    #' @param id (`character(1)`)\cr
    #'   Identifier of resulting object, default `"fda.bsignal"`.
    #' @param param_vals (named `list()`)\cr
    #'   List of hyperparameter settings, overwriting the hyperparameter settings that would
    #'   otherwise be set during construction. Default `list()`.
    initialize = function(id = "fda.bsignal", param_vals = list()) {
      param_set = ps(
        inS = p_fct(default = "smooth", c("smooth", "linear", "constant"), tags = c("train", "predict")),
        knots = p_uty(
          default = 10L,
          tags = c("train", "predict"),
          custom_check = crate(\(x) check_numeric(x, min.len = 1))
        ),
        boundary.knots = p_uty(
          default = NULL,
          special_vals = list(NULL),
          tags = c("train", "predict"),
          custom_check = crate(\(x) check_numeric(x, len = 2L, null.ok = TRUE))
        ),
        degree = p_int(default = 3L, tags = c("train", "predict")),
        differences = p_int(1L, default = 1L, tags = c("train", "predict")),
        df = p_dbl(default = 4, tags = c("train", "predict")),
        lambda = p_uty(default = NULL, tags = c("train", "predict")),
        center = p_lgl(default = FALSE, tags = c("train", "predict")),
        cyclic = p_lgl(default = FALSE, tags = c("train", "predict")),
        Z = p_uty(default = NULL, tags = c("train", "predict")),
        penalty = p_fct(default = "ps", c("ps", "pss"), tags = c("train", "predict")),
        check.ident = p_lgl(default = FALSE, tags = c("train", "predict"))
      )

      super$initialize(
        id = id,
        param_set = param_set,
        param_vals = param_vals,
        packages = c("mlr3fda", "mlr3pipelines", "tf", "mboost", "FDboost"),
        feature_types = "tfd_reg",
        tags = "fda"
      )
    }
  ),

  private = list(
    .transform_dt = function(dt, levels) {
      pars = self$param_set$get_values()

      setcbindlist(imap(dt, function(x, nm) {
        x = as.matrix(x)
        blrn = invoke(FDboost::bsignal, x = x, s = seq_col(x), .args = pars)
        bsignal = mboost::extract(object = blrn, what = "design") # get the design matrix of the base learner
        feats = as.data.table(bsignal)
        setnames(feats, sprintf("%s_bsig_%i", nm, seq_col(feats)))
      }))
    }
  )
)

#' @include zzz.R
register_po("fda.bsignal", PipeOpFDABsignal)

#' @title Discrete Wavelet Transform Features
#'
#' @name mlr_pipeops_fda.wavelets
#'
#' @description
#' This `PipeOp` extracts discrete wavelet transform coefficients from functional columns.
#' For more details, see [wavelets::dwt()], which is called internally.
#'
#' @section Parameters:
#' The parameters are the parameters inherited from [`PipeOpTaskPreprocSimple`][mlr3pipelines::PipeOpTaskPreprocSimple],
#' as well as the following parameters:
#' * `filter` :: `character(1)` | `numeric()` | [wavelets::wt.filter()]\cr
#'   Specifies which filter should be used. Must be either [wavelets::wt.filter()] object, an even numeric vector or a
#'   string. In case of a string must be one of `"d"`|`"la"`|`"bl"`|`"c"` followed by an even number for the level of
#'   the filter. The level of the filter needs to be smaller or equal than the time-series length.
#'   For more information and acceptable filters see `help(wt.filter)`. Default is `"la8"`.
#' * `n.levels` :: `integer(1)`\cr
#'   An integer specifying the level of the decomposition.
#' * `boundary` :: `character(1)`\cr
#'   Boundary to be used. `"periodic"` assumes circular time series, for `"reflection"` the series is extended to twice
#'   its length. Default is `"periodic"`.
#' * `fast` :: `logical(1)`\cr
#'   Should the pyramid algorithm be calculated with an internal C function? Default is `TRUE`.
#'
#' @export
#' @examples
#' task = tsk("fuel")
#' po_wavelets = po("fda.wavelets")
#' task_wavelets = po_wavelets$train(list(task))[[1L]]
#' task_wavelets$data()
PipeOpFDAWavelets = R6Class(
  "PipeOpFDAWavelets",
  inherit = PipeOpTaskPreprocSimple,
  public = list(
    #' @description Initializes a new instance of this Class.
    #' @param id (`character(1)`)\cr
    #'   Identifier of resulting object, default `"fda.wavelets"`.
    #' @param param_vals (named `list()`)\cr
    #'   List of hyperparameter settings, overwriting the hyperparameter settings that would
    #'   otherwise be set during construction. Default `list()`.
    initialize = function(id = "fda.wavelets", param_vals = list()) {
      param_set = ps(
        filter = p_uty(
          default = "la8",
          tags = c("train", "predict"),
          custom_check = crate(function(x) {
            if (test_class(x, "wt.filter")) {
              return(TRUE)
            }
            if (test_string(x)) {
              choices = c(
                paste0("d", c(2, 4, 6, 8, 10, 12, 14, 16, 18, 20)),
                paste0("la", c(8, 10, 12, 14, 16, 18, 20)),
                paste0("bl", c(14, 18, 20)),
                paste0("c", c(6, 12, 18, 24, 30)),
                "haar"
              )
              return(check_choice(x, choices))
            }
            if (test_numeric(x) && length(x) %% 2L == 0L) {
              return(TRUE)
            }
            "Must be either a string, an even numeric vector or wavelet filter object"
          })
        ),
        n.levels = p_int(tags = c("train", "predict")),
        boundary = p_fct(default = "periodic", c("periodic", "reflection"), tags = c("train", "predict")),
        fast = p_lgl(default = TRUE, tags = c("train", "predict"))
      )

      super$initialize(
        id = id,
        param_set = param_set,
        param_vals = param_vals,
        packages = c("mlr3fda", "mlr3pipelines", "tf", "wavelets"),
        feature_types = c("tfd_reg", "tfd_irreg"),
        tags = "fda"
      )
    }
  ),

  private = list(
    .transform_dt = function(dt, levels) {
      pars = self$param_set$get_values()
      filter = pars$filter %??% "la8"

      setcbindlist(imap(dt, function(x, nm) {
        feats = map_dtr(
          tf::tf_evaluations(x),
          function(x) {
            wt = invoke(wavelets::dwt, X = x, .args = pars)
            feats = unlist(c(wt@W, wt@V[[wt@level]]), use.names = FALSE)
            as.data.table(t(feats))
          },
          .fill = TRUE
        )
        setnames(feats, sprintf("%s_wav_%s_%i", nm, filter, seq_col(feats)))
      }))
    }
  )
)

#' @include zzz.R
register_po("fda.wavelets", PipeOpFDAWavelets)

#' @title Register (Align) Functional Columns
#'
#' @name mlr_pipeops_fda.register
#'
#' @description
#' Aligns functional features by estimating warping functions against a template using [tf::tf_register()].
#' Registration reduces phase (horizontal) variability while preserving amplitude (vertical) variability,
#' which is useful when curves share a common shape but differ in the timing of their features.
#'
#' During training, a template is learned for each functional column (either estimated iteratively from the
#' data or supplied via `args`). The template is stored as part of the `$state` and reused at predict time
#' so that new observations are aligned to the same reference as the training data.
#'
#' Supported methods are `"srvf"`, `"cc"`, and `"affine"`. The `"landmark"` method from [tf::tf_register()]
#' is not supported because it requires per-observation landmark positions for both training and prediction
#' data, which does not fit a stateful preprocessing step. For landmark registration, use [tf::tf_register()]
#' directly and feed the aligned data into the task.
#'
#' @section Parameters:
#' The parameters are the parameters inherited from [`PipeOpTaskPreproc`][mlr3pipelines::PipeOpTaskPreproc],
#' as well as the following parameters:
#' * `method` :: `character(1)`\cr
#'   Registration method. One of:
#'   * `"srvf"`: elastic registration using the Square Root Velocity Framework via [fdasrvf::time_warping()].
#'     Requires regular grids and the \CRANpkg{fdasrvf} package. Default template is the Karcher mean.
#'   * `"cc"`: continuous-criterion registration with monotone spline warps. Requires regular grids.
#'     Default template is the arithmetic mean.
#'   * `"affine"`: affine (shift and/or scale) registration with warps of the form \eqn{h(t) = a \cdot t + b}.
#'     Supports regular and irregular grids. Default template is the arithmetic mean.
#'
#'   Default is `"srvf"`.
#' * `args` :: named `list()`\cr
#'   Method-specific arguments passed to [tf::tf_register()] via `...`. See the help page of
#'   [tf::tf_estimate_warps()] for valid arguments (e.g. `lambda` / `penalty_method` for `"srvf"`;
#'   `nbasis`, `lambda`, `crit`, `conv`, `iterlim` for `"cc"`; `type`, `shift_range`, `scale_range`
#'   for `"affine"`). An optional `template` entry is honored at training time and stored in the state.
#' * `max_iter` :: `integer(1)`\cr
#'   Maximum number of Procrustes-style template refinement iterations during training. Default is `3`.
#'   Ignored at predict time because the stored template is used directly.
#' * `tol` :: `numeric(1)`\cr
#'   Convergence tolerance for template refinement during training. Default is `0.01`.
#'
#' @section State:
#' `$state$templates` contains the learned template (as a length-1 `tf` vector) for each functional column.
#'
#' @export
#' @examples
#' \donttest{
#' set.seed(1)
#' task = tsk("fuel")
#' po_reg = po("fda.register", method = "affine", args = list(type = "shift_scale"))
#' task_reg = po_reg$train(list(task))[[1L]]
#' task_reg$data(cols = c("NIR", "UVVIS"))
#' }
PipeOpFDARegister = R6Class(
  "PipeOpFDARegister",
  inherit = PipeOpTaskPreproc,
  public = list(
    #' @description Initializes a new instance of this Class.
    #' @param id (`character(1)`)\cr
    #'   Identifier of resulting object, default `"fda.register"`.
    #' @param param_vals (named `list()`)\cr
    #'   List of hyperparameter settings, overwriting the hyperparameter settings that would
    #'   otherwise be set during construction. Default `list()`.
    initialize = function(id = "fda.register", param_vals = list()) {
      param_set = ps(
        method = p_fct(default = "srvf", c("srvf", "cc", "affine"), tags = c("train", "predict")),
        args = p_uty(
          tags = c("train", "predict", "required"),
          custom_check = crate(\(x) check_list(x, names = "unique"))
        ),
        max_iter = p_int(1L, default = 3L, tags = "train"),
        tol = p_dbl(default = 1e-2, lower = 0, tags = "train")
      )
      param_set$set_values(args = list())

      super$initialize(
        id = id,
        param_set = param_set,
        param_vals = param_vals,
        packages = c("mlr3fda", "mlr3pipelines", "tf"),
        feature_types = c("tfd_reg", "tfd_irreg"),
        tags = "fda"
      )
    }
  ),
  private = list(
    .train_dt = function(dt, levels, target) {
      pars = self$param_set$get_values(tags = "train")
      args = c(remove_named(pars, "args"), pars$args)
      templates = vector("list", length(dt))
      names(templates) = names(dt)
      for (j in names(dt)) {
        reg = invoke(tf::tf_register, x = dt[[j]], store_x = FALSE, .args = args)
        templates[[j]] = tf::tf_template(reg)
        set(dt, j = j, value = tf::tf_aligned(reg))
      }
      self$state = list(templates = templates)
      dt
    },

    .predict_dt = function(dt, levels) {
      pars = self$param_set$get_values(tags = "predict")
      args = c(remove_named(pars, "args"), pars$args)
      templates = self$state$templates
      for (j in names(dt)) {
        args$template = templates[[j]]
        reg = invoke(tf::tf_register, x = dt[[j]], store_x = FALSE, .args = args)
        set(dt, j = j, value = tf::tf_aligned(reg))
      }
      dt
    }
  )
)

#' @include zzz.R
register_po("fda.register", PipeOpFDARegister)

#' @title Time Series Feature Extraction
#'
#' @name mlr_pipeops_fda.tsfeats
#'
#' @description
#' This `PipeOp` extracts time series features from functional columns.
#'
#' For more details, see [tsfeatures::tsfeatures()], which is called internally.
#'
#' @section Parameters:
#' The parameters are the parameters inherited from [`PipeOpTaskPreprocSimple`][mlr3pipelines::PipeOpTaskPreprocSimple],
#' as well as the following parameters:
#' * `features` :: `character()`\cr
#'   Function names which return numeric vectors of features.
#'   All features returned by these functions must be named if they return more than one feature.
#'   Default is `c("frequency", "stl_features", "entropy", "acf_features")`.
#' * `scale` :: `logical(1)`\cr
#'   If `TRUE`, data is scaled to mean 0 and sd 1 before features are computed. Default is `TRUE`.
#' * `trim` :: `logical(1)`\cr
#'   If `TRUE`, data is trimmed by `trim_amount` before features are computed.
#'   Values larger than `trim_amount` in absolute value are set to `NA`. Default is `FALSE`.
#' * `trim_amount` :: `numeric(1)`\cr
#'   Default level of trimming. Default is `0.1`.
#' * `parallel` :: `logical(1)`\cr
#'   If `TRUE`, the features are computed in parallel. Default is `FALSE`.
#' * `multiprocess` :: `any`\cr
#'   The function from the future package to use for parallel processing. Default is [future::multisession()].
#' * `na.action` :: `any`\cr
#'   A function to handle missing values. Default is [stats::na.pass()].
#'
#' @section Naming:
#' The new names generally append a `_{feature}` to the corresponding column name.
#' If a column was called `"x"` and the feature is `"trend"`, the corresponding new column will
#' be called `"x_trend"`.
#'
#' @export
#' @examples
#' task = tsk("fuel")
#' po_tsfeats = po("fda.tsfeats")
#' task_tsfeats = po_tsfeats$train(list(task))[[1L]]
#' task_tsfeats$data()
PipeOpFDATsfeatures = R6Class(
  "PipeOpFDATsfeatures",
  inherit = PipeOpTaskPreprocSimple,
  public = list(
    #' @description Initializes a new instance of this Class.
    #' @param id (`character(1)`)\cr
    #'   Identifier of resulting object, default `"fda.tsfeats"`.
    #' @param param_vals (named `list()`)\cr
    #'   List of hyperparameter settings, overwriting the hyperparameter settings that would
    #'   otherwise be set during construction. Default `list()`.
    initialize = function(id = "fda.tsfeats", param_vals = list()) {
      param_set = ps(
        features = p_uty(
          default = c("frequency", "stl_features", "entropy", "acf_features"),
          tags = "train",
          custom_check = crate(\(x) check_character(x, any.missing = FALSE, min.len = 1L))
        ),
        scale = p_lgl(default = TRUE, tags = c("train", "predict")),
        trim = p_lgl(default = FALSE, tags = c("train", "predict")),
        trim_amount = p_dbl(default = 0.1, tags = c("train", "predict"), depends = quote(trim == TRUE)), # nolint
        parallel = p_lgl(default = FALSE, tags = c("train", "predict")),
        multiprocess = p_uty(
          default = future::multisession,
          tags = c("train", "predict"),
          depends = quote(parallel == TRUE), # nolint
          custom_check = check_function
        ),
        na.action = p_uty(default = stats::na.pass, tags = c("train", "predict"), custom_check = check_function)
      )

      super$initialize(
        id = id,
        param_set = param_set,
        param_vals = param_vals,
        packages = c("mlr3fda", "mlr3pipelines", "tf", "tsfeatures"),
        feature_types = c("tfd_reg", "tfd_irreg"),
        tags = "fda"
      )
    }
  ),

  private = list(
    .transform_dt = function(dt, levels) {
      pars = self$param_set$get_values()

      setcbindlist(imap(dt, function(x, nm) {
        tslist = tf::tf_evaluations(x)
        feats = invoke(tsfeatures::tsfeatures, tslist = tslist, .args = pars)
        setDT(feats)
        setnames(feats, sprintf("%s_%s", nm, names(feats)))
      }))
    }
  )
)

#' @include zzz.R
register_po("fda.tsfeats", PipeOpFDATsfeatures)

#' @title Fuel Regression Task
#'
#' @name mlr_tasks_fuel
#' @format [R6::R6Class] inheriting from [mlr3::TaskRegr].
#'
#' @description
#' This dataset contains two functional covariates and one scalar covariate. The goal is
#' to predict the heat value of some fuel based on the ultraviolet radiation spectrum and
#' infrared ray radiation and one scalar column called h2o.
#'
#' This is a subset of the full dataset, which is contained in the package `FDboost`.
#'
#' @templateVar id fuel
#' @template task
#'
#' @references `r format_bib("brockhaus2015functional")`
#'
#' @template seealso_task
NULL

load_task_fuel = function(id = "fuel") {
  fuel = load_dataset("fuel", package = "mlr3fda")
  fuel = data.table(
    heatan = fuel$heatan,
    h2o = fuel$h2o,
    UVVIS = tf::tfd(fuel$UVVIS),
    NIR = tf::tfd(fuel$NIR)
  )
  b = as_data_backend(fuel)

  task = TaskRegr$new(
    id = id,
    backend = b,
    target = "heatan",
    label = "Spectral Data of Fossil Fuels"
  )

  b$hash = task$man = "mlr3fda::mlr_tasks_fuel"
  task
}

#' @include zzz.R
register_task("fuel", load_task_fuel)

#' @title Interpolate Functional Columns
#'
#' @name mlr_pipeops_fda.interpol
#'
#' @description
#' Interpolate functional features (e.g. all individuals are observed at different time-points) to a common grid.
#' This is useful if you want to compare functional features across observations.
#' The interpolation is done using the `tf` package. See [`tfd()`][tf::tfd] for details.
#'
#' @section Parameters:
#' The parameters are the parameters inherited from [`PipeOpTaskPreprocSimple`][mlr3pipelines::PipeOpTaskPreprocSimple],
#' as well as the following parameters:
#' * `grid` :: `character(1)` | `numeric()`\cr
#'   The grid to use for interpolation.
#'   If `grid` is numeric, it must be a sequence of values to use for the grid or a single value that
#'   specifies the number of points to use for the grid, requires `left` and `right` to be specified in the latter case.
#'   If `grid` is a character, it must be one of:
#'   * `"union"`: This option creates a grid based on the union of all argument points from the provided functional
#'     features. This means that if the argument points across features are \(t_1, t_2, ..., t_n\), then the grid will
#'     be the combined unique set of these points. This option is generally used when the argument points vary across
#'     observations and a common grid is needed for comparison or further analysis.
#'   * `"intersect"`: Creates a grid using the intersection of all argument points of a feature.
#'     This grid includes only those points that are common across all functional features,
#'     facilitating direct comparison on a shared set of points.
#'   * `"minmax"`: Generates a grid within the range of the maximum of the minimum argument points to the minimum of the
#'     maximum argument points across features.
#'     This bounded grid encapsulates the argument point range common to all features.
#'   Note: For regular functional data this has no effect as all argument points are the same.
#'   Initial value is `"union"`.
#' * `method` :: `character(1)`\cr
#'   One of:
#'   * `"linear"`: applies linear interpolation without extrapolation (see [tf::tf_approx_linear()]).
#'   * `"spline"`: applies cubic spline interpolation (see [tf::tf_approx_spline()]).
#'   * `"fill_extend"`: applies linear interpolation with constant extrapolation (see [tf::tf_approx_fill_extend()]).
#'   * `"locf"`: applies "last observation carried forward" interpolation (see [tf::tf_approx_locf()]).
#'   * `"nocb"`: applies "next observation carried backward" interpolation (see [tf::tf_approx_nocb()]).
#'
#'   Default is `"linear"`.
#' * `left` :: `numeric()`\cr
#'   The left boundary of the window.
#'   The window is specified such that all values >=left and <=right are kept for the computations.
#' * `right` :: `numeric()`\cr
#'   The right boundary of the window.
#'
#' @export
#' @examples
#' task = tsk("fuel")
#' pop = po("fda.interpol")
#' task_interpol = pop$train(list(task))[[1L]]
#' task_interpol$data()
PipeOpFDAInterpol = R6Class(
  "PipeOpFDAInterpol",
  inherit = PipeOpTaskPreprocSimple,
  public = list(
    #' @description Initializes a new instance of this Class.
    #' @param id (`character(1)`)\cr
    #'   Identifier of resulting object, default `"fda.interpol"`.
    #' @param param_vals (named `list()`)\cr
    #'   List of hyperparameter settings, overwriting the hyperparameter settings that would
    #'   otherwise be set during construction. Default `list()`.
    initialize = function(id = "fda.interpol", param_vals = list()) {
      param_set = ps(
        grid = p_uty(
          tags = c("train", "predict", "required"),
          custom_check = crate(function(x) {
            if (test_string(x)) {
              return(check_choice(x, choices = c("union", "intersect", "minmax")))
            }
            if (test_numeric(x, any.missing = FALSE, min.len = 1L)) {
              return(TRUE)
            }
            "Must be either a string or numeric vector"
          })
        ),
        method = p_fct(
          c("linear", "spline", "fill_extend", "locf", "nocb"),
          default = "linear",
          tags = c("train", "predict")
        ),
        left = p_dbl(tags = c("train", "predict")),
        right = p_dbl(tags = c("train", "predict"))
      )
      param_set$set_values(grid = "union")

      super$initialize(
        id = id,
        param_set = param_set,
        param_vals = param_vals,
        packages = c("mlr3fda", "mlr3pipelines", "tf"),
        feature_types = c("tfd_reg", "tfd_irreg"),
        tags = "fda"
      )
    }
  ),
  private = list(
    .transform_dt = function(dt, levels) {
      pars = self$param_set$get_values()
      grid = pars$grid
      method = pars$method
      left = pars$left
      right = pars$right
      has_left = !is.null(left)
      has_right = !is.null(right)
      if (xor(has_left, has_right)) {
        error_config("Either both or none of 'left' and 'right' must be specified.")
      }
      if (has_left && has_right) {
        assert_count(grid)
        assert_true(left <= right)
      }
      method = method %??% "linear"
      evaluator = sprintf("tf_approx_%s", method)

      if (is.numeric(grid)) {
        if (length(grid) > 1L && !has_left && !has_right) {
          max_grid = max(grid)
          min_grid = min(grid)
          for (j in seq_along(dt)) {
            x = dt[[j]]
            domain = tf::tf_domain(x)
            if (min_grid < domain[[1L]] || max_grid > domain[[2L]]) {
              error_config("The grid must be within the range of the domain.")
            }
            set(dt, j = j, value = invoke(tf::tfd, data = x, arg = grid, .args = list(evaluator = evaluator)))
          }
          return(dt)
        }
        arg = seq(left, right, length.out = grid)
        for (j in seq_along(dt)) {
          set(dt, j = j, value = invoke(tf::tfd, data = dt[[j]], arg = arg, .args = list(evaluator = evaluator)))
        }
        return(dt)
      }

      for (j in seq_along(dt)) {
        x = dt[[j]]
        if (tf::is_irreg(x)) {
          arg = tf::tf_arg(x)
          arg = switch(
            grid,
            union = sort(unique(unlist(arg, recursive = FALSE, use.names = FALSE))),
            intersect = Reduce(intersect, arg),
            minmax = {
              lower = max(map_dbl(arg, 1L))
              upper = min(map_dbl(arg, \(arg) arg[[length(arg)]]))
              arg = sort(unique(unlist(arg, recursive = FALSE, use.names = FALSE)))
              arg[seq(which(lower == arg), which(upper == arg))]
            }
          )
          set(dt, j = j, value = invoke(tf::tfd, data = x, arg = arg, .args = list(evaluator = evaluator)))
        }
      }
      dt
    }
  )
)

#' @include zzz.R
register_po("fda.interpol", PipeOpFDAInterpol)

#' @title Flatten Functional Columns
#'
#' @name mlr_pipeops_fda.flatten
#'
#' @description
#' Convert regular functional features (e.g. all individuals are observed at the same time-points)
#' to new columns, one for each input value to the function.
#'
#' @section Parameters:
#' The parameters are the parameters inherited from [`PipeOpTaskPreprocSimple`][mlr3pipelines::PipeOpTaskPreprocSimple].
#'
#' @section Naming:
#' The new names generally append `_1`, `_2`, ... to the corresponding column name.
#' However this can lead to name clashes with existing columns.
#' This is solved as follows:
#' If a column was called `"x"`, the corresponding new columns will
#' be called `"x_1"`, `"x_2"`, etc. In case of duplicates, unique names are obtained using `make.unique()` and
#' a warning is given.
#'
#' @export
#' @examples
#' task = tsk("fuel")
#' pop = po("fda.flatten")
#' task_flat = pop$train(list(task))[[1L]]
PipeOpFDAFlatten = R6Class(
  "PipeOpFDAFlatten",
  inherit = PipeOpTaskPreprocSimple,
  public = list(
    #' @description Initializes a new instance of this Class.
    #' @param id (`character(1)`)\cr
    #'   Identifier of resulting object, default `"fda.flatten"`.
    #' @param param_vals (named `list()`)\cr
    #'   List of hyperparameter settings, overwriting the hyperparameter settings that would
    #'   otherwise be set during construction. Default `list()`.
    initialize = function(id = "fda.flatten", param_vals = list()) {
      super$initialize(
        id = id,
        param_vals = param_vals,
        packages = c("mlr3fda", "mlr3pipelines", "tf"),
        feature_types = c("tfd_reg", "tfd_irreg"),
        tags = "fda"
      )
    }
  ),
  private = list(
    .transform = function(task) {
      cols = self$state$dt_columns
      if (length(cols) == 0L) {
        return(task)
      }
      dt = task$data(cols = cols)

      dt_flat = setcbindlist(imap(dt, function(x, nm) {
        if (tf::is_irreg(x)) {
          flat = suppressWarnings(as.matrix(x))
        } else {
          flat = as.matrix(x)
        }
        d = as.data.table(flat)
        setnames(d, sprintf("%s_%i", nm, seq_col(flat)))
      }))
      feature_names = names(dt_flat)

      if (anyDuplicated(c(task$col_info$id, feature_names))) {
        unique_names = make.unique(c(task$col_info$id, feature_names), sep = "_")
        feature_names = tail(unique_names, length(feature_names))
        setnames(dt_flat, feature_names)
        lg$debug(sprintf("Duplicate names found in pipeop %s", self$id), feature_names = feature_names)
      }

      task$select(setdiff(task$feature_names, cols))$cbind(dt_flat)
    }
  )
)

#' @include zzz.R
register_po("fda.flatten", PipeOpFDAFlatten)

library(mlr3)
library(checkmate)
library(mlr3misc)
library(mlr3pipelines)
library(R6)
library(paradox)

walk(list.files(system.file("testthat", package = "mlr3"), pattern = "^helper.*\\.[rR]", full.names = TRUE), source)
# filter out helper_compat.R for weird reason to overload testthat functions to v2 behavior
map_if(
  list.files(system.file("testthat", package = "mlr3pipelines"), pattern = "^helper.*\\.[rR]", full.names = TRUE),
  function(x) !endsWith(x, "helper_compat.R"),
  source
)

#' @title Extract Simple Features from Functional Columns
#'
#' @name mlr_pipeops_fda.extract
#'
#' @description
#' This is the class that extracts simple features from functional columns.
#' Note that it only operates on values that were actually observed and does not interpolate.
#'
#' @section Parameters:
#' The parameters are the parameters inherited from [`PipeOpTaskPreprocSimple`][mlr3pipelines::PipeOpTaskPreprocSimple],
#' as well as the following parameters:
#' * `drop` :: `logical(1)`\cr
#'   Whether to drop the original `functional` features and only keep the extracted features.
#'   Note that this does not remove the features from the backend, but only from the active
#'   column role `feature`. Initial value is `TRUE`.
#' * `features` :: `list()` | `character()`\cr
#'   A list of features to extract. Each element can be either a function or a string.
#'   If the element is a function it requires the following arguments: `arg` and `value` and returns a `numeric`.
#'   For string elements, the following predefined features are available:
#'   `"mean"`, `"max"`, `"min"`, `"slope"`, `"median"`, `"var"`.
#'   Initial value is `c("mean", "max", "min", "slope", "median", "var")`.
#' * `left` :: `numeric()`\cr
#'   The left boundary of the window. Initial value is `-Inf`.
#'   The window is specified such that all values >=left and <=right are kept for the computations.
#' * `right` :: `numeric()`\cr
#'   The right boundary of the window. Initial value is `Inf`.
#'
#' @section Naming:
#' The new names generally append a `_{feature}` to the corresponding column name.
#' However this can lead to name clashes with existing columns.
#' This is solved as follows:
#' If a column was called `"x"` and the feature is `"mean"`, the corresponding new column will
#' be called `"x_mean"`. In case of duplicates, unique names are obtained using `make.unique()` and
#' a warning is given.
#'
#' @export
#' @examples
#' task = tsk("fuel")
#' po_fmean = po("fda.extract", features = "mean")
#' task_fmean = po_fmean$train(list(task))[[1L]]
#'
#' # add more than one feature
#' pop = po("fda.extract", features = c("mean", "median", "var"))
#' task_features = pop$train(list(task))[[1L]]
#'
#' # add a custom feature
#' po_custom = po("fda.extract",
#'   features = list(mean = function(arg, value) mean(value, na.rm = TRUE))
#' )
#' task_custom = po_custom$train(list(task))[[1L]]
#' task_custom
PipeOpFDAExtract = R6Class(
  "PipeOpFDAExtract",
  inherit = PipeOpTaskPreprocSimple,
  public = list(
    #' @description Initializes a new instance of this Class.
    #' @param id (`character(1)`)\cr
    #'   Identifier of resulting object, default `"fda.extract"`.
    #' @param param_vals (named `list()`)\cr
    #'   List of hyperparameter settings, overwriting the hyperparameter settings that would
    #'   otherwise be set during construction. Default `list()`.
    initialize = function(id = "fda.extract", param_vals = list()) {
      param_set = ps(
        drop = p_lgl(tags = c("train", "predict", "required")),
        left = p_dbl(tags = c("train", "predict", "required")),
        right = p_dbl(tags = c("train", "predict", "required")),
        features = p_uty(
          tags = c("train", "predict", "required"),
          custom_check = crate(function(x) {
            if (test_character(x)) {
              return(check_subset(x, choices = c("mean", "median", "min", "max", "slope", "var")))
            }
            if (test_list(x)) {
              res = check_list(x, types = c("character", "function"), any.missing = FALSE, unique = TRUE)
              if (!isTRUE(res)) {
                return(res)
              }
              nms = names2(x)
              res = check_names(nms[!is.na(nms)], "unique")
              if (!isTRUE(res)) {
                return(res)
              }
              for (i in seq_along(x)) {
                if (is.function(x[[i]])) {
                  res = check_function(x[[i]], args = c("arg", "value"))
                  if (!isTRUE(res)) {
                    return(res)
                  }
                  res = check_names(nms[i])
                  if (!isTRUE(res)) {
                    return(res)
                  }
                } else {
                  res = check_choice(x[[i]], choices = c("mean", "median", "min", "max", "slope", "var"))
                  # nolint next
                  if (!isTRUE(res)) {
                    return(res)
                  }
                }
              }
              return(TRUE)
            }
            "Features must be a character or list"
          })
        )
      )
      param_set$set_values(
        drop = TRUE,
        left = -Inf,
        right = Inf,
        features = c("mean", "max", "min", "slope", "median", "var")
      )

      super$initialize(
        id = id,
        param_set = param_set,
        param_vals = param_vals,
        packages = c("mlr3fda", "mlr3pipelines", "tf"),
        feature_types = c("tfd_reg", "tfd_irreg"),
        tags = "fda"
      )
    }
  ),
  private = list(
    .transform = function(task) {
      cols = self$state$dt_columns
      if (length(cols) == 0L) {
        return(task)
      }
      dt = task$data(cols = cols)
      pars = self$param_set$get_values()
      drop = pars$drop
      features = pars$features
      left = pars$left
      right = pars$right
      assert_true(left <= right)

      # handle name clashes of generated features with existing columns
      feature_names = imap_chr(features, \(value, nm) if (is.function(value)) nm else value)
      feature_names = as.vector(t(outer(cols, feature_names, paste, sep = "_")))

      if (anyDuplicated(c(task$col_info$id, feature_names))) {
        unique_names = make.unique(c(task$col_info$id, feature_names), sep = "_")
        feature_names = tail(unique_names, length(feature_names))
        lg$debug(sprintf("Duplicate names found in pipeop %s", self$id), feature_names = feature_names)
      }

      features = map(features, function(feature) {
        if (is.function(feature)) {
          return(feature)
        }
        switch(feature, mean = fmean, median = fmedian, min = fmin, max = fmax, slope = fslope, var = fvar)
      })
      fextractor = make_fextractor(features)

      features = map(cols, \(col) invoke(fextractor, x = dt[[col]], left = left, right = right))
      features = setDT(unlist(features, recursive = FALSE, use.names = FALSE))
      setnames(features, feature_names)

      if (!drop) {
        features = cbind(dt, features)
      }

      task$select(setdiff(task$feature_names, cols))$cbind(features)
      task
    }
  )
)

make_fextractor = function(features) {
  function(x, left = -Inf, right = Inf) {
    args = tf::tf_arg(x)

    if (tf::is_reg(x)) {
      interval = ffind(args, left = left, right = right)
      lower = interval[[1L]]
      upper = interval[[2L]]

      if (is.na(lower) || is.na(upper)) {
        res = map(features, \(f) rep(NA_real_, length(x))) # no observation in the given interval [left, right]
        return(res)
      }

      values = tf::tf_evaluations(x)
      arg = args[lower:upper]
      res = map(seq_along(x), function(i) {
        value = values[[i]]
        map(features, \(f) f(arg = arg, value = value[lower:upper]))
      })
      return(transform_list(res))
    }

    values = tf::tf_evaluations(x)
    res = map(seq_along(x), function(i) {
      arg = args[[i]]
      value = values[[i]]

      interval = ffind(arg, left = left, right = right)
      lower = interval[[1L]]
      upper = interval[[2L]]

      if (is.na(lower) || is.na(upper)) {
        rep(NA_real_, length(features)) # no observation in the given interval [left, right]
      } else {
        map(features, \(f) f(arg = arg[lower:upper], value = value[lower:upper]))
      }
    })
    transform_list(res)
  }
}

transform_list = function(x) {
  res = transpose(x)
  map(res, \(x) unlist(x, recursive = FALSE, use.names = FALSE))
}

ffind = function(x, left = -Inf, right = Inf) {
  len = length(x)
  if (left <= x[[1L]] && right >= x[[len]]) {
    return(c(1L, len))
  }
  if (left > x[[len]] || right < x[[1L]]) {
    return(rep(NA_integer_, 2L))
  }
  it = findInterval(c(left, right), x)
  # in case there are no values in the interval, it contains the index of the smallest value below
  # and both values in it are identical,
  # e.g. searching the interval (1.1, 1.2) in c(1, 2) returns c(1, 1) which we here convert to an NA interval
  if (it[[1L]] == it[[2L]] && left > x[[it[[1L]]]]) {
    return(rep(NA_integer_, 2L))
  }
  if (it[[1L]] == 0L) {
    it[[1L]] = 1L
  } else if (x[[it[[1L]]]] < left) {
    it[[1L]] = it[[1L]] + 1L
  }
  it
}

fmean = function(arg, value) mean(value, na.rm = TRUE)
fmin = function(arg, value) min(value, na.rm = TRUE)
fmax = function(arg, value) max(value, na.rm = TRUE)
fmedian = function(arg, value) stats::median(value, na.rm = TRUE)
fslope = function(arg, value) stats::coefficients(stats::lm(value ~ arg))[[2L]]
fvar = function(arg, value) stats::var(value, na.rm = TRUE)

#' @include zzz.R
register_po("fda.extract", PipeOpFDAExtract)

test_that("PipeOpFDAExtract - basic properties", {
  pop = po("fda.extract")
  expect_pipeop(pop)
  expect_identical(pop$id, "fda.extract")
})

test_that("PipeOpFDAExtract works", {
  # tf_reg works
  dt = data.table(
    id = c("Ann", "Ann", "Ann", "Bob", "Bob", "Bob"),
    arg = rep(1:3, 2L),
    value = 1:6
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  y = 1:2
  dt = data.table(f = f, y = y)
  task = as_task_regr(dt, target = "y")

  po_fmean = po("fda.extract", features = "mean", drop = TRUE)
  task_fmean = train_pipeop(po_fmean, list(task))[[1L]]
  fmean = task_fmean$data()$f_mean
  expect_identical(fmean, c(2, 5))

  # multiple functions work
  pop = po("fda.extract", features = c("mean", "median", "var"), drop = TRUE)
  task_pop = train_pipeop(pop, list(task))[[1L]]
  expect = data.table(y = 1:2, f_mean = c(2, 5), f_median = c(2, 5), f_var = c(1, 1))
  expect_equal(task_pop$data(), expect)

  pop = po("fda.extract", features = list("mean", "median", "var"), drop = TRUE)
  task_pop = train_pipeop(pop, list(task))[[1L]]
  expect = data.table(y = 1:2, f_mean = c(2, 5), f_median = c(2, 5), f_var = c(1, 1))
  expect_equal(task_pop$data(), expect)

  # custom function works
  custom = function(arg, value) {
    mean(value, na.rm = TRUE)
  }
  pop = po("fda.extract", features = list("mean", custom = custom), drop = TRUE)
  task_pop = train_pipeop(pop, list(task))[[1L]]
  expect = data.table(y = 1:2, f_mean = c(2, 5), f_custom = c(2, 5))
  expect_equal(task_pop$data(), expect)

  # return NA if not in interval
  po_fmean = po("fda.extract", features = "mean", drop = TRUE, left = 100, right = 200)
  task_fmean = train_pipeop(po_fmean, list(task))[[1L]]
  fmean = task_fmean$data()$f_mean
  expect_equal(fmean, rep(NA_real_, 2L))

  pop = po("fda.extract", features = c("mean", "median", "min"), drop = TRUE, left = 100, right = 200)
  task_pop = train_pipeop(pop, list(task))[[1L]]
  expected = data.table(
    y = 1:2,
    f_mean = rep(NA_real_, 2L),
    f_median = rep(NA_real_, 2L),
    f_min = rep(NA_real_, 2L)
  )
  expect_identical(task_pop$data(), expected)

  # tf_irreg works
  dt = data.table(
    id = c("Ann", "Ann", "Ann", "Bob", "Bob"),
    arg = c(1, 7, 2, 3, 5),
    value = c(1, 2, 3, 4, 5)
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  y = c(1, 2)
  dt = data.table(f = f, y = y)
  task = as_task_regr(dt, target = "y")

  po_fmean = po("fda.extract", features = list("mean", "median", custom = custom), drop = TRUE)
  task_fmean = train_pipeop(po_fmean, list(task))[[1L]]
  fmean = task_fmean$data()$f_mean
  expect_equal(fmean, c(2, 4.5))

  po_fmean = po("fda.extract", features = "mean", drop = TRUE, left = 1, right = 3)
  task_fmean = train_pipeop(po_fmean, list(task))[[1L]]
  fmean = task_fmean$data()$f_mean
  expect_equal(fmean, c(2, 4))

  # return NA if not in interval
  po_fmean = po("fda.extract", features = list("mean"), drop = TRUE, left = 100, right = 200)
  task_fmean = train_pipeop(po_fmean, list(task))[[1L]]
  fmean = task_fmean$data()$f_mean
  expect_identical(fmean, rep(NA_real_, 2L))

  # drop works
  po_fmean = po("fda.extract", features = list("mean"), drop = FALSE)
  task_fmean = train_pipeop(po_fmean, list(task))[[1L]]
  expect_set_equal(task_fmean$feature_names, c("f", "f_mean"))

  po_fmean = po("fda.extract", features = list("mean"), drop = TRUE)
  task_fmean = train_pipeop(po_fmean, list(task))[[1L]]
  expect_set_equal(task_fmean$feature_names, "f_mean")

  # affect_columns works
  po_fmean = po("fda.extract", features = list("mean"), drop = TRUE, affect_columns = selector_name("abc"))
  task_fmean = train_pipeop(po_fmean, list(task))[[1L]]
  expect_set_equal(task_fmean$feature_names, "f")
})

test_that("PipeOpFDAExtract works (simple test) for all features", {
  task = tsk("fuel")
  pop = po("fda.extract")
  features = list("mean", "min", "max", "slope", "median")
  pop$param_set$values$features = features
  expect_list(train_pipeop(pop, list(task)), len = 1L)
})

test_that("PipeOpFDAExtract input validation works", {
  # features not a list or character
  expect_error(po("fda.extract", features = 2L))
  # wrong features
  expect_error(po("fda.extract", features = list("mean", "fmean")))
  expect_error(po("fda.extract", features = c("mean", "fmean")))
  # duplicate features
  expect_error(po("fda.extract", features = list("mean", "mean")))
  # value other than function or string
  expect_error(po("fda.extract", features = list("mean", 2L)))
  # wrong params for feature function
  expect_error(po("fda.extract", features = list(custom = function(arg) mean(arg, na.rm = TRUE))))
  expect_error(po("fda.extract", features = list(custom = function(value) mean(value, na.rm = TRUE))))
  expect_error(po("fda.extract", features = list(custom = function(x, y) sum(x, y))))
  # missing name for custom function
  expect_error(po("fda.extract", features = list(function(arg, value) mean(value, na.rm = TRUE))))
})

test_that("PipeOpFDAExtract works with name clashes", {
  dt = data.table(
    id = c("Ann", "Ann", "Ann", "Bob", "Bob"),
    arg = c(1, 7, 2, 3, 5),
    value = c(1, 2, 3, 4, 5)
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  y = c(1, 2)
  dt = data.table(f = f, y = y)
  dt$f_mean = c(-1, -1)
  task = as_task_regr(dt, target = "y")
  pop = po("fda.extract", features = list("mean"), drop = FALSE)
  taskout = train_pipeop(pop, list(task))[[1L]]
  expect_permutation(taskout$feature_names, c("f", "f_mean", "f_mean_1"))
})

test_that("ffind works", {
  expect_identical(ffind(1:5, 2, 4), c(2L, 4L))
  x = 0:10 * 3
  expect_identical(ffind(x, 9, 25), c(4L, 9L))
  expect_identical(ffind(x, 9.5, 24.5), c(5L, 9L))
  expect_identical(ffind(x, 9.5, 21.5), c(5L, 8L))
  expect_identical(ffind(-5:5, -3, 3), c(3L, 9L))
  expect_identical(ffind(1:10, 1, 10), c(1L, 10L))
  # non-integer
  x = c(1.2, 2.3, 3.4, 4.5)
  expect_identical(ffind(x, 2.5, 4), c(3L, 3L))
  # min/max
  x = c(2, 3, 4, 4)
  expect_identical(ffind(x), c(1L, 4L))
  expect_identical(ffind(1:10, -1, 11), c(1L, 10L))
  # negative numbers
  x = c(-5, -3, -1, 1, 3)
  expect_identical(ffind(x, -3, 1), c(2L, 4L))
  # not in interval
  expect_identical(ffind(2:5, 6, 10), c(NA_integer_, NA_integer_))
  expect_identical(ffind(2:5, 1, 1), c(NA_integer_, NA_integer_))
  expect_identical(ffind(1:10, 20, 30), c(NA_integer_, NA_integer_))
  # single element
  expect_identical(ffind(5, 5, 5), c(1L, 1L))
  expect_identical(ffind(5, 4, 6), c(1L, 1L))
  expect_identical(ffind(5, 6, 7), c(NA_integer_, NA_integer_))
  # NA values
  x = c(2, NA, 4, 5)
  expect_identical(ffind(x, 2, 5), c(1L, 4L))
  # large vector
  x = 1:1e6
  expect_equal(ffind(x, 1e5, 1e6), c(1e5, 1e6))
  # lower and upper same value
  expect_identical(ffind(1:3, 1, 1), c(1L, 1L))
  # not in interval
  expect_identical(ffind(1:3, 1.1, 1.2), c(NA_integer_, NA_integer_))
  expect_identical(ffind(1:3, 1.3, 1.8), c(NA_integer_, NA_integer_))
  # left not in interval
  expect_identical(ffind(1:3, 1.2, 2), c(2L, 2L))
  # right not in interval
  expect_identical(ffind(1:3, 1, 1.2), c(1L, 1L))
  # one boundary outside
  expect_identical(ffind(1:10, 0, 5), c(1L, 5L))
  expect_identical(ffind(1:10, 5, 15), c(5L, 10L))

  expect_identical(
    ffind(c(-3876, -3798, -3453, -3363, -2974, -2953, -2871, -1917, -1335, -1304, -725, 10), left = -200, right = 0),
    rep(NA_integer_, 2L)
  )
})

#' @title Fast Fourier Transform Features
#'
#' @name mlr_pipeops_fda.fourier
#'
#' @description
#' This `PipeOp` extracts Fourier coefficients from functional columns.
#' For more details, see [stats::fft()], which is called internally.
#' Only the one-sided spectrum is returned since the input is real-valued (Oppenheim and Schafer, 2010).
#'
#' @section Parameters:
#' The parameters are the parameters inherited from [`PipeOpTaskPreprocSimple`][mlr3pipelines::PipeOpTaskPreprocSimple],
#' as well as the following parameters:
#' * `type` :: `character(1)`\cr
#'   Which feature to extract from the Fourier coefficients. `"amplitude"` returns the magnitude.
#'   `"phase"` returns the phase shift in degrees (values in \[-180, 180\]). Initial value is `"phase"`.
#'
#' @references `r format_bib("oppenheim2010discrete")`
#'
#' @export
#' @examples
#' task = tsk("fuel")
#' po_fourier = po("fda.fourier")
#' task_fourier = po_fourier$train(list(task))[[1L]]
#' task_fourier$data()
PipeOpFDAFourier = R6Class(
  "PipeOpFDAFourier",
  inherit = PipeOpTaskPreprocSimple,
  public = list(
    #' @description Initializes a new instance of this Class.
    #' @param id (`character(1)`)\cr
    #'   Identifier of resulting object, default `"fda.fourier"`.
    #' @param param_vals (named `list()`)\cr
    #'   List of hyperparameter settings, overwriting the hyperparameter settings that would
    #'   otherwise be set during construction. Default `list()`.
    initialize = function(id = "fda.fourier", param_vals = list()) {
      param_set = ps(
        type = p_fct(levels = c("phase", "amplitude"), tags = c("train", "predict"))
      )
      param_set$set_values(type = "phase")

      super$initialize(
        id = id,
        param_set = param_set,
        param_vals = param_vals,
        packages = c("mlr3fda", "mlr3pipelines", "tf"),
        feature_types = "tfd_reg",
        tags = "fda"
      )
    }
  ),

  private = list(
    .transform_dt = function(dt, levels) {
      type = self$param_set$get_values()$type

      setcbindlist(imap(dt, function(x, nm) {
        values = tf::tf_evaluations(x)
        n = length(values[[1L]])
        n_onesided = n %/% 2L + 1L
        fft_coeff = map_dtr(values, function(v) {
          coeff = stats::fft(v)[seq_len(n_onesided)] / n
          res = switch(
            type,
            amplitude = {
              amp = Mod(coeff)
              # double interior positive frequencies (DC and Nyquist have no conjugate pair)
              double_end = if (n %% 2L == 0L) n_onesided - 1L else n_onesided
              if (double_end >= 2L) {
                amp[2L:double_end] = amp[2L:double_end] * 2
              }
              amp
            },
            phase = {
              mag = Mod(coeff)
              phase = Arg(coeff) * 180 / pi
              phase[mag < max(mag) * 1e-4] = 0
              phase
            }
          )
          as.data.table(t(res))
        })
        setnames(fft_coeff, sprintf("%s_fft_%s_%i", nm, type, seq_col(fft_coeff)))
      }))
    }
  )
)

#' @include zzz.R
register_po("fda.fourier", PipeOpFDAFourier)

#' @title Linearly Transform the Domain of Functional Data
#'
#' @name mlr_pipeops_fda.scalerange
#'
#' @description
#' Linearly transform the domain of functional data so they are between `lower` and `upper`.
#' The formula for this is \eqn{x' = offset + x * scale},
#' where \eqn{scale} is \eqn{(upper - lower) / (max(x) - min(x))} and
#' \eqn{offset} is \eqn{-min(x) * scale + lower}. The same transformation is applied during training and prediction.
#'
#' @section Parameters:
#' The parameters are the parameters inherited from [`PipeOpTaskPreproc`][mlr3pipelines::PipeOpTaskPreproc],
#' as well as the following parameters:
#' * `lower` :: `numeric(1)`\cr
#'   Target value of smallest item of input data. Initialized to `0`.
#' * `upper` :: `numeric(1)`\cr
#'   Target value of greatest item of input data. Initialized to `1`.
#'
#' @export
#' @examples
#' task = tsk("fuel")
#' po_scale = po("fda.scalerange", lower = -1, upper = 1)
#' task_scale = po_scale$train(list(task))[[1L]]
#' task_scale$data()
PipeOpFDAScaleRange = R6Class(
  "PipeOpFDAScaleRange",
  inherit = PipeOpTaskPreproc,
  public = list(
    #' @description Initializes a new instance of this Class.
    #' @param id (`character(1)`)\cr
    #'   Identifier of resulting object, default `"fda.scalerange"`.
    #' @param param_vals (named `list()`)\cr
    #'   List of hyperparameter settings, overwriting the hyperparameter settings that would
    #'   otherwise be set during construction. Default `list()`.
    initialize = function(id = "fda.scalerange", param_vals = list()) {
      param_set = ps(
        lower = p_dbl(tags = c("required", "train")),
        upper = p_dbl(tags = c("required", "train"))
      )
      param_set$set_values(lower = 0, upper = 1)

      super$initialize(
        id = id,
        param_set = param_set,
        param_vals = param_vals,
        packages = c("mlr3fda", "mlr3pipelines", "tf"),
        feature_types = c("tfd_reg", "tfd_irreg"),
        tags = "fda"
      )
    }
  ),
  private = list(
    .train_dt = function(dt, levels, target) {
      pars = self$param_set$get_values(tags = "train")

      for (j in names(dt)) {
        x = dt[[j]]
        domain = tf::tf_domain(x)
        scale = (pars$upper - pars$lower) / (domain[2L] - domain[1L])
        offset = -domain[1L] * scale + pars$lower
        self$state[[j]] = list(domain = domain, scale = scale, offset = offset)

        args = tf::tf_arg(x)
        if (tf::is_reg(x)) {
          new_args = offset + args * scale
        } else {
          new_args = map(args, \(arg) offset + arg * scale)
        }
        set(dt, j = j, value = invoke(tf::tfd, data = tf::tf_evaluations(x), arg = new_args))
      }
      dt
    },

    .predict_dt = function(dt, levels) {
      for (j in names(dt)) {
        x = dt[[j]]
        trafo = self$state[[j]]
        if (!all(trafo$domain == tf::tf_domain(x))) {
          error_input("Domain of new data does not match the domain of the training data.")
        }
        args = tf::tf_arg(x)
        if (tf::is_reg(x)) {
          new_args = trafo$offset + args * trafo$scale
        } else {
          new_args = map(args, \(arg) trafo$offset + arg * trafo$scale)
        }
        set(dt, j = j, value = invoke(tf::tfd, data = tf::tf_evaluations(x), arg = new_args))
      }
      dt
    }
  )
)

#' @include zzz.R
register_po("fda.scalerange", PipeOpFDAScaleRange)

test_that("PipeOpFDAZoom - basic properties", {
  pop = po("fda.zoom")
  expect_pipeop(pop)
  expect_identical(pop$id, "fda.zoom")
})

test_that("PipeOpFDAZoom works", {
  task = tsk("fuel")
  pop = po("fda.zoom", begin = 50, end = 100)
  task_zoom = train_pipeop(pop, list(task))[[1L]]
  new_data = task_zoom$data()
  expect_task(task_zoom)
  expect_shape(new_data, dim = c(129L, 4L))
  expect_identical(task_zoom$n_features, task$n_features)
  expect_named(new_data, names(new_data))
  expect_numeric(tf::tf_arg(new_data$NIR), lower = 50, upper = 100)
  expect_numeric(tf::tf_arg(new_data$UVVIS), lower = 50, upper = 100)
  expect_identical(tf::tf_domain(new_data$NIR), c(50, 100))
  expect_identical(tf::tf_domain(new_data$UVVIS), c(50, 100))
  expect_identical(new_data, predict_pipeop(pop, list(task))[[1L]]$data())

  # irregular data works
  task = tsk("dti")
  pop = po("fda.zoom", begin = 0.2, end = 0.8)
  new_data = train_pipeop(pop, list(task))[[1L]]$data()
  expect_identical(tf::tf_domain(new_data$cca), c(0.2, 0.8))
  expect_identical(tf::tf_domain(new_data$rcst), c(0.2, 0.8))
})

test_that("PipeOpFDAFourier - basic properties", {
  pop = po("fda.fourier")
  expect_pipeop(pop)
  expect_identical(pop$id, "fda.fourier")
})

test_that("PipeOpFDAFourier works", {
  task = tsk("fuel")

  pop = po("fda.fourier", type = "phase")
  task_result = train_pipeop(pop, list(task))[[1L]]
  new_data = task_result$data()
  expect_task(task_result)
  # fuel NIR has 231 points (one-sided = 116), UVVIS has 134 points (one-sided = 68)
  expect_shape(new_data, dim = c(task$nrow, 2L + 116L + 68L))
  expect_match(setdiff(names(new_data), c("heatan", "h2o")), "_fft_phase_[0-9]+$")
  walk(new_data, expect_numeric)

  pop = po("fda.fourier", type = "amplitude")
  task_result = train_pipeop(pop, list(task))[[1L]]
  new_data = task_result$data()
  expect_task(task_result)
  expect_shape(new_data, dim = c(task$nrow, 2L + 116L + 68L))
  expect_match(setdiff(names(new_data), c("heatan", "h2o")), "_fft_amplitude_[0-9]+$")
  walk(new_data, expect_numeric)
})

test_that("PipeOpFDAFourier amplitude doubling is correct", {
  # odd n: all positive frequencies doubled, no Nyquist
  pop = po("fda.fourier", type = "amplitude")
  x_odd = tf::tfd(list(c(1, 2, 3, 4, 5)), arg = 1:5)
  task_odd = as_task_regr(data.table(x = x_odd, y = 1), target = "y")
  res_odd = train_pipeop(pop, list(task_odd))[[1L]]$data()
  fft_raw = stats::fft(c(1, 2, 3, 4, 5)) / 5
  expected = Mod(fft_raw[1:3])
  expected[2:3] = expected[2:3] * 2
  actual = as.numeric(res_odd[1L, grep("^x_fft", names(res_odd)), with = FALSE])
  expect_equal(actual, expected, tolerance = 1e-10)

  # even n: Nyquist (last) not doubled
  pop = po("fda.fourier", type = "amplitude")
  x_even = tf::tfd(list(c(1, 2, 3, 4)), arg = 1:4)
  task_even = as_task_regr(data.table(x = x_even, y = 1), target = "y")
  res_even = train_pipeop(pop, list(task_even))[[1L]]$data()
  fft_raw4 = stats::fft(c(1, 2, 3, 4)) / 4
  expected4 = Mod(fft_raw4[1:3])
  expected4[2] = expected4[2] * 2
  actual4 = as.numeric(res_even[1L, grep("^x_fft", names(res_even)), with = FALSE])
  expect_equal(actual4, expected4, tolerance = 1e-10)

  # n = 2: DC + Nyquist only, nothing doubled
  pop = po("fda.fourier", type = "amplitude")
  x_two = tf::tfd(list(c(1, 3)), arg = 1:2)
  task_two = as_task_regr(data.table(x = x_two, y = 1), target = "y")
  res_two = train_pipeop(pop, list(task_two))[[1L]]$data()
  fft_raw2 = stats::fft(c(1, 3)) / 2
  expected2 = Mod(fft_raw2[1:2])
  actual2 = as.numeric(res_two[1L, grep("^x_fft", names(res_two)), with = FALSE])
  expect_equal(actual2, expected2, tolerance = 1e-10)
})

test_that("PipeOpFDAFourier predict works", {
  task = tsk("fuel")
  pop = po("fda.fourier")
  train_pipeop(pop, list(task))
  task_result = predict_pipeop(pop, list(task))[[1L]]
  expect_task(task_result)
})

test_that("PipeOpFDARegister - basic properties", {
  pop = po("fda.register")
  expect_pipeop(pop)
  expect_identical(pop$id, "fda.register")
})

test_that("PipeOpFDARegister - affine shift registration works", {
  skip_if_not_installed("withr")
  withr::local_seed(1L)

  t = seq(0, 2 * pi, length.out = 51L)
  shifts = c(-0.3, -0.1, 0, 0.2, 0.4)
  mat = t(sapply(shifts, function(s) sin(t + s)))
  f = tf::tfd(mat, arg = t)
  dt = data.table(y = rnorm(length(shifts)), f = f)
  task = as_task_regr(dt, target = "y")

  pop = po("fda.register", method = "affine", args = list(type = "shift"))
  task_reg = train_pipeop(pop, list(task))[[1L]]
  expect_task(task_reg)
  new_data = task_reg$data()
  expect_named(new_data, c("y", "f"))
  expect_class(new_data$f, "tfd")
  expect_length(new_data$f, length(shifts))

  # template stored in state
  expect_list(pop$state$templates, len = 1L)
  expect_class(pop$state$templates$f, "tfd")
  expect_length(pop$state$templates$f, 1L)

  # predict returns same length
  new_pred = predict_pipeop(pop, list(task))[[1L]]$data()
  expect_length(new_pred$f, length(shifts))
})

test_that("PipeOpFDARegister - predict uses stored template", {
  skip_if_not_installed("withr")
  withr::local_seed(42L)

  t = seq(0, 1, length.out = 41L)
  make_curves = function(shifts) {
    mat = t(sapply(shifts, function(s) dnorm(t, mean = 0.5 + s, sd = 0.1)))
    tf::tfd(mat, arg = t)
  }
  train_f = make_curves(c(-0.1, 0, 0.1))
  test_f = make_curves(c(-0.05, 0.05))

  train_dt = data.table(y = 1:3 + 0, f = train_f)
  test_dt = data.table(y = 1:2 + 0, f = test_f)
  train_task = as_task_regr(train_dt, target = "y")
  test_task = as_task_regr(test_dt, target = "y")

  pop = po("fda.register", method = "affine", args = list(type = "shift"))
  train_pipeop(pop, list(train_task))
  tmpl_before = pop$state$templates$f

  predict_pipeop(pop, list(test_task))
  # state unchanged after predict
  expect_equal(pop$state$templates$f, tmpl_before)

  # aligning test data via predict_pipeop matches calling tf_register directly with the stored template
  reg_direct = tf::tf_register(
    test_f,
    method = "affine",
    template = tmpl_before,
    type = "shift",
    store_x = FALSE
  )
  expected = tf::tf_aligned(reg_direct)
  pred = predict_pipeop(pop, list(test_task))[[1L]]$data()
  expect_equal(pred$f, expected)
})

test_that("PipeOpFDARegister - cc method works", {
  skip_if_not_installed("withr")
  withr::local_seed(7L)

  t = seq(0, 1, length.out = 81L)
  mat = t(sapply(c(-0.08, -0.04, 0, 0.04, 0.08), function(s) dnorm(t, 0.5 + s, 0.12)))
  f = tf::tfd(mat, arg = t)
  dt = data.table(y = seq_len(5L) + 0, f = f)
  task = as_task_regr(dt, target = "y")

  pop = po("fda.register", method = "cc", args = list(nbasis = 6L, lambda = 1e-3), max_iter = 2L)
  new_data = suppressMessages(train_pipeop(pop, list(task)))[[1L]]$data()
  expect_class(new_data$f, "tfd")
  expect_length(new_data$f, 5L)
  expect_class(pop$state$templates$f, "tfd")
})

test_that("PipeOpFDARegister - srvf method works", {
  skip_if_not_installed("fdasrvf")
  skip_if_not_installed("withr")
  withr::local_seed(3L)

  t = seq(0, 1, length.out = 41L)
  mat = t(sapply(c(-0.05, 0, 0.05, 0.1), function(s) dnorm(t, 0.5 + s, 0.12)))
  f = tf::tfd(mat, arg = t)
  dt = data.table(y = 1:4 + 0, f = f)
  task = as_task_regr(dt, target = "y")

  pop = po("fda.register", method = "srvf")
  new_data = suppressMessages(train_pipeop(pop, list(task)))[[1L]]$data()
  expect_class(new_data$f, "tfd")
  expect_length(new_data$f, 4L)
  expect_class(pop$state$templates$f, "tfd")

  new_pred = suppressMessages(predict_pipeop(pop, list(task)))[[1L]]$data()
  expect_length(new_pred$f, 4L)
})

test_that("PipeOpFDARegister - multiple columns and fuel task", {
  task = tsk("fuel")
  pop = po("fda.register", method = "affine", args = list(type = "shift"))
  task_reg = suppressWarnings(suppressMessages(train_pipeop(pop, list(task))))[[1L]]
  expect_task(task_reg)
  expect_set_equal(task$feature_names, task_reg$feature_names)
  expect_list(pop$state$templates, len = 2L)
  expect_named(pop$state$templates, c("NIR", "UVVIS"))
})

test_that("PipeOpFDATsfeatures - basic properties", {
  pop = po("fda.tsfeats")
  expect_pipeop(pop)
  expect_identical(pop$id, "fda.tsfeats")
})

test_that("PipeOpFDATsfeatures works", {
  skip_if_not_installed("tsfeatures")

  task = tsk("fuel")
  pop = po("fda.tsfeats")
  task_tsfeats = train_pipeop(pop, list(task))[[1L]]
  new_data = task_tsfeats$data()
  expect_task(task_tsfeats)
  expect_shape(new_data, dim = c(129L, 34L))
  expect_named(new_data, names(new_data))

  # single feature works
  pop = po("fda.tsfeats", features = "entropy")
  task_tsfeats = train_pipeop(pop, list(task))[[1L]]
  new_data = task_tsfeats$data()
  walk(new_data, expect_numeric)
  expect_shape(new_data, dim = c(129L, 4L))
  expect_named(new_data, c("heatan", "h2o", "NIR_entropy", "UVVIS_entropy"))

  # multiple features work
  pop = po("fda.tsfeats", features = c("frequency", "stl_features"))
  task_tsfeats = train_pipeop(pop, list(task))[[1L]]
  new_data = task_tsfeats$data()
  walk(new_data, expect_numeric)
  expect_shape(new_data, dim = c(129L, 20L))
  expect_match(setdiff(names(new_data), c("heatan", "h2o")), "NIR_|UVVIS_")

  # irregular data works
  task = tsk("dti")
  pop = po("fda.tsfeats")
  task_tsfeats = train_pipeop(pop, list(task))[[1L]]
  new_data = task_tsfeats$data()
  expect_task(task_tsfeats)
})

test_that("PipeOpFPCA - basic properties", {
  pop = po("fda.fpca")
  expect_pipeop(pop)
  expect_identical(pop$id, "fda.fpca")
})

test_that("PipeOpFPCA works", {
  skip_if_not_installed("withr")
  withr::local_seed(1234L)
  # single col works
  dt = data.table(
    id = c("Ann", "Ann", "Ann", "Bob", "Bob", "Bob"),
    arg = rep(1:3, 2L),
    value = 1:6
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  y = c(1, 2)
  dt = data.table(f = f, y = y)
  task = as_task_regr(dt, target = "y")

  pop = po("fda.fpca")
  task_fpc = train_pipeop(pop, list(task))[[1L]]
  expect_task(task_fpc)
  new_data = task_fpc$data()
  expect_shape(new_data, dim = c(2L, 2L))
  expect_named(new_data, c("y", "f_pc_1"))
  expect_numeric(new_data$f_pc_1, len = 2)
  expect_identical(new_data, predict_pipeop(pop, list(task))[[1L]]$data())

  # n_components works
  dt = data.table(y = rnorm(15L), f = tf::tf_rgp(15L))
  task = as_task_regr(dt, target = "y")
  pop = po("fda.fpca", n_components = 2L)
  new_data = train_pipeop(pop, list(task))[[1L]]$data()
  expect_shape(new_data, dim = c(15L, 3L))
  expect_named(new_data, c("y", "f_pc_1", "f_pc_2"))

  # multiple cols work
  dt = data.table(
    id = rep(1:10, each = 3L),
    arg = rep(1:3, 10L),
    value = 1:10 + rnorm(1L)
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  dt = data.table(y = rnorm(10L), f = f, g = f, h = f)
  task = as_task_regr(dt, target = "y")
  pop = po("fda.fpca")
  task_fpc = train_pipeop(pop, list(task))[[1L]]
  new_data = task_fpc$data()
  expect_task(task_fpc)
  expect_shape(new_data, dim = c(10L, 10L))
  nms = c(
    "y",
    "f_pc_1",
    "f_pc_2",
    "f_pc_3",
    "g_pc_1",
    "g_pc_2",
    "g_pc_3",
    "h_pc_1",
    "h_pc_2",
    "h_pc_3"
  )
  expect_named(new_data, nms)
  expect_identical(new_data, predict_pipeop(pop, list(task))[[1L]]$data())

  # n_components works
  pop = po("fda.fpca", n_components = 2L)
  new_data = train_pipeop(pop, list(task))[[1L]]$data()
  expect_shape(new_data, dim = c(10L, 7L))
  nms = c(
    "y",
    "f_pc_1",
    "f_pc_2",
    "g_pc_1",
    "g_pc_2",
    "h_pc_1",
    "h_pc_2"
  )
  expect_named(new_data, nms)

  # affect_columns works
  pop = po("fda.fpca", affect_columns = selector_name("f"))
  task_fpc = train_pipeop(pop, list(task))[[1L]]
  expect_set_equal(task_fpc$feature_names, c("f_pc_1", "f_pc_2", "f_pc_3", "g", "h"))

  # does not touch irreg
  dt = data.table(
    id = c("Ann", "Ann", "Ann", "Bob", "Bob"),
    arg = c(1L, 7L, 2L, 3L, 5L),
    value = 1:5
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  y = 1:2
  dt = data.table(f = f, y = y)
  task = as_task_regr(dt, target = "y")
  pop = po("fda.fpca")
  task_fpca = train_pipeop(pop, list(task))[[1L]]
  expect_set_equal(task$feature_names, task_fpca$feature_names)
})

test_that("tf does not support NAs", {
  # Various PipeOps assume that there are no functional NAs
  # This test will inform us whether this feature is implemented,
  # in which case we then have to adress this case in the PipeOps
  # https://github.com/tidyfun/tf/issues/33
  # Currently, NA functions are dropped
  d = data.frame(time = 1, value = NA_real_, id = "1", stringsAsFactors = FALSE)
  x = tf::tfd(d, arg = "time", value = "value", id = "id")
  expect_length(x, 0L)
})

test_that("pofu has no surprises: irreg", {
  # pofu can successfully determine that the two cca columns are identical
  task = tsk("dti")$select("cca")
  task1 = task$clone(deep = TRUE)
  task1$id = "test"
  taskout = po("featureunion")$train(list(task, task1))[[1L]]
  expect_permutation(taskout$feature_names, "cca")
})

test_that("pofu has no surprises: reg", {
  # pofu can successfully determine that the two cca columns are identical
  task = tsk("fuel")$select("NIR")
  task1 = task$clone(deep = TRUE)
  task1$id = "test"
  taskout = po("featureunion")$train(list(task, task1))[[1L]]
  expect_permutation(taskout$feature_names, "NIR")
})

test_that("PipeOpFDAScaleRange - basic properties", {
  pop = po("fda.scalerange")
  expect_pipeop(pop)
  expect_identical(pop$id, "fda.scalerange")
})

test_that("PipeOpFDAScaleRange works", {
  task = tsk("fuel")
  pop = po("fda.scalerange")
  task_scale = train_pipeop(pop, list(task))[[1L]]
  new_data = task_scale$data()
  expect_task(task_scale)
  expect_shape(new_data, dim = c(129L, 4L))
  expect_identical(task_scale$n_features, task$n_features)
  expect_named(new_data, names(new_data))
  expect_numeric(tf::tf_arg(new_data$NIR), lower = 0, upper = 1)
  expect_numeric(tf::tf_arg(new_data$UVVIS), lower = 0, upper = 1)
  expect_identical(new_data, predict_pipeop(pop, list(task))[[1L]]$data())

  # different range works
  pop = po("fda.scalerange", lower = -1, upper = 1)
  task_scale = train_pipeop(pop, list(task))[[1L]]
  new_data = task_scale$data()
  expect_equal(tf::tf_domain(new_data$NIR), c(-1, 1))
  expect_equal(range(tf::tf_arg(new_data$NIR)), c(-1, 1))
  expect_equal(tf::tf_domain(new_data$UVVIS), c(-1, 1))
  expect_equal(range(tf::tf_arg(new_data$UVVIS)), c(-1, 1))

  # throws error if new data has different domain
  pop = po("fda.scalerange")
  train_pipeop(pop, list(task))
  expect_error(
    predict_pipeop(pop, list(task_scale)),
    "Domain of new data does not match the domain of the training data."
  )

  # irregular data works
  task = tsk("dti")
  pop = po("fda.scalerange", lower = -1, upper = 1)
  new_data = train_pipeop(pop, list(task))[[1L]]$data()
  expect_equal(tf::tf_domain(new_data$cca), c(-1, 1))
  expect_equal(tf::tf_domain(new_data$rcst), c(-1, 1))
})

test_that("hash input for tfd vectors", {
  # this tests that the hash of the DataBackendDataTable is not influenced by baggage of the evaluator attribute
  # of tfd vectors
  x_reg = tsk("fuel")$data(cols = "NIR")[[1]][1:5]
  e_reg = attr(x_reg, "evaluator")
  x_irreg = tsk("dti")$data(cols = "cca")[[1]][1:5]
  e_irreg = attr(x_irreg, "evaluator")

  e_reg1 = e_reg
  environment(e_reg1) = new.env()
  x_reg1 = x_reg
  attr(x_reg1, "evaluator") = e_reg1

  d1 = as_data_backend(data.table(
    y = 1:5,
    x_reg = x_reg1,
    x_irreg = x_irreg
  ))

  e_irreg1 = e_irreg
  environment(e_irreg1) = new.env()
  x_irreg1 = x_irreg
  attr(x_irreg1, "evaluator") = e_irreg1

  d2 = as_data_backend(data.table(
    y = 1:5,
    x_reg = x_reg,
    x_irreg = x_irreg1
  ))

  d3 = as_data_backend(data.table(
    y = 1:5,
    x_reg = x_reg,
    x_irreg = x_irreg
  ))

  expect_identical(d1$hash, d2$hash)
  expect_identical(d2$hash, d3$hash)
})

test_that("PipeOpFDACor - basic properties", {
  pop = po("fda.cor")
  expect_pipeop(pop)
  expect_identical(pop$id, "fda.cor")
})

test_that("PipeOpFDACor works", {
  skip_if_not_installed("withr")
  withr::local_seed(1234L)
  dt = data.table(y = 1:100, x1 = tf::tf_rgp(100L), x2 = tf::tf_rgp(100L), x3 = tf::tf_rgp(100L))
  task = as_task_regr(dt, target = "y")

  pop = po("fda.cor")
  task_cor = train_pipeop(pop, list(task))[[1L]]
  expect_task(task_cor)
  new_data = task_cor$data()
  expect_shape(new_data, dim = c(100L, 4L))
  expect_named(new_data, c("y", "x1_x2_cor", "x1_x3_cor", "x2_x3_cor"))
  expect_numeric(new_data$x1_x2_cor, lower = -1, upper = 1, len = 100)
  expect_numeric(new_data$x1_x3_cor, lower = -1, upper = 1, len = 100)
  expect_numeric(new_data$x2_x3_cor, lower = -1, upper = 1, len = 100)

  # single col gives warning
  task$select("x1")
  pop = po("fda.cor")
  expect_warning(task_cor <- train_pipeop(pop, list(task))[[1L]], "task has fewer than 2 columns")
  expect_identical(task$data(), task_cor$data())

  # different domain throws error
  dt_domain = copy(dt)[, x1 := tf::tf_rgp(100L, 20:120)]
  task = as_task_regr(dt_domain, target = "y")
  pop = po("fda.cor")
  expect_error(train_pipeop(pop, list(task)), "Domain of x1 and x2 do not match")

  # does not touch irreg
  dt[, x1 := tf::tf_sparsify(x1)]
  task = as_task_regr(dt, target = "y")
  task_cor = train_pipeop(pop, list(task))[[1L]]
  expect_set_equal(task_cor$feature_names, c("x1", "x2_x3_cor"))
})

test_that("PipeOpFDABsignal - basic properties", {
  pop = po("fda.bsignal")
  expect_pipeop(pop)
  expect_identical(pop$id, "fda.bsignal")
})

test_that("PipeOpFDABsignal works", {
  skip_if_not_installed("FDboost")

  task = tsk("fuel")
  pop = po("fda.bsignal")
  task_bsignal = train_pipeop(pop, list(task))[[1L]]
  new_data = task_bsignal$data()
  expect_task(task_bsignal)
  expect_shape(new_data, dim = c(129L, 30L))
  expect_named(new_data, names(new_data))

  # irregular data works
  task = tsk("dti")
  pop = po("fda.bsignal")
  task_bsignal = train_pipeop(pop, list(task))[[1L]]
  new_data = task_bsignal$data()
  expect_task(task_bsignal)
})

test_that("PipeOpFDAInterpol - basic properties", {
  pop = po("fda.interpol")
  expect_pipeop(pop)
  expect_identical(pop$id, "fda.interpol")
})

test_that("PipeOpFDAInterpol input validation works", {
  expect_error(po("fda.interpol", grid = c("union", "intersect")))
  expect_error(po("fda.interpol", grid = "unionh"))
  expect_error(po("fda.interpol", grid = list(1L)))
  expect_error(po("fda.interpol", grid = logical(1L)))
  expect_error(po("fda.interpol", grid = factor(1L)))
  expect_error(po("fda.interpol", grid = numeric(0L)))
  expect_error(po("fda.interpol", grid = 1:3, method = c("linear", "spline")))
  expect_error(po("fda.interpol", grid = 1:3, method = "cube"))
  task = tsk("fuel")
  pop = po("fda.interpol", grid = 1:3, left = 1, right = 2)
  expect_error(train_pipeop(pop, list(task)))
  pop = po("fda.interpol", grid = 10L, left = 2, right = 1)
  expect_error(train_pipeop(pop, list(task)))
  pop = po("fda.interpol", grid = 10L, left = 2)
  expect_error(
    train_pipeop(pop, list(task)),
    "Either both or none of 'left' and 'right' must be specified."
  )
  pop = po("fda.interpol", grid = 10L, right = 2)
  expect_error(
    train_pipeop(pop, list(task)),
    "Either both or none of 'left' and 'right' must be specified."
  )
})

test_that("PipeOpFDAInterpol extrapolation works", {
  # extrapolate with fill_extend method
  dt = data.table(
    id = rep(1:2, each = 5L),
    arg = rep(1:5, 2L),
    value = c(NA, 2, 5, 5, 7, 3, 5, 10, 2, NA)
  )
  dt_in = data.table(y = 1:2, f = tf::tfd(dt, id = "id", arg = "arg", value = "value"))
  task = as_task_regr(dt_in, target = "y")
  pop = po("fda.interpol", grid = 1:5, method = "fill_extend")
  actual = train_pipeop(pop, list(task))[[1L]]$data()
  setnafill(dt, fill = 2L)
  expected = data.table(y = 1:2, f = tf::tfd(dt, id = "id", arg = "arg", value = "value"))
  expect_equal(actual, expected, ignore_attr = TRUE)
  # throw warning if extrapolation is not possible
  pop = po("fda.interpol", grid = 1:5)
  expect_warning(train_pipeop(pop, list(task)))
})

test_that("PipeOpFDAInterpol works with minmax", {
  # tfr doesnt't have an effect
  dt = data.table(
    id = rep(1:2, each = 5L),
    arg = rep(1:5, 2L),
    value = c(1, 2, 5, 5, 7, 3, 5, 10, 2, 12)
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  dt = data.table(y = 1:2, f = f)
  task = as_task_regr(dt, target = "y")
  pop = po("fda.interpol", grid = "minmax")
  task_interpol = train_pipeop(pop, list(task))[[1L]]
  expect_equal(task_interpol$data(), dt)

  # tfi works with same min and max
  dt = data.table(
    id = rep(1:2, each = 5L),
    arg = rep(1:5, 2L),
    value = c(1, NA, 5, 5, 7, 3, 5, 10, NA, 12)
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  dt = data.table(y = 1:2, f = f)
  task = as_task_regr(dt, target = "y")
  pop = po("fda.interpol", grid = "minmax")
  task_interpol = train_pipeop(pop, list(task))[[1L]]
  dt = data.table(
    id = rep(1:2, each = 5L),
    arg = rep(1:5, 2L),
    value = c(1, 3, 5, 5, 7, 3, 5, 10, 11, 12)
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  expected = data.table(y = 1:2, f = f)
  expect_equal(task_interpol$data(), expected)

  # tfi works with different min and max
  dt = data.table(
    id = c(rep(1L, 3L), rep(2L, 6L)),
    arg = c(3:5, 1:6),
    value = c(2, 5, 6, 1, 3, 4, 5, 6, 7)
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  dt = data.table(y = 1:2, f = f)
  task = as_task_regr(dt, target = "y")
  pop = po("fda.interpol", grid = "minmax")
  task_interpol = train_pipeop(pop, list(task))[[1L]]
  dt = data.table(
    id = rep(1:2, each = 3L),
    arg = rep(3:5, 2L),
    value = c(2, 5, 6, 4, 5, 6)
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  expected = data.table(y = 1:2, f = f)
  expect_equal(task_interpol$data(), expected)
})

test_that("PipeOpFDAInterpol works with intersect", {
  # tfr works
  dt = data.table(
    id = rep(1:2, each = 5L),
    arg = rep(1:5, 2L),
    value = c(1, 2, 5, 5, 7, 3, 5, 10, 2, 12)
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  dt = data.table(y = 1:2, f = f)
  task = as_task_regr(dt, target = "y")
  pop = po("fda.interpol", grid = "intersect")
  task_interpol = train_pipeop(pop, list(task))[[1L]]
  expect_equal(task_interpol$data(), dt)

  # tfi works with same min and max
  dt = data.table(
    id = rep(1:2, each = 5L),
    arg = rep(1:5, 2L),
    value = c(1, NA, 5, 5, 7, 3, 5, 10, NA, 12)
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  dt = data.table(y = 1:2, f = f)
  task = as_task_regr(dt, target = "y")
  pop = po("fda.interpol", grid = "intersect")
  task_interpol = train_pipeop(pop, list(task))[[1L]]
  dt = data.table(
    id = rep(1:2, each = 3L),
    arg = rep(c(1, 3, 5), 2L),
    value = c(1, 5, 7, 3, 10, 12)
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  expected = data.table(y = 1:2, f = f)
  expect_equal(task_interpol$data(), expected)

  # tfi works with different min and max
  dt = data.table(
    id = c(rep(1L, 3L), rep(2L, 6L)),
    arg = c(3:5, 1:6),
    value = c(2, 5, 6, 1, 3, 4, 5, 6, 7)
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  dt = data.table(y = 1:2, f = f)
  task = as_task_regr(dt, target = "y")
  pop = po("fda.interpol", grid = "intersect")
  task_interpol = train_pipeop(pop, list(task))[[1L]]
  dt = data.table(
    id = rep(1:2, each = 3L),
    arg = rep(3:5, 2L),
    value = c(2, 5, 6, 4, 5, 6)
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  expected = data.table(y = 1:2, f = f)
  expect_equal(task_interpol$data(), expected)
})

test_that("PipeOpFDAInterpol works with union", {
  # tfr works
  dt = data.table(
    id = rep(1:2, each = 5L),
    arg = rep(1:5, 2L),
    value = c(1, 2, 5, 5, 7, 3, 5, 10, 2, 12)
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  dt = data.table(y = 1:2, f = f)
  task = as_task_regr(dt, target = "y")
  pop = po("fda.interpol", grid = "union")
  task_interpol = train_pipeop(pop, list(task))[[1L]]
  expect_equal(task_interpol$data(), dt)

  # works with default
  pop = po("fda.interpol")
  task_interpol = train_pipeop(pop, list(task))[[1L]]
  expect_equal(task_interpol$data(), dt)

  # tfi works with same min and max
  dt = data.table(
    id = rep(1:2, each = 5L),
    arg = rep(1:5, 2L),
    value = c(1, NA, 5, 5, 7, 3, 5, 10, NA, 12)
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  dt = data.table(y = 1:2, f = f)
  task = as_task_regr(dt, target = "y")
  pop = po("fda.interpol", grid = "union")
  task_interpol = train_pipeop(pop, list(task))[[1L]]
  dt = data.table(
    id = rep(1:2, each = 5L),
    arg = rep(1:5, 2L),
    value = c(1, 3, 5, 5, 7, 3, 5, 10, 11, 12)
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  expected = data.table(y = 1:2, f = f)
  expect_equal(task_interpol$data(), expected)

  # tfi works with different min and max
  dt = data.table(
    id = c(rep(1L, 3L), rep(2L, 6L)),
    arg = c(3:5, 1:6),
    value = c(2, NA, 5, 1, 3, 4, 5, 6, 7)
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  dt = data.table(y = 1:2, f = f)
  task = as_task_regr(dt, target = "y")
  pop = po("fda.interpol", grid = "union")
  expect_warning(task_interpol <- train_pipeop(pop, list(task))[[1L]])
  dt = data.table(
    id = c(rep(1L, 3L), rep(2L, 6L)),
    arg = c(3:5, 1:6),
    value = c(2, 3.5, 5, 1, 3, 4, 5, 6, 7)
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  expected = data.table(y = 1:2, f = f)
  expect_equal(task_interpol$data(), expected)
})

test_that("PipeOpFDAInterpol works with custom grid", {
  # tfr works
  dt = data.table(
    id = rep(1:2, each = 5L),
    arg = rep(1:5, 2L),
    value = c(1, 2, 5, 5, 7, 3, 5, 10, 2, 12)
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  dt = data.table(y = 1:2, f = f)
  task = as_task_regr(dt, target = "y")
  pop = po("fda.interpol", grid = 3:5)
  task_interpol = train_pipeop(pop, list(task))[[1L]]
  dt = data.table(
    id = rep(1:2, each = 3L),
    arg = rep(3:5, 2L),
    value = c(5, 5, 7, 10, 2, 12)
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  expected = data.table(y = 1:2, f = f)
  expect_equal(task_interpol$data(), expected)

  # outside of range
  pop = po("fda.interpol", grid = 3:7)
  expect_error(train_pipeop(pop, list(task)), "The grid must be within the range of the domain.")
  pop = po("fda.interpol", grid = -1:3)
  expect_error(train_pipeop(pop, list(task)), "The grid must be within the range of the domain.")

  # tfi works with same min and max
  dt = data.table(
    id = rep(1:2, each = 5L),
    arg = rep(1:5, 2L),
    value = c(1, NA, 5, 5, 7, 3, 5, 10, NA, 12)
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  dt = data.table(y = 1:2, f = f)
  task = as_task_regr(dt, target = "y")
  pop = po("fda.interpol", grid = 3:5)
  task_interpol = train_pipeop(pop, list(task))[[1L]]
  dt = data.table(
    id = rep(1:2, each = 3L),
    arg = rep(3:5, 2L),
    value = c(5, 5, 7, 10, 11, 12)
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  expected = data.table(y = 1:2, f = f)
  expect_equal(task_interpol$data(), expected)

  # tfi works with different min and max
  dt = data.table(
    id = c(rep(1L, 3L), rep(2L, 6L)),
    arg = c(3:5, 1:6),
    value = c(2, 5, 6, 1, 3, 4, 5, 6, 7)
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  dt = data.table(y = 1:2, f = f)
  task = as_task_regr(dt, target = "y")
  pop = po("fda.interpol", grid = 3:5)
  task_interpol = train_pipeop(pop, list(task))[[1L]]
  dt = data.table(
    id = rep(1:2, each = 3L),
    arg = rep(3:5, 2L),
    value = c(2, 5, 6, 4, 5, 6)
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  expected = data.table(y = 1:2, f = f)
  expect_equal(task_interpol$data(), expected)
})

test_that("PipeOpFDAInterpol works with grid length + left and right", {
  # tfr works with integer output grid
  dt = data.table(
    id = rep(1:2, each = 5L),
    arg = rep(1:5, 2L),
    value = c(1, 2, 5, 5, 7, 3, 5, 10, 2, 12)
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  dt = data.table(y = 1:2, f = f)
  task = as_task_regr(dt, target = "y")
  pop = po("fda.interpol", grid = 3L, left = 2, right = 4)
  task_interpol = train_pipeop(pop, list(task))[[1L]]
  dt = data.table(
    id = rep(1:2, each = 3L),
    arg = rep(2:4, 2L),
    value = c(2, 5, 5, 5, 10, 2)
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  expected = data.table(y = 1:2, f = f)
  expect_equal(task_interpol$data(), expected)

  # tfr works with numeric output grid
  dt = data.table(
    id = rep(1:2, each = 5L),
    arg = rep(1:5, 2L),
    value = c(1, 2, 5, 5, 7, 3, 5, 10, 2, 12)
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  dt = data.table(y = 1:2, f = f)
  task = as_task_regr(dt, target = "y")
  pop = po("fda.interpol", grid = 3L, left = 2, right = 5)
  task_interpol = train_pipeop(pop, list(task))[[1L]]
  dt = data.table(
    id = rep(1:2, each = 3L),
    arg = rep(c(2, 3.5, 5), 2L),
    value = c(2, 5, 7, 5, 6, 12)
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  expected = data.table(y = 1:2, f = f)
  expect_equal(task_interpol$data(), expected)

  # tfi works
  dt = data.table(
    id = c(rep(1L, 3L), rep(2L, 6L)),
    arg = c(3:5, 1:6),
    value = c(2, 5, 6, 1, 3, 4, 5, 6, 7)
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  dt = data.table(y = 1:2, f = f)
  task = as_task_regr(dt, target = "y")
  pop = po("fda.interpol", grid = 3L, left = 3, right = 5)
  task_interpol = train_pipeop(pop, list(task))[[1L]]
  dt = data.table(
    id = rep(1:2, each = 3L),
    arg = rep(3:5, 2L),
    value = c(2, 5, 6, 4, 5, 6)
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  expected = data.table(y = 1:2, f = f)
  expect_equal(task_interpol$data(), expected)
})

test_that("PipeOpFDAInterpol method arg works", {
  dt_in = data.table(
    id = rep(1:2, each = 5L),
    arg = rep(1:5, 2L),
    value = c(1, 2, 5, 5, 7, 3, 5, 10, 2, 12)
  )
  dt_out = data.table(
    id = rep(1:2, each = 3L),
    arg = rep(3:5, 2L),
    value = c(5, 5, 7, 10, 2, 12)
  )
  methods = c("linear", "spline", "fill_extend", "locf", "nocb")
  walk(methods, function(method) {
    f = tf::tfd(dt_in, id = "id", arg = "arg", value = "value")
    dt = data.table(y = 1:2, f = f)
    task = as_task_regr(dt, target = "y")
    pop = po("fda.interpol", grid = 3:5, method = method)
    task_interpol = train_pipeop(pop, list(task))[[1L]]

    evaluator = paste0("tf_approx_", method)
    f = do.call(tf::tfd, list(data = dt_out, id = "id", arg = "arg", value = "value", evaluator = evaluator))
    expected = data.table(y = 1:2, f = f)
    expect_equal(task_interpol$data(), expected)
  })
})

test_that("PipeOpFDARandomEffect - basic properties", {
  pop = po("fda.random_effect")
  expect_pipeop(pop)
  expect_identical(pop$id, "fda.random_effect")
})

test_that("PipeOpFDARandomEffect works", {
  skip_if_not_installed("lme4")

  task = tsk("fuel")
  pop = po("fda.random_effect")
  task_fre = suppressWarnings(train_pipeop(pop, list(task))[[1L]])
  new_data = task_fre$data()
  expect_task(task_fre)
  expect_shape(new_data, dim = c(129L, 6L))
  expect_named(
    new_data,
    c("heatan", "h2o", "NIR_random_intercept", "NIR_random_slope", "UVVIS_random_intercept", "UVVIS_random_slope")
  )

  # irregular data works
  task = tsk("dti")
  pop = po("fda.random_effect")
  task_fre = train_pipeop(pop, list(task))[[1L]]
  new_data = task_fre$data()
  expect_task(task_fre)
  expect_shape(new_data, dim = c(340L, 6L))
  expect_named(
    new_data,
    c("pasat", "sex", "cca_random_intercept", "cca_random_slope", "rcst_random_intercept", "rcst_random_slope")
  )
})

test_that("PipeOpFDAFlatten - basic properties", {
  pop = po("fda.flatten")
  expect_pipeop(pop)
  expect_identical(pop$id, "fda.flatten")
})

test_that("PipeOpFDAFlatten works", {
  task = tsk("fuel")
  pop = po("fda.flatten")
  x = train_pipeop(pop, list(task))[[1L]]
  expected_features = c(
    "h2o",
    paste0("UVVIS_", 1:134),
    paste0("NIR_", 1:231)
  )
  expect_set_equal(x$feature_names, expected_features)

  pop = po("fda.flatten")
  pop$param_set$values$affect_columns = selector_name("UVVIS")
  x = train_pipeop(pop, list(task))[[1L]]
  expected_features = c(
    "h2o",
    paste0("UVVIS_", 1:134),
    "NIR"
  )
  expect_set_equal(x$feature_names, expected_features)

  pop = po("fda.flatten")
  pop$param_set$values$affect_columns = selector_name("..xyz")
  x = train_pipeop(pop, list(task))[[1L]]
  expected_features = c("h2o", "UVVIS", "NIR")
  expect_set_equal(x$feature_names, expected_features)
})

test_that("PipeOpFDAFlatten works with name clashes", {
  dt = tsk("fuel")$select("NIR")$data(1)
  dt$NIR_1 = 1
  task = as_task_regr(dt, target = "heatan")
  pop = po("fda.flatten")
  taskout = train_pipeop(pop, train_pipeop(pop, list(task)))[[1L]]
  expect_true("NIR_1_1" %in% taskout$feature_names)
})

test_that("PipeOpFDAFlatten works with tfr and tfi", {
  # tfr works
  dt = data.table(
    id = rep(1:2, each = 5L),
    arg = rep(1:5, 2L),
    value = c(1, 2, 5, 5, 7, 3, 5, 10, 2, 12)
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  dt = data.table(y = 1:2, f = f)
  task = as_task_regr(dt, target = "y")
  pop = po("fda.flatten")
  task_flat = train_pipeop(pop, list(task))[[1L]]
  expected = data.table(
    y = 1:2,
    f_1 = c(1, 3),
    f_2 = c(2, 5),
    f_3 = c(5, 10),
    f_4 = c(5, 2),
    f_5 = c(7, 12)
  )
  expect_set_equal(c("f_1", "f_2", "f_3", "f_4", "f_5"), task_flat$feature_names)
  expect_equal(task_flat$data(), expected)

  # tfi works
  dt = data.table(
    id = c(rep(1L, 3L), rep(2L, 6L)),
    arg = c(3:5, 1:6),
    value = c(2, 5, 6, 1, 3, 4, 5, 6, 7)
  )
  f = tf::tfd(dt, id = "id", arg = "arg", value = "value")
  dt = data.table(y = 1:2, f = f)
  task = as_task_regr(dt, target = "y")
  pop = po("fda.flatten")
  task_flat = train_pipeop(pop, list(task))[[1L]]
  expected = data.table(
    y = 1:2,
    f_1 = c(NA, 1),
    f_2 = c(NA, 3),
    f_3 = c(2, 4),
    f_4 = c(5, 5),
    f_5 = c(6, 6),
    f_6 = c(NA, 7)
  )
  expect_set_equal(c("f_1", "f_2", "f_3", "f_4", "f_5", "f_6"), task_flat$feature_names)
  expect_equal(task_flat$data(), expected)
})

test_that("PipeOpFDAWavelets - basic properties", {
  pop = po("fda.wavelets")
  expect_pipeop(pop)
  expect_identical(pop$id, "fda.wavelets")
})

test_that("PipeOpFDAWavelets input validation", {
  skip_if_not_installed("wavelets")
  expect_no_error(po("fda.wavelets", filter = wavelets::wt.filter()))
  expect_no_error(po("fda.wavelets", filter = "la8"))
  expect_no_error(po("fda.wavelets", filter = 1:10))
  expect_snapshot(po("fda.wavelets", filter = "la4"), error = TRUE)
  expect_snapshot(po("fda.wavelets", filter = "invalid_filter"), error = TRUE)
  expect_snapshot(po("fda.wavelets", filter = c(1, 2, 3)), error = TRUE)
  expect_snapshot(po("fda.wavelets", filter = list("la8")), error = TRUE)
})

test_that("PipeOpFDAWavelets works", {
  skip_if_not_installed("wavelets")
  task = tsk("fuel")

  pop = po("fda.wavelets")
  task_wav = train_pipeop(pop, list(task))[[1L]]
  new_data = task_wav$data()
  expect_task(task_wav)
  expect_shape(new_data, dim = c(task$nrow, 362L))
  expect_match(setdiff(names(new_data), c("heatan", "h2o")), "_wav_la8_[0-9]+$")

  pop = po("fda.wavelets", filter = "haar", boundary = "reflection")
  task_wav = train_pipeop(pop, list(task))[[1L]]
  new_data = task_wav$data()
  expect_task(task_wav)
  walk(new_data, expect_numeric)
  expect_shape(new_data, dim = c(task$nrow, 726L))
  expect_match(setdiff(names(new_data), c("heatan", "h2o")), "_wav_haar_[0-9]+$")

  # irregular data works
  task = tsk("dti")
  task$select(setdiff(task$feature_names, "sex"))
  pop = po("fda.wavelets")
  task_wav = train_pipeop(pop, list(task))[[1L]]
  new_data = task_wav$data()
  expect_task(task_wav)
  walk(new_data, expect_numeric)
  expect_shape(new_data, dim = c(task$nrow, 144L))
  expect_match(setdiff(names(new_data), "pasat"), "_wav_la8_[0-9]+$")
})

test_that("PipeOpFDASmooth - basic properties", {
  pop = po("fda.smooth")
  expect_pipeop(pop)
  expect_identical(pop$id, "fda.smooth")
})

test_that("PipeOpFDASmooth", {
  test_cca = function(method, args) {
    po_smooth = po("fda.smooth")
    x = task2$data(cols = "cca")$cca

    po_smooth$param_set$set_values(method = method, args = args)
    observed_train = train_pipeop(po_smooth, list(task2))[[1L]]$data(cols = "cca")[[1L]]
    observed_predict = predict_pipeop(po_smooth, list(task2))[[1L]]$data(cols = "cca")[[1L]]

    expected = suppressMessages(invoke(tf::tf_smooth, x = x, method = method, .args = args))

    expect_equal(observed_train, expected)
    expect_equal(observed_predict, expected)
  }

  test_nir = function(method, args) {
    po_smooth = po("fda.smooth")
    x = task1$data(cols = "NIR")$NIR

    po_smooth$param_set$set_values(method = method, args = args)
    observed_train = train_pipeop(po_smooth, list(task1))[[1L]]$data(cols = "NIR")[[1L]]
    observed_predict = predict_pipeop(po_smooth, list(task1))[[1L]]$data(cols = "NIR")[[1L]]

    expected = suppressMessages(invoke(tf::tf_smooth, x = x, method = method, .args = args))

    expect_equal(observed_train, expected)
    expect_equal(observed_predict, expected)
  }

  # regular
  task1 = tsk("fuel")
  task1$col_roles$feature = "NIR"
  # irregular
  task2 = tsk("dti")
  task2$col_roles$feature = "cca"

  test_nir("lowess", list(f = 0.3))
  test_cca("lowess", list(f = 0.2))

  test_nir("rollmean", list(k = 5))
  test_nir("rollmedian", list())
  test_nir("savgol", list())

  # verbose parameter is respected
  po_smooth = po("fda.smooth")
  po_smooth$param_set$set_values(verbose = TRUE)
  expect_snapshot(train_pipeop(po_smooth, list(task1)))
  po_smooth = po("fda.smooth")
  po_smooth$param_set$set_values(verbose = FALSE)
  expect_no_message(train_pipeop(po_smooth, list(task1)))
})

# PipeOpFDASmooth

    Code
      train_pipeop(po_smooth, list(task1))
    Message
      Using `f = 0.15` as smoother span for `lowess()`.
    Output
      $output
      
      -- <TaskRegr> (129x2): Spectral Data of Fossil Fuels ---------------------------
      * Target: heatan
      * Properties: -
      * Features (1):
        * tfr (1): NIR
      

# PipeOpFDAWavelets input validation

    Code
      po("fda.wavelets", filter = "la4")
    Condition
      Error in `self$assert()`:
      ! Assertion on 'xs' failed: filter: Must be element of set {'d2','d4','d6','d8','d10','d12','d14','d16','d18','d20','la8','la10','la12','la14','la16','la18','la20','bl14','bl18','bl20','c6','c12','c18','c24','c30','haar'}, but is 'la4'.

---

    Code
      po("fda.wavelets", filter = "invalid_filter")
    Condition
      Error in `self$assert()`:
      ! Assertion on 'xs' failed: filter: Must be element of set {'d2','d4','d6','d8','d10','d12','d14','d16','d18','d20','la8','la10','la12','la14','la16','la18','la20','bl14','bl18','bl20','c6','c12','c18','c24','c30','haar'}, but is 'invalid_filter'.

---

    Code
      po("fda.wavelets", filter = c(1, 2, 3))
    Condition
      Error in `self$assert()`:
      ! Assertion on 'xs' failed: filter: Must be either a string, an even numeric vector or wavelet filter object.

---

    Code
      po("fda.wavelets", filter = list("la8"))
    Condition
      Error in `self$assert()`:
      ! Assertion on 'xs' failed: filter: Must be either a string, an even numeric vector or wavelet filter object.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PipeOpFDAScaleRange.R
\name{mlr_pipeops_fda.scalerange}
\alias{mlr_pipeops_fda.scalerange}
\alias{PipeOpFDAScaleRange}
\title{Linearly Transform the Domain of Functional Data}
\description{
Linearly transform the domain of functional data so they are between \code{lower} and \code{upper}.
The formula for this is \eqn{x' = offset + x * scale},
where \eqn{scale} is \eqn{(upper - lower) / (max(x) - min(x))} and
\eqn{offset} is \eqn{-min(x) * scale + lower}. The same transformation is applied during training and prediction.
}
\section{Parameters}{

The parameters are the parameters inherited from \code{\link[mlr3pipelines:PipeOpTaskPreproc]{PipeOpTaskPreproc}},
as well as the following parameters:
\itemize{
\item \code{lower} :: \code{numeric(1)}\cr
Target value of smallest item of input data. Initialized to \code{0}.
\item \code{upper} :: \code{numeric(1)}\cr
Target value of greatest item of input data. Initialized to \code{1}.
}
}

\examples{
task = tsk("fuel")
po_scale = po("fda.scalerange", lower = -1, upper = 1)
task_scale = po_scale$train(list(task))[[1L]]
task_scale$data()
}
\section{Super classes}{
\code{\link[mlr3pipelines:PipeOp]{mlr3pipelines::PipeOp}} -> \code{\link[mlr3pipelines:PipeOpTaskPreproc]{mlr3pipelines::PipeOpTaskPreproc}} -> \code{PipeOpFDAScaleRange}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-PipeOpFDAScaleRange-initialize}{\code{PipeOpFDAScaleRange$new()}}
    \item \href{#method-PipeOpFDAScaleRange-clone}{\code{PipeOpFDAScaleRange$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="help"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-help'><code>mlr3pipelines::PipeOp$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="predict"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-predict'><code>mlr3pipelines::PipeOp$predict()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="print"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-print'><code>mlr3pipelines::PipeOp$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="train"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-train'><code>mlr3pipelines::PipeOp$train()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PipeOpFDAScaleRange-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-PipeOpFDAScaleRange-initialize}{}}}
\subsection{\code{PipeOpFDAScaleRange$new()}}{
  Initializes a new instance of this Class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PipeOpFDAScaleRange$new(id = "fda.scalerange", param_vals = list())}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{id}}{(\code{character(1)})\cr
Identifier of resulting object, default \code{"fda.scalerange"}.}
      \item{\code{param_vals}}{(named \code{list()})\cr
List of hyperparameter settings, overwriting the hyperparameter settings that would
otherwise be set during construction. Default \code{list()}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PipeOpFDAScaleRange-clone"></a>}}
\if{latex}{\out{\hypertarget{method-PipeOpFDAScaleRange-clone}{}}}
\subsection{\code{PipeOpFDAScaleRange$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PipeOpFDAScaleRange$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PipeOpFDAFourier.R
\name{mlr_pipeops_fda.fourier}
\alias{mlr_pipeops_fda.fourier}
\alias{PipeOpFDAFourier}
\title{Fast Fourier Transform Features}
\description{
This \code{PipeOp} extracts Fourier coefficients from functional columns.
For more details, see \code{\link[stats:fft]{stats::fft()}}, which is called internally.
Only the one-sided spectrum is returned since the input is real-valued (Oppenheim and Schafer, 2010).
}
\section{Parameters}{

The parameters are the parameters inherited from \code{\link[mlr3pipelines:PipeOpTaskPreprocSimple]{PipeOpTaskPreprocSimple}},
as well as the following parameters:
\itemize{
\item \code{type} :: \code{character(1)}\cr
Which feature to extract from the Fourier coefficients. \code{"amplitude"} returns the magnitude.
\code{"phase"} returns the phase shift in degrees (values in [-180, 180]). Initial value is \code{"phase"}.
}
}

\examples{
task = tsk("fuel")
po_fourier = po("fda.fourier")
task_fourier = po_fourier$train(list(task))[[1L]]
task_fourier$data()
}
\references{
Oppenheim, V A, Schafer, W R (2010).
\emph{Discrete-Time Signal Processing}, 3rd edition.
Pearson.
ISBN 978-0131988422.
}
\section{Super classes}{
\code{\link[mlr3pipelines:PipeOp]{mlr3pipelines::PipeOp}} -> \code{\link[mlr3pipelines:PipeOpTaskPreproc]{mlr3pipelines::PipeOpTaskPreproc}} -> \code{\link[mlr3pipelines:PipeOpTaskPreprocSimple]{mlr3pipelines::PipeOpTaskPreprocSimple}} -> \code{PipeOpFDAFourier}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-PipeOpFDAFourier-initialize}{\code{PipeOpFDAFourier$new()}}
    \item \href{#method-PipeOpFDAFourier-clone}{\code{PipeOpFDAFourier$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="help"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-help'><code>mlr3pipelines::PipeOp$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="predict"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-predict'><code>mlr3pipelines::PipeOp$predict()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="print"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-print'><code>mlr3pipelines::PipeOp$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="train"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-train'><code>mlr3pipelines::PipeOp$train()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PipeOpFDAFourier-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-PipeOpFDAFourier-initialize}{}}}
\subsection{\code{PipeOpFDAFourier$new()}}{
  Initializes a new instance of this Class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PipeOpFDAFourier$new(id = "fda.fourier", param_vals = list())}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{id}}{(\code{character(1)})\cr
Identifier of resulting object, default \code{"fda.fourier"}.}
      \item{\code{param_vals}}{(named \code{list()})\cr
List of hyperparameter settings, overwriting the hyperparameter settings that would
otherwise be set during construction. Default \code{list()}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PipeOpFDAFourier-clone"></a>}}
\if{latex}{\out{\hypertarget{method-PipeOpFDAFourier-clone}{}}}
\subsection{\code{PipeOpFDAFourier$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PipeOpFDAFourier$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PipeOpFDAFlatten.R
\name{mlr_pipeops_fda.flatten}
\alias{mlr_pipeops_fda.flatten}
\alias{PipeOpFDAFlatten}
\title{Flatten Functional Columns}
\description{
Convert regular functional features (e.g. all individuals are observed at the same time-points)
to new columns, one for each input value to the function.
}
\section{Parameters}{

The parameters are the parameters inherited from \code{\link[mlr3pipelines:PipeOpTaskPreprocSimple]{PipeOpTaskPreprocSimple}}.
}

\section{Naming}{

The new names generally append \verb{_1}, \verb{_2}, ... to the corresponding column name.
However this can lead to name clashes with existing columns.
This is solved as follows:
If a column was called \code{"x"}, the corresponding new columns will
be called \code{"x_1"}, \code{"x_2"}, etc. In case of duplicates, unique names are obtained using \code{make.unique()} and
a warning is given.
}

\examples{
task = tsk("fuel")
pop = po("fda.flatten")
task_flat = pop$train(list(task))[[1L]]
}
\section{Super classes}{
\code{\link[mlr3pipelines:PipeOp]{mlr3pipelines::PipeOp}} -> \code{\link[mlr3pipelines:PipeOpTaskPreproc]{mlr3pipelines::PipeOpTaskPreproc}} -> \code{\link[mlr3pipelines:PipeOpTaskPreprocSimple]{mlr3pipelines::PipeOpTaskPreprocSimple}} -> \code{PipeOpFDAFlatten}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-PipeOpFDAFlatten-initialize}{\code{PipeOpFDAFlatten$new()}}
    \item \href{#method-PipeOpFDAFlatten-clone}{\code{PipeOpFDAFlatten$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="help"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-help'><code>mlr3pipelines::PipeOp$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="predict"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-predict'><code>mlr3pipelines::PipeOp$predict()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="print"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-print'><code>mlr3pipelines::PipeOp$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="train"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-train'><code>mlr3pipelines::PipeOp$train()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PipeOpFDAFlatten-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-PipeOpFDAFlatten-initialize}{}}}
\subsection{\code{PipeOpFDAFlatten$new()}}{
  Initializes a new instance of this Class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PipeOpFDAFlatten$new(id = "fda.flatten", param_vals = list())}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{id}}{(\code{character(1)})\cr
Identifier of resulting object, default \code{"fda.flatten"}.}
      \item{\code{param_vals}}{(named \code{list()})\cr
List of hyperparameter settings, overwriting the hyperparameter settings that would
otherwise be set during construction. Default \code{list()}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PipeOpFDAFlatten-clone"></a>}}
\if{latex}{\out{\hypertarget{method-PipeOpFDAFlatten-clone}{}}}
\subsection{\code{PipeOpFDAFlatten$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PipeOpFDAFlatten$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PipeOpFDARandomEffect.R
\name{mlr_pipeops_fda.random_effect}
\alias{mlr_pipeops_fda.random_effect}
\alias{PipeOpFDARandomEffect}
\title{Extract Random Effects from Functional Columns}
\description{
This is the class that extracts random effects, specifically random intercepts and
random slopes, from functional columns. This PipeOp fits a linear mixed model, specifically
a random intercept and random slope model, using the \code{\link[lme4:lmer]{lme4::lmer()}} function.
The target variable is the value of the functional feature which is regressed on the functional feature's argument
while subject id determines the grouping structure. After model estimation, the random effects are extracted and
assigned to the correct id.
}
\section{Parameters}{

The parameters are the parameters inherited from \code{\link[mlr3pipelines:PipeOpTaskPreprocSimple]{PipeOpTaskPreprocSimple}}.
}

\section{Naming}{

The new names append \verb{_random_intercept} and \verb{_random_slope} to the corresponding column name of the
functional feature.
}

\examples{
task = tsk("dti")
po_fre = po("fda.random_effect")
task_fre = po_fre$train(list(task))[[1L]]
}
\section{Super classes}{
\code{\link[mlr3pipelines:PipeOp]{mlr3pipelines::PipeOp}} -> \code{\link[mlr3pipelines:PipeOpTaskPreproc]{mlr3pipelines::PipeOpTaskPreproc}} -> \code{\link[mlr3pipelines:PipeOpTaskPreprocSimple]{mlr3pipelines::PipeOpTaskPreprocSimple}} -> \code{PipeOpFDARandomEffect}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-PipeOpFDARandomEffect-initialize}{\code{PipeOpFDARandomEffect$new()}}
    \item \href{#method-PipeOpFDARandomEffect-clone}{\code{PipeOpFDARandomEffect$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="help"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-help'><code>mlr3pipelines::PipeOp$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="predict"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-predict'><code>mlr3pipelines::PipeOp$predict()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="print"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-print'><code>mlr3pipelines::PipeOp$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="train"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-train'><code>mlr3pipelines::PipeOp$train()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PipeOpFDARandomEffect-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-PipeOpFDARandomEffect-initialize}{}}}
\subsection{\code{PipeOpFDARandomEffect$new()}}{
  Initializes a new instance of this Class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PipeOpFDARandomEffect$new(id = "fda.random_effect", param_vals = list())}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{id}}{(\code{character(1)})\cr
Identifier of resulting object, default \code{"fda.random_effect"}.}
      \item{\code{param_vals}}{(named \code{list()})\cr
List of hyperparameter settings, overwriting the hyperparameter settings that would
otherwise be set during construction. Default \code{list()}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PipeOpFDARandomEffect-clone"></a>}}
\if{latex}{\out{\hypertarget{method-PipeOpFDARandomEffect-clone}{}}}
\subsection{\code{PipeOpFDARandomEffect$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PipeOpFDARandomEffect$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TaskRegr_dti.R
\name{mlr_tasks_dti}
\alias{mlr_tasks_dti}
\title{Diffusion Tensor Imaging (DTI) Regression Task}
\format{
\link[R6:R6Class]{R6::R6Class} inheriting from \link[mlr3:TaskRegr]{mlr3::TaskRegr}.
}
\description{
This dataset contains two functional covariates and three scalar covariates. The goal is
to predict the PASAT score. \code{pasat} represents the PASAT score at each visit.
\code{subject_id} represents the subject ID. \code{cca} represents the fractional anisotropy tract profiles from the corpus
callosum. \code{sex} indicates subject's sex. \code{rcst} represents the fractional anisotropy tract profiles from the right
corticospinal tract. Rows containing NAs are removed.

This is a subset of the full dataset, which is contained in the package \code{refund}.
}
\section{Dictionary}{

This \link[mlr3:Task]{mlr3::Task} can be instantiated via the \link[mlr3misc:Dictionary]{dictionary} \link[mlr3:mlr_tasks]{mlr3::mlr_tasks} or with the associated sugar function \code{\link[mlr3:tsk]{mlr3::tsk()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_tasks$get("dti")
tsk("dti")
}\if{html}{\out{</div>}}
}

\section{Meta Information}{

\itemize{
\item Task type: \dQuote{regr}
\item Dimensions: 340x4
\item Properties: \dQuote{groups}
\item Has Missings: \code{FALSE}
\item Target: \dQuote{pasat}
\item Features: \dQuote{cca}, \dQuote{rcst}, \dQuote{sex}
}
}

\references{
Goldsmith, Jeff, Bobb, Jennifer, Crainiceanu, M C, Caffo, Brian, Reich, Daniel (2011).
\dQuote{Penalized functional regression.}
\emph{Journal of Computational and Graphical Statistics}, \bold{20}(4), 830--851.

Brain dataset courtesy of Gordon Kindlmann at the Scientific Computing and Imaging Institute, University of Utah,
and Andrew Alexander, W. M. Keck Laboratory for Functional Brain Imaging and Behavior, University of
Wisconsin-Madison.
}
\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter2/data_and_basic_modeling.html}
\item Package \CRANpkg{mlr3data} for more toy tasks.
\item Package \CRANpkg{mlr3oml} for downloading tasks from \url{https://www.openml.org}.
\item Package \CRANpkg{mlr3viz} for some generic visualizations.
\item \link[mlr3misc:Dictionary]{Dictionary} of \link[mlr3:Task]{Tasks}: \link[mlr3:mlr_tasks]{mlr3::mlr_tasks}
\item \code{as.data.table(mlr_tasks)} for a table of available \link[mlr3:Task]{Tasks} in the running session (depending on the loaded packages).
\item \CRANpkg{mlr3fselect} and \CRANpkg{mlr3filters} for feature selection and feature filtering.
\item Extension packages for additional task types:
\itemize{
\item Unsupervised clustering: \CRANpkg{mlr3cluster}
\item Probabilistic supervised regression and survival analysis: \url{https://mlr3proba.mlr-org.com/}.
}
}

Other Task:
\code{\link{mlr_tasks_fuel}},
\code{\link{mlr_tasks_phoneme}}
}
\concept{Task}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PipeOpFDAInterpol.R
\name{mlr_pipeops_fda.interpol}
\alias{mlr_pipeops_fda.interpol}
\alias{PipeOpFDAInterpol}
\title{Interpolate Functional Columns}
\description{
Interpolate functional features (e.g. all individuals are observed at different time-points) to a common grid.
This is useful if you want to compare functional features across observations.
The interpolation is done using the \code{tf} package. See \code{\link[tf:tfd]{tfd()}} for details.
}
\section{Parameters}{

The parameters are the parameters inherited from \code{\link[mlr3pipelines:PipeOpTaskPreprocSimple]{PipeOpTaskPreprocSimple}},
as well as the following parameters:
\itemize{
\item \code{grid} :: \code{character(1)} | \code{numeric()}\cr
The grid to use for interpolation.
If \code{grid} is numeric, it must be a sequence of values to use for the grid or a single value that
specifies the number of points to use for the grid, requires \code{left} and \code{right} to be specified in the latter case.
If \code{grid} is a character, it must be one of:
\itemize{
\item \code{"union"}: This option creates a grid based on the union of all argument points from the provided functional
features. This means that if the argument points across features are \(t_1, t_2, ..., t_n\), then the grid will
be the combined unique set of these points. This option is generally used when the argument points vary across
observations and a common grid is needed for comparison or further analysis.
\item \code{"intersect"}: Creates a grid using the intersection of all argument points of a feature.
This grid includes only those points that are common across all functional features,
facilitating direct comparison on a shared set of points.
\item \code{"minmax"}: Generates a grid within the range of the maximum of the minimum argument points to the minimum of the
maximum argument points across features.
This bounded grid encapsulates the argument point range common to all features.
Note: For regular functional data this has no effect as all argument points are the same.
Initial value is \code{"union"}.
}
\item \code{method} :: \code{character(1)}\cr
One of:
\itemize{
\item \code{"linear"}: applies linear interpolation without extrapolation (see \code{\link[tf:tf_approx_linear]{tf::tf_approx_linear()}}).
\item \code{"spline"}: applies cubic spline interpolation (see \code{\link[tf:tf_approx_spline]{tf::tf_approx_spline()}}).
\item \code{"fill_extend"}: applies linear interpolation with constant extrapolation (see \code{\link[tf:tf_approx_fill_extend]{tf::tf_approx_fill_extend()}}).
\item \code{"locf"}: applies "last observation carried forward" interpolation (see \code{\link[tf:tf_approx_locf]{tf::tf_approx_locf()}}).
\item \code{"nocb"}: applies "next observation carried backward" interpolation (see \code{\link[tf:tf_approx_nocb]{tf::tf_approx_nocb()}}).
}

Default is \code{"linear"}.
\item \code{left} :: \code{numeric()}\cr
The left boundary of the window.
The window is specified such that all values >=left and <=right are kept for the computations.
\item \code{right} :: \code{numeric()}\cr
The right boundary of the window.
}
}

\examples{
task = tsk("fuel")
pop = po("fda.interpol")
task_interpol = pop$train(list(task))[[1L]]
task_interpol$data()
}
\section{Super classes}{
\code{\link[mlr3pipelines:PipeOp]{mlr3pipelines::PipeOp}} -> \code{\link[mlr3pipelines:PipeOpTaskPreproc]{mlr3pipelines::PipeOpTaskPreproc}} -> \code{\link[mlr3pipelines:PipeOpTaskPreprocSimple]{mlr3pipelines::PipeOpTaskPreprocSimple}} -> \code{PipeOpFDAInterpol}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-PipeOpFDAInterpol-initialize}{\code{PipeOpFDAInterpol$new()}}
    \item \href{#method-PipeOpFDAInterpol-clone}{\code{PipeOpFDAInterpol$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="help"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-help'><code>mlr3pipelines::PipeOp$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="predict"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-predict'><code>mlr3pipelines::PipeOp$predict()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="print"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-print'><code>mlr3pipelines::PipeOp$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="train"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-train'><code>mlr3pipelines::PipeOp$train()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PipeOpFDAInterpol-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-PipeOpFDAInterpol-initialize}{}}}
\subsection{\code{PipeOpFDAInterpol$new()}}{
  Initializes a new instance of this Class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PipeOpFDAInterpol$new(id = "fda.interpol", param_vals = list())}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{id}}{(\code{character(1)})\cr
Identifier of resulting object, default \code{"fda.interpol"}.}
      \item{\code{param_vals}}{(named \code{list()})\cr
List of hyperparameter settings, overwriting the hyperparameter settings that would
otherwise be set during construction. Default \code{list()}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PipeOpFDAInterpol-clone"></a>}}
\if{latex}{\out{\hypertarget{method-PipeOpFDAInterpol-clone}{}}}
\subsection{\code{PipeOpFDAInterpol$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PipeOpFDAInterpol$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TaskRegr_fuel.R
\name{mlr_tasks_fuel}
\alias{mlr_tasks_fuel}
\title{Fuel Regression Task}
\format{
\link[R6:R6Class]{R6::R6Class} inheriting from \link[mlr3:TaskRegr]{mlr3::TaskRegr}.
}
\description{
This dataset contains two functional covariates and one scalar covariate. The goal is
to predict the heat value of some fuel based on the ultraviolet radiation spectrum and
infrared ray radiation and one scalar column called h2o.

This is a subset of the full dataset, which is contained in the package \code{FDboost}.
}
\section{Dictionary}{

This \link[mlr3:Task]{mlr3::Task} can be instantiated via the \link[mlr3misc:Dictionary]{dictionary} \link[mlr3:mlr_tasks]{mlr3::mlr_tasks} or with the associated sugar function \code{\link[mlr3:tsk]{mlr3::tsk()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_tasks$get("fuel")
tsk("fuel")
}\if{html}{\out{</div>}}
}

\section{Meta Information}{

\itemize{
\item Task type: \dQuote{regr}
\item Dimensions: 129x4
\item Properties: -
\item Has Missings: \code{FALSE}
\item Target: \dQuote{heatan}
\item Features: \dQuote{NIR}, \dQuote{UVVIS}, \dQuote{h2o}
}
}

\references{
Brockhaus, Sarah, Scheipl, Fabian, Hothorn, Torsten, Greven, Sonja (2015).
\dQuote{The functional linear array model.}
\emph{Statistical Modelling}, \bold{15}(3), 279--300.
}
\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter2/data_and_basic_modeling.html}
\item Package \CRANpkg{mlr3data} for more toy tasks.
\item Package \CRANpkg{mlr3oml} for downloading tasks from \url{https://www.openml.org}.
\item Package \CRANpkg{mlr3viz} for some generic visualizations.
\item \link[mlr3misc:Dictionary]{Dictionary} of \link[mlr3:Task]{Tasks}: \link[mlr3:mlr_tasks]{mlr3::mlr_tasks}
\item \code{as.data.table(mlr_tasks)} for a table of available \link[mlr3:Task]{Tasks} in the running session (depending on the loaded packages).
\item \CRANpkg{mlr3fselect} and \CRANpkg{mlr3filters} for feature selection and feature filtering.
\item Extension packages for additional task types:
\itemize{
\item Unsupervised clustering: \CRANpkg{mlr3cluster}
\item Probabilistic supervised regression and survival analysis: \url{https://mlr3proba.mlr-org.com/}.
}
}

Other Task:
\code{\link{mlr_tasks_dti}},
\code{\link{mlr_tasks_phoneme}}
}
\concept{Task}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PipeOpFDAZoom.R
\name{mlr_pipeops_fda.zoom}
\alias{mlr_pipeops_fda.zoom}
\alias{PipeOpFDAZoom}
\title{Zoom In/Out on Functional Columns}
\description{
Zoom in or out on functional features by restricting their domain to a specified window.
This operation extracts a subset of each function by defining new lower and upper boundaries,
effectively cropping the functional data to focus on a specific region of interest.
Calls \code{\link[tf:tf_zoom]{tf::tf_zoom()}} from package \CRANpkg{tf}.
}
\section{Parameters}{

The parameters are the parameters inherited from \code{\link[mlr3pipelines:PipeOpTaskPreprocSimple]{PipeOpTaskPreprocSimple}},
as well as the following parameters:
\itemize{
\item \code{begin} :: \code{numeric()}\cr
The lower limit of the domain. Can be a single value applied to all
functional columns, or a numeric of length equal to the number of observations.
The window includes all values where argument >= \code{begin}. If not specified,
defaults to the lower limit of each function's domain.
\item \code{end} :: \code{numeric()}\cr
The upper limit of the domain.
}
}

\examples{
task = tsk("fuel")
pop = po("fda.zoom", begin = 50, end = 100)
task_zoom = pop$train(list(task))[[1L]]
task_zoom$data()
}
\section{Super classes}{
\code{\link[mlr3pipelines:PipeOp]{mlr3pipelines::PipeOp}} -> \code{\link[mlr3pipelines:PipeOpTaskPreproc]{mlr3pipelines::PipeOpTaskPreproc}} -> \code{\link[mlr3pipelines:PipeOpTaskPreprocSimple]{mlr3pipelines::PipeOpTaskPreprocSimple}} -> \code{PipeOpFDAZoom}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-PipeOpFDAZoom-initialize}{\code{PipeOpFDAZoom$new()}}
    \item \href{#method-PipeOpFDAZoom-clone}{\code{PipeOpFDAZoom$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="help"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-help'><code>mlr3pipelines::PipeOp$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="predict"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-predict'><code>mlr3pipelines::PipeOp$predict()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="print"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-print'><code>mlr3pipelines::PipeOp$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="train"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-train'><code>mlr3pipelines::PipeOp$train()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PipeOpFDAZoom-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-PipeOpFDAZoom-initialize}{}}}
\subsection{\code{PipeOpFDAZoom$new()}}{
  Initializes a new instance of this Class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PipeOpFDAZoom$new(id = "fda.zoom", param_vals = list())}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{id}}{(\code{character(1)})\cr
Identifier of resulting object, default \code{"fda.zoom"}.}
      \item{\code{param_vals}}{(named \code{list()})\cr
List of hyperparameter settings, overwriting the hyperparameter settings that would
otherwise be set during construction. Default \code{list()}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PipeOpFDAZoom-clone"></a>}}
\if{latex}{\out{\hypertarget{method-PipeOpFDAZoom-clone}{}}}
\subsection{\code{PipeOpFDAZoom$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PipeOpFDAZoom$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TaskClassif_phoneme.R
\name{mlr_tasks_phoneme}
\alias{mlr_tasks_phoneme}
\title{Phoneme Classification Task}
\format{
\link[R6:R6Class]{R6::R6Class} inheriting from \link[mlr3:TaskClassif]{mlr3::TaskClassif}.
}
\description{
The task contains a single functional covariate and 5 equally big classes (aa, ao, dcl, iy, sh).
The aim is to predict the class of the phoneme from the functional feature, which is a
log-periodogram.\cr
This is a subset of the full dataset, which is contained in the package \code{fda.usc}.
}
\section{Dictionary}{

This \link[mlr3:Task]{mlr3::Task} can be instantiated via the \link[mlr3misc:Dictionary]{dictionary} \link[mlr3:mlr_tasks]{mlr3::mlr_tasks} or with the associated sugar function \code{\link[mlr3:tsk]{mlr3::tsk()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_tasks$get("phoneme")
tsk("phoneme")
}\if{html}{\out{</div>}}
}

\section{Meta Information}{

\itemize{
\item Task type: \dQuote{classif}
\item Dimensions: 250x2
\item Properties: \dQuote{multiclass}
\item Has Missings: \code{FALSE}
\item Target: \dQuote{class}
\item Features: \dQuote{X}
}
}

\references{
Ferraty, Frédéric, Vieu, Philippe (2003).
\dQuote{Curves discrimination: a nonparametric functional approach.}
\emph{Computational Statistics & Data Analysis}, \bold{44}(1-2), 161--173.
}
\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter2/data_and_basic_modeling.html}
\item Package \CRANpkg{mlr3data} for more toy tasks.
\item Package \CRANpkg{mlr3oml} for downloading tasks from \url{https://www.openml.org}.
\item Package \CRANpkg{mlr3viz} for some generic visualizations.
\item \link[mlr3misc:Dictionary]{Dictionary} of \link[mlr3:Task]{Tasks}: \link[mlr3:mlr_tasks]{mlr3::mlr_tasks}
\item \code{as.data.table(mlr_tasks)} for a table of available \link[mlr3:Task]{Tasks} in the running session (depending on the loaded packages).
\item \CRANpkg{mlr3fselect} and \CRANpkg{mlr3filters} for feature selection and feature filtering.
\item Extension packages for additional task types:
\itemize{
\item Unsupervised clustering: \CRANpkg{mlr3cluster}
\item Probabilistic supervised regression and survival analysis: \url{https://mlr3proba.mlr-org.com/}.
}
}

Other Task:
\code{\link{mlr_tasks_dti}},
\code{\link{mlr_tasks_fuel}}
}
\concept{Task}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\docType{package}
\name{mlr3fda-package}
\alias{mlr3fda}
\alias{mlr3fda-package}
\title{mlr3fda: Extending 'mlr3' to Functional Data Analysis}
\description{
Extends the 'mlr3' ecosystem to functional analysis by adding support for irregular and regular functional data as defined in the 'tf' package. The package provides 'PipeOps' for preprocessing functional columns and for extracting scalar features, thereby allowing standard machine learning algorithms to be applied afterwards. Available operations include simple functional features such as the mean or maximum, smoothing, interpolation, flattening, and functional 'PCA'.
}
\section{Data types}{

To extend mlr3 to functional data, two data types from the tf package are added:
\itemize{
\item \code{tfd_irreg} - Irregular functional data, i.e. the functions are observed for
potentially different inputs for each observation.
\item \code{tfd_reg} - Regular functional data, i.e. the functions are observed for the same input
for each individual.
}

Lang M, Binder M, Richter J, Schratz P, Pfisterer F, Coors S, Au Q, Casalicchio G, Kotthoff L, Bischl B (2019).
\dQuote{mlr3: A modern object-oriented machine learning framework in R.}
\emph{Journal of Open Source Software}.
\doi{10.21105/joss.01903}.
\url{https://joss.theoj.org/papers/10.21105/joss.01903}.
}

\seealso{
Useful links:
\itemize{
  \item \url{https://mlr3fda.mlr-org.com}
  \item \url{https://github.com/mlr-org/mlr3fda}
  \item Report bugs at \url{https://github.com/mlr-org/mlr3fda/issues}
}

}
\author{
\strong{Maintainer}: Maximilian Mücke \email{muecke.maximilian@gmail.com} (\href{https://orcid.org/0009-0000-9432-9795}{ORCID})

Authors:
\itemize{
  \item Maximilian Mücke \email{muecke.maximilian@gmail.com} (\href{https://orcid.org/0009-0000-9432-9795}{ORCID})
  \item Sebastian Fischer \email{sebf.fischer@gmail.com} (\href{https://orcid.org/0000-0002-9609-3197}{ORCID})
}

Other contributors:
\itemize{
  \item Fabian Scheipl \email{fabian.scheipl@googlemail.com} (\href{https://orcid.org/0000-0001-8172-3603}{ORCID}) [contributor]
  \item Bernd Bischl \email{bernd_bischl@gmx.net} (\href{https://orcid.org/0000-0001-6002-6980}{ORCID}) [contributor]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PipeOpFDABsignal.R
\name{mlr_pipeops_fda.bsignal}
\alias{mlr_pipeops_fda.bsignal}
\alias{PipeOpFDABsignal}
\title{B-spline Feature Extraction}
\description{
This \code{PipeOp} extracts features from functional data using B-spline basis functions.
The extracted features are B-spline coefficients that represent the functional data in the B-spline basis space.
For more details, see \code{\link[FDboost:bsignal]{FDboost::bsignal()}}, which is called internally.
}
\section{Parameters}{

The parameters are the parameters inherited from \code{\link[mlr3pipelines:PipeOpTaskPreprocSimple]{PipeOpTaskPreprocSimple}},
as well as the following parameters:
\itemize{
\item \code{inS} :: \code{character(1)}\cr
Type of effect in the covariate index: one of \code{"smooth"}, \code{"linear"}, \code{"constant"}. Default is \code{"smooth"}.
\item \code{knots} :: \code{numeric()}\cr
Either the number of interior knots or a vector of their positions.
\item \code{boundary.knots} :: \code{numeric(2)}\cr
Boundary points at which to anchor the B-spline basis.
Lower and upper boundary points for the spline basis. Defaults to the range of the data.
\item \code{degree} :: \code{integer(1)}\cr
The degree of the regression spline. Default is \code{3L}.
\item \code{differences} :: \code{integer(1)}\cr
Order of difference penalty. Default is \code{1L}.
\item \code{df} :: \code{numeric(1)}\cr
Trace of the hat matrix, controlling smoothness. Default is \code{4}.
\item \code{lambda} :: \code{any}\cr
Smoothing parameter of the penalty term.
\item \code{center} :: \code{logical(1)}\cr
Reparameterize the unpenalized part to zero-mean? Default is \code{FALSE}.
\item \code{cyclic} :: \code{logical(1)}\cr
If true the fitted coefficient function coincides at the boundaries.
\item \code{Z} :: \code{any}\cr
Custom transformation matrix for the spline design.
\item \code{penalty} :: \code{character(1)}\cr
The penalty type: \code{"ps"} (P-spline) or \code{"pss"} (shrinkage). Default is \code{"ps"}.
\item \code{check.ident} :: \code{logical(1)}\cr
Use checks for identifiability of the effect. Default is \code{FALSE}.
}
}

\examples{
task = tsk("fuel")
po_bsignal = po("fda.bsignal")
task_bsignal = po_bsignal$train(list(task))[[1L]]
task_bsignal$data()
}
\section{Super classes}{
\code{\link[mlr3pipelines:PipeOp]{mlr3pipelines::PipeOp}} -> \code{\link[mlr3pipelines:PipeOpTaskPreproc]{mlr3pipelines::PipeOpTaskPreproc}} -> \code{\link[mlr3pipelines:PipeOpTaskPreprocSimple]{mlr3pipelines::PipeOpTaskPreprocSimple}} -> \code{PipeOpFDABsignal}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-PipeOpFDABsignal-initialize}{\code{PipeOpFDABsignal$new()}}
    \item \href{#method-PipeOpFDABsignal-clone}{\code{PipeOpFDABsignal$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="help"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-help'><code>mlr3pipelines::PipeOp$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="predict"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-predict'><code>mlr3pipelines::PipeOp$predict()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="print"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-print'><code>mlr3pipelines::PipeOp$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="train"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-train'><code>mlr3pipelines::PipeOp$train()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PipeOpFDABsignal-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-PipeOpFDABsignal-initialize}{}}}
\subsection{\code{PipeOpFDABsignal$new()}}{
  Initializes a new instance of this Class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PipeOpFDABsignal$new(id = "fda.bsignal", param_vals = list())}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{id}}{(\code{character(1)})\cr
Identifier of resulting object, default \code{"fda.bsignal"}.}
      \item{\code{param_vals}}{(named \code{list()})\cr
List of hyperparameter settings, overwriting the hyperparameter settings that would
otherwise be set during construction. Default \code{list()}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PipeOpFDABsignal-clone"></a>}}
\if{latex}{\out{\hypertarget{method-PipeOpFDABsignal-clone}{}}}
\subsection{\code{PipeOpFDABsignal$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PipeOpFDABsignal$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PipeOpFDACor.R
\name{mlr_pipeops_fda.cor}
\alias{mlr_pipeops_fda.cor}
\alias{PipeOpFDACor}
\title{Cross-Correlation of Functional Data}
\description{
Calculates the cross-correlation between two functional vectors using \code{\link[tf:tf_crosscor]{tf::tf_crosscor()}}.
Note that it only operates on regular data and that the cross-correlation assumes that each column
has the same domain.

To apply this \code{PipeOp} to irregular data, convert it to a regular grid first using \code{\link{PipeOpFDAInterpol}}.
If you need to change the domain of the columns, use \code{\link{PipeOpFDAScaleRange}}.
}
\section{Parameters}{

The parameters are the parameters inherited from \code{\link[mlr3pipelines:PipeOpTaskPreprocSimple]{PipeOpTaskPreprocSimple}},
as well as the following parameters:
\itemize{
\item \code{arg} :: \code{numeric()}\cr
Grid to use for the cross-correlation.
}
}

\examples{
set.seed(1234L)
dt = data.table(y = 1:100, x1 = tf::tf_rgp(100L), x2 = tf::tf_rgp(100L))
task = as_task_regr(dt, target = "y")
po_cor = po("fda.cor")
task_cor = po_cor$train(list(task))[[1L]]
task_cor
}
\section{Super classes}{
\code{\link[mlr3pipelines:PipeOp]{mlr3pipelines::PipeOp}} -> \code{\link[mlr3pipelines:PipeOpTaskPreproc]{mlr3pipelines::PipeOpTaskPreproc}} -> \code{\link[mlr3pipelines:PipeOpTaskPreprocSimple]{mlr3pipelines::PipeOpTaskPreprocSimple}} -> \code{PipeOpFDACor}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-PipeOpFDACor-initialize}{\code{PipeOpFDACor$new()}}
    \item \href{#method-PipeOpFDACor-clone}{\code{PipeOpFDACor$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="help"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-help'><code>mlr3pipelines::PipeOp$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="predict"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-predict'><code>mlr3pipelines::PipeOp$predict()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="print"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-print'><code>mlr3pipelines::PipeOp$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="train"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-train'><code>mlr3pipelines::PipeOp$train()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PipeOpFDACor-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-PipeOpFDACor-initialize}{}}}
\subsection{\code{PipeOpFDACor$new()}}{
  Initializes a new instance of this Class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PipeOpFDACor$new(id = "fda.cor", param_vals = list())}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{id}}{(\code{character(1)})\cr
Identifier of resulting object, default \code{"fda.cor"}.}
      \item{\code{param_vals}}{(named \code{list()})\cr
List of hyperparameter settings, overwriting the hyperparameter settings that would
otherwise be set during construction. Default \code{list()}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PipeOpFDACor-clone"></a>}}
\if{latex}{\out{\hypertarget{method-PipeOpFDACor-clone}{}}}
\subsection{\code{PipeOpFDACor$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PipeOpFDACor$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PipeOpFDARegister.R
\name{mlr_pipeops_fda.register}
\alias{mlr_pipeops_fda.register}
\alias{PipeOpFDARegister}
\title{Register (Align) Functional Columns}
\description{
Aligns functional features by estimating warping functions against a template using \code{\link[tf:tf_register]{tf::tf_register()}}.
Registration reduces phase (horizontal) variability while preserving amplitude (vertical) variability,
which is useful when curves share a common shape but differ in the timing of their features.

During training, a template is learned for each functional column (either estimated iteratively from the
data or supplied via \code{args}). The template is stored as part of the \verb{$state} and reused at predict time
so that new observations are aligned to the same reference as the training data.

Supported methods are \code{"srvf"}, \code{"cc"}, and \code{"affine"}. The \code{"landmark"} method from \code{\link[tf:tf_register]{tf::tf_register()}}
is not supported because it requires per-observation landmark positions for both training and prediction
data, which does not fit a stateful preprocessing step. For landmark registration, use \code{\link[tf:tf_register]{tf::tf_register()}}
directly and feed the aligned data into the task.
}
\section{Parameters}{

The parameters are the parameters inherited from \code{\link[mlr3pipelines:PipeOpTaskPreproc]{PipeOpTaskPreproc}},
as well as the following parameters:
\itemize{
\item \code{method} :: \code{character(1)}\cr
Registration method. One of:
\itemize{
\item \code{"srvf"}: elastic registration using the Square Root Velocity Framework via \code{\link[fdasrvf:time_warping]{fdasrvf::time_warping()}}.
Requires regular grids and the \CRANpkg{fdasrvf} package. Default template is the Karcher mean.
\item \code{"cc"}: continuous-criterion registration with monotone spline warps. Requires regular grids.
Default template is the arithmetic mean.
\item \code{"affine"}: affine (shift and/or scale) registration with warps of the form \eqn{h(t) = a \cdot t + b}.
Supports regular and irregular grids. Default template is the arithmetic mean.
}

Default is \code{"srvf"}.
\item \code{args} :: named \code{list()}\cr
Method-specific arguments passed to \code{\link[tf:tf_register]{tf::tf_register()}} via \code{...}. See the help page of
\code{\link[tf:tf_estimate_warps]{tf::tf_estimate_warps()}} for valid arguments (e.g. \code{lambda} / \code{penalty_method} for \code{"srvf"};
\code{nbasis}, \code{lambda}, \code{crit}, \code{conv}, \code{iterlim} for \code{"cc"}; \code{type}, \code{shift_range}, \code{scale_range}
for \code{"affine"}). An optional \code{template} entry is honored at training time and stored in the state.
\item \code{max_iter} :: \code{integer(1)}\cr
Maximum number of Procrustes-style template refinement iterations during training. Default is \code{3}.
Ignored at predict time because the stored template is used directly.
\item \code{tol} :: \code{numeric(1)}\cr
Convergence tolerance for template refinement during training. Default is \code{0.01}.
}
}

\section{State}{

\verb{$state$templates} contains the learned template (as a length-1 \code{tf} vector) for each functional column.
}

\examples{
\donttest{
set.seed(1)
task = tsk("fuel")
po_reg = po("fda.register", method = "affine", args = list(type = "shift_scale"))
task_reg = po_reg$train(list(task))[[1L]]
task_reg$data(cols = c("NIR", "UVVIS"))
}
}
\section{Super classes}{
\code{\link[mlr3pipelines:PipeOp]{mlr3pipelines::PipeOp}} -> \code{\link[mlr3pipelines:PipeOpTaskPreproc]{mlr3pipelines::PipeOpTaskPreproc}} -> \code{PipeOpFDARegister}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-PipeOpFDARegister-initialize}{\code{PipeOpFDARegister$new()}}
    \item \href{#method-PipeOpFDARegister-clone}{\code{PipeOpFDARegister$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="help"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-help'><code>mlr3pipelines::PipeOp$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="predict"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-predict'><code>mlr3pipelines::PipeOp$predict()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="print"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-print'><code>mlr3pipelines::PipeOp$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="train"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-train'><code>mlr3pipelines::PipeOp$train()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PipeOpFDARegister-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-PipeOpFDARegister-initialize}{}}}
\subsection{\code{PipeOpFDARegister$new()}}{
  Initializes a new instance of this Class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PipeOpFDARegister$new(id = "fda.register", param_vals = list())}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{id}}{(\code{character(1)})\cr
Identifier of resulting object, default \code{"fda.register"}.}
      \item{\code{param_vals}}{(named \code{list()})\cr
List of hyperparameter settings, overwriting the hyperparameter settings that would
otherwise be set during construction. Default \code{list()}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PipeOpFDARegister-clone"></a>}}
\if{latex}{\out{\hypertarget{method-PipeOpFDARegister-clone}{}}}
\subsection{\code{PipeOpFDARegister$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PipeOpFDARegister$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{dti}
\alias{dti}
\title{The dti dataset}
\format{
A data frame with 340 rows and 5 variables
}
\usage{
data(dti)
}
\description{
See \link{mlr_tasks_dti} for a description of the dataset.
}
\keyword{datasets}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PipeOpFDAExtract.R
\name{mlr_pipeops_fda.extract}
\alias{mlr_pipeops_fda.extract}
\alias{PipeOpFDAExtract}
\title{Extract Simple Features from Functional Columns}
\description{
This is the class that extracts simple features from functional columns.
Note that it only operates on values that were actually observed and does not interpolate.
}
\section{Parameters}{

The parameters are the parameters inherited from \code{\link[mlr3pipelines:PipeOpTaskPreprocSimple]{PipeOpTaskPreprocSimple}},
as well as the following parameters:
\itemize{
\item \code{drop} :: \code{logical(1)}\cr
Whether to drop the original \code{functional} features and only keep the extracted features.
Note that this does not remove the features from the backend, but only from the active
column role \code{feature}. Initial value is \code{TRUE}.
\item \code{features} :: \code{list()} | \code{character()}\cr
A list of features to extract. Each element can be either a function or a string.
If the element is a function it requires the following arguments: \code{arg} and \code{value} and returns a \code{numeric}.
For string elements, the following predefined features are available:
\code{"mean"}, \code{"max"}, \code{"min"}, \code{"slope"}, \code{"median"}, \code{"var"}.
Initial value is \code{c("mean", "max", "min", "slope", "median", "var")}.
\item \code{left} :: \code{numeric()}\cr
The left boundary of the window. Initial value is \code{-Inf}.
The window is specified such that all values >=left and <=right are kept for the computations.
\item \code{right} :: \code{numeric()}\cr
The right boundary of the window. Initial value is \code{Inf}.
}
}

\section{Naming}{

The new names generally append a \verb{_\{feature\}} to the corresponding column name.
However this can lead to name clashes with existing columns.
This is solved as follows:
If a column was called \code{"x"} and the feature is \code{"mean"}, the corresponding new column will
be called \code{"x_mean"}. In case of duplicates, unique names are obtained using \code{make.unique()} and
a warning is given.
}

\examples{
task = tsk("fuel")
po_fmean = po("fda.extract", features = "mean")
task_fmean = po_fmean$train(list(task))[[1L]]

# add more than one feature
pop = po("fda.extract", features = c("mean", "median", "var"))
task_features = pop$train(list(task))[[1L]]

# add a custom feature
po_custom = po("fda.extract",
  features = list(mean = function(arg, value) mean(value, na.rm = TRUE))
)
task_custom = po_custom$train(list(task))[[1L]]
task_custom
}
\section{Super classes}{
\code{\link[mlr3pipelines:PipeOp]{mlr3pipelines::PipeOp}} -> \code{\link[mlr3pipelines:PipeOpTaskPreproc]{mlr3pipelines::PipeOpTaskPreproc}} -> \code{\link[mlr3pipelines:PipeOpTaskPreprocSimple]{mlr3pipelines::PipeOpTaskPreprocSimple}} -> \code{PipeOpFDAExtract}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-PipeOpFDAExtract-initialize}{\code{PipeOpFDAExtract$new()}}
    \item \href{#method-PipeOpFDAExtract-clone}{\code{PipeOpFDAExtract$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="help"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-help'><code>mlr3pipelines::PipeOp$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="predict"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-predict'><code>mlr3pipelines::PipeOp$predict()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="print"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-print'><code>mlr3pipelines::PipeOp$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="train"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-train'><code>mlr3pipelines::PipeOp$train()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PipeOpFDAExtract-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-PipeOpFDAExtract-initialize}{}}}
\subsection{\code{PipeOpFDAExtract$new()}}{
  Initializes a new instance of this Class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PipeOpFDAExtract$new(id = "fda.extract", param_vals = list())}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{id}}{(\code{character(1)})\cr
Identifier of resulting object, default \code{"fda.extract"}.}
      \item{\code{param_vals}}{(named \code{list()})\cr
List of hyperparameter settings, overwriting the hyperparameter settings that would
otherwise be set during construction. Default \code{list()}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PipeOpFDAExtract-clone"></a>}}
\if{latex}{\out{\hypertarget{method-PipeOpFDAExtract-clone}{}}}
\subsection{\code{PipeOpFDAExtract$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PipeOpFDAExtract$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PipeOpFDATsfeatures.R
\name{mlr_pipeops_fda.tsfeats}
\alias{mlr_pipeops_fda.tsfeats}
\alias{PipeOpFDATsfeatures}
\title{Time Series Feature Extraction}
\description{
This \code{PipeOp} extracts time series features from functional columns.

For more details, see \code{\link[tsfeatures:tsfeatures]{tsfeatures::tsfeatures()}}, which is called internally.
}
\section{Parameters}{

The parameters are the parameters inherited from \code{\link[mlr3pipelines:PipeOpTaskPreprocSimple]{PipeOpTaskPreprocSimple}},
as well as the following parameters:
\itemize{
\item \code{features} :: \code{character()}\cr
Function names which return numeric vectors of features.
All features returned by these functions must be named if they return more than one feature.
Default is \code{c("frequency", "stl_features", "entropy", "acf_features")}.
\item \code{scale} :: \code{logical(1)}\cr
If \code{TRUE}, data is scaled to mean 0 and sd 1 before features are computed. Default is \code{TRUE}.
\item \code{trim} :: \code{logical(1)}\cr
If \code{TRUE}, data is trimmed by \code{trim_amount} before features are computed.
Values larger than \code{trim_amount} in absolute value are set to \code{NA}. Default is \code{FALSE}.
\item \code{trim_amount} :: \code{numeric(1)}\cr
Default level of trimming. Default is \code{0.1}.
\item \code{parallel} :: \code{logical(1)}\cr
If \code{TRUE}, the features are computed in parallel. Default is \code{FALSE}.
\item \code{multiprocess} :: \code{any}\cr
The function from the future package to use for parallel processing. Default is \code{\link[future:multisession]{future::multisession()}}.
\item \code{na.action} :: \code{any}\cr
A function to handle missing values. Default is \code{\link[stats:na.pass]{stats::na.pass()}}.
}
}

\section{Naming}{

The new names generally append a \verb{_\{feature\}} to the corresponding column name.
If a column was called \code{"x"} and the feature is \code{"trend"}, the corresponding new column will
be called \code{"x_trend"}.
}

\examples{
task = tsk("fuel")
po_tsfeats = po("fda.tsfeats")
task_tsfeats = po_tsfeats$train(list(task))[[1L]]
task_tsfeats$data()
}
\section{Super classes}{
\code{\link[mlr3pipelines:PipeOp]{mlr3pipelines::PipeOp}} -> \code{\link[mlr3pipelines:PipeOpTaskPreproc]{mlr3pipelines::PipeOpTaskPreproc}} -> \code{\link[mlr3pipelines:PipeOpTaskPreprocSimple]{mlr3pipelines::PipeOpTaskPreprocSimple}} -> \code{PipeOpFDATsfeatures}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-PipeOpFDATsfeatures-initialize}{\code{PipeOpFDATsfeatures$new()}}
    \item \href{#method-PipeOpFDATsfeatures-clone}{\code{PipeOpFDATsfeatures$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="help"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-help'><code>mlr3pipelines::PipeOp$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="predict"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-predict'><code>mlr3pipelines::PipeOp$predict()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="print"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-print'><code>mlr3pipelines::PipeOp$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="train"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-train'><code>mlr3pipelines::PipeOp$train()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PipeOpFDATsfeatures-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-PipeOpFDATsfeatures-initialize}{}}}
\subsection{\code{PipeOpFDATsfeatures$new()}}{
  Initializes a new instance of this Class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PipeOpFDATsfeatures$new(id = "fda.tsfeats", param_vals = list())}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{id}}{(\code{character(1)})\cr
Identifier of resulting object, default \code{"fda.tsfeats"}.}
      \item{\code{param_vals}}{(named \code{list()})\cr
List of hyperparameter settings, overwriting the hyperparameter settings that would
otherwise be set during construction. Default \code{list()}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PipeOpFDATsfeatures-clone"></a>}}
\if{latex}{\out{\hypertarget{method-PipeOpFDATsfeatures-clone}{}}}
\subsection{\code{PipeOpFDATsfeatures$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PipeOpFDATsfeatures$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PipeOpFPCA.R
\name{mlr_pipeops_fda.fpca}
\alias{mlr_pipeops_fda.fpca}
\alias{PipeOpFPCA}
\title{Functional Principal Component Analysis}
\description{
This \code{PipeOp} applies a functional principal component analysis (FPCA) to functional columns and then
extracts the principal components as features. This is done using a (truncated) weighted SVD.

To apply this \code{PipeOp} to irregular data, convert it to a regular grid first using \code{\link{PipeOpFDAInterpol}}.

For more details, see \code{\link[tf:tfb_fpc]{tf::tfb_fpc()}}, which is called internally.
}
\section{Parameters}{

The parameters are the parameters inherited from \code{\link[mlr3pipelines:PipeOpTaskPreproc]{PipeOpTaskPreproc}},
as well as the following parameters:
\itemize{
\item \code{pve} :: \code{numeric(1)}\cr
The percentage of variance explained that should be retained. Default is \code{0.995}.
\item \code{n_components} :: \code{integer(1)}\cr
The number of principal components to extract. This parameter is initialized to \code{Inf}.
}
}

\section{Naming}{

The new names generally append a \verb{_pc_\{number\}} to the corresponding column name.
If a column was called \code{"x"} and there are three principal components, the corresponding
new columns will be called \verb{"x_pc_1", "x_pc_2", "x_pc_3"}.
}

\examples{
task = tsk("fuel")
po_fpca = po("fda.fpca", n_components = 3L)
task_fpca = po_fpca$train(list(task))[[1L]]
task_fpca$data()
}
\section{Super classes}{
\code{\link[mlr3pipelines:PipeOp]{mlr3pipelines::PipeOp}} -> \code{\link[mlr3pipelines:PipeOpTaskPreproc]{mlr3pipelines::PipeOpTaskPreproc}} -> \code{PipeOpFPCA}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-PipeOpFPCA-initialize}{\code{PipeOpFPCA$new()}}
    \item \href{#method-PipeOpFPCA-clone}{\code{PipeOpFPCA$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="help"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-help'><code>mlr3pipelines::PipeOp$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="predict"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-predict'><code>mlr3pipelines::PipeOp$predict()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="print"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-print'><code>mlr3pipelines::PipeOp$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="train"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-train'><code>mlr3pipelines::PipeOp$train()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PipeOpFPCA-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-PipeOpFPCA-initialize}{}}}
\subsection{\code{PipeOpFPCA$new()}}{
  Initializes a new instance of this Class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PipeOpFPCA$new(id = "fda.fpca", param_vals = list())}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{id}}{(\code{character(1)})\cr
Identifier of resulting object, default \code{"fda.fpca"}.}
      \item{\code{param_vals}}{(named \code{list()})\cr
List of hyperparameter settings, overwriting the hyperparameter settings that would
otherwise be set during construction. Default \code{list()}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PipeOpFPCA-clone"></a>}}
\if{latex}{\out{\hypertarget{method-PipeOpFPCA-clone}{}}}
\subsection{\code{PipeOpFPCA$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PipeOpFPCA$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PipeOpFDAWavelets.R
\name{mlr_pipeops_fda.wavelets}
\alias{mlr_pipeops_fda.wavelets}
\alias{PipeOpFDAWavelets}
\title{Discrete Wavelet Transform Features}
\description{
This \code{PipeOp} extracts discrete wavelet transform coefficients from functional columns.
For more details, see \code{\link[wavelets:dwt]{wavelets::dwt()}}, which is called internally.
}
\section{Parameters}{

The parameters are the parameters inherited from \code{\link[mlr3pipelines:PipeOpTaskPreprocSimple]{PipeOpTaskPreprocSimple}},
as well as the following parameters:
\itemize{
\item \code{filter} :: \code{character(1)} | \code{numeric()} | \code{\link[wavelets:wt.filter]{wavelets::wt.filter()}}\cr
Specifies which filter should be used. Must be either \code{\link[wavelets:wt.filter]{wavelets::wt.filter()}} object, an even numeric vector or a
string. In case of a string must be one of \code{"d"}|\code{"la"}|\code{"bl"}|\code{"c"} followed by an even number for the level of
the filter. The level of the filter needs to be smaller or equal than the time-series length.
For more information and acceptable filters see \code{help(wt.filter)}. Default is \code{"la8"}.
\item \code{n.levels} :: \code{integer(1)}\cr
An integer specifying the level of the decomposition.
\item \code{boundary} :: \code{character(1)}\cr
Boundary to be used. \code{"periodic"} assumes circular time series, for \code{"reflection"} the series is extended to twice
its length. Default is \code{"periodic"}.
\item \code{fast} :: \code{logical(1)}\cr
Should the pyramid algorithm be calculated with an internal C function? Default is \code{TRUE}.
}
}

\examples{
task = tsk("fuel")
po_wavelets = po("fda.wavelets")
task_wavelets = po_wavelets$train(list(task))[[1L]]
task_wavelets$data()
}
\section{Super classes}{
\code{\link[mlr3pipelines:PipeOp]{mlr3pipelines::PipeOp}} -> \code{\link[mlr3pipelines:PipeOpTaskPreproc]{mlr3pipelines::PipeOpTaskPreproc}} -> \code{\link[mlr3pipelines:PipeOpTaskPreprocSimple]{mlr3pipelines::PipeOpTaskPreprocSimple}} -> \code{PipeOpFDAWavelets}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-PipeOpFDAWavelets-initialize}{\code{PipeOpFDAWavelets$new()}}
    \item \href{#method-PipeOpFDAWavelets-clone}{\code{PipeOpFDAWavelets$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="help"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-help'><code>mlr3pipelines::PipeOp$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="predict"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-predict'><code>mlr3pipelines::PipeOp$predict()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="print"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-print'><code>mlr3pipelines::PipeOp$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="train"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-train'><code>mlr3pipelines::PipeOp$train()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PipeOpFDAWavelets-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-PipeOpFDAWavelets-initialize}{}}}
\subsection{\code{PipeOpFDAWavelets$new()}}{
  Initializes a new instance of this Class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PipeOpFDAWavelets$new(id = "fda.wavelets", param_vals = list())}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{id}}{(\code{character(1)})\cr
Identifier of resulting object, default \code{"fda.wavelets"}.}
      \item{\code{param_vals}}{(named \code{list()})\cr
List of hyperparameter settings, overwriting the hyperparameter settings that would
otherwise be set during construction. Default \code{list()}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PipeOpFDAWavelets-clone"></a>}}
\if{latex}{\out{\hypertarget{method-PipeOpFDAWavelets-clone}{}}}
\subsection{\code{PipeOpFDAWavelets$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PipeOpFDAWavelets$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PipeOpFDASmooth.R
\name{mlr_pipeops_fda.smooth}
\alias{mlr_pipeops_fda.smooth}
\alias{PipeOpFDASmooth}
\title{Smooth Functional Columns}
\description{
Smoothes functional data using \code{\link[tf:tf_smooth]{tf::tf_smooth()}}.
This preprocessing operator is similar to \code{\link{PipeOpFDAInterpol}}, however it does not interpolate to unobserved
x-values, but rather smooths the observed values.
}
\section{Parameters}{

The parameters are the parameters inherited from \code{\link[mlr3pipelines:PipeOpTaskPreprocSimple]{PipeOpTaskPreprocSimple}},
as well as the following parameters:
\itemize{
\item \code{method} :: \code{character(1)}\cr
One of:
\itemize{
\item \code{"lowess"}: locally weighted scatterplot smoothing (default)
\item \code{"rollmean"}: rolling mean
\item \code{"rollmedian"}: rolling median
\item \code{"savgol"}: Savitzky-Golay filtering
}

All methods but \code{"lowess"} ignore non-equidistant arg values.
\item \code{args} :: named \code{list()}\cr
List of named arguments that is passed to \code{tf_smooth()}. See the help page of \code{tf_smooth()} for
default values.
\item \code{verbose} :: \code{logical(1)}\cr
Whether to print messages during the transformation.
Is initialized to \code{FALSE}.
}
}

\examples{
task = tsk("fuel")
po_smooth = po("fda.smooth", method = "rollmean", args = list(k = 5))
task_smooth = po_smooth$train(list(task))[[1L]]
task_smooth
task_smooth$data(cols = c("NIR", "UVVIS"))
}
\section{Super classes}{
\code{\link[mlr3pipelines:PipeOp]{mlr3pipelines::PipeOp}} -> \code{\link[mlr3pipelines:PipeOpTaskPreproc]{mlr3pipelines::PipeOpTaskPreproc}} -> \code{\link[mlr3pipelines:PipeOpTaskPreprocSimple]{mlr3pipelines::PipeOpTaskPreprocSimple}} -> \code{PipeOpFDASmooth}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-PipeOpFDASmooth-initialize}{\code{PipeOpFDASmooth$new()}}
    \item \href{#method-PipeOpFDASmooth-clone}{\code{PipeOpFDASmooth$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="help"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-help'><code>mlr3pipelines::PipeOp$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="predict"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-predict'><code>mlr3pipelines::PipeOp$predict()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="print"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-print'><code>mlr3pipelines::PipeOp$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3pipelines" data-topic="PipeOp" data-id="train"><a href='../../mlr3pipelines/html/PipeOp.html#method-PipeOp-train'><code>mlr3pipelines::PipeOp$train()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PipeOpFDASmooth-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-PipeOpFDASmooth-initialize}{}}}
\subsection{\code{PipeOpFDASmooth$new()}}{
  Initializes a new instance of this Class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PipeOpFDASmooth$new(id = "fda.smooth", param_vals = list())}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{id}}{(\code{character(1)})\cr
Identifier of resulting object, default \code{"fda.smooth"}.}
      \item{\code{param_vals}}{(named \code{list()})\cr
List of hyperparameter settings, overwriting the hyperparameter settings that would
otherwise be set during construction. Default \code{list()}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PipeOpFDASmooth-clone"></a>}}
\if{latex}{\out{\hypertarget{method-PipeOpFDASmooth-clone}{}}}
\subsection{\code{PipeOpFDASmooth$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PipeOpFDASmooth$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{fuel}
\alias{fuel}
\title{The fuel dataset}
\format{
A data frame with 129 rows and 4 variables
}
\usage{
data(fuel)
}
\description{
See \link{mlr_tasks_fuel} for a description of the dataset.
}
\keyword{datasets}
\keyword{internal}
